/*
 * $Id: Corpserv.cpp,v 1.12 2006-10-22 14:59:13 tonydodd Exp $
 *
 * $Log: not supported by cvs2svn $
 * Revision 1.10  1996/05/17 14:32:49  scotty
 * Caught the last remnant of the old logging scheme which was
 * trying to fflush() an unitialized FILE pointer.
 *
 * Revision 1.9  1996/05/13 01:23:55  scotty
 * v927 fixes from Tony.
 *
 * Revision 1.8  1996/05/12 22:13:55  scotty
 * Added a signal handler to catch children when they exit;
 * modified the call to accept() to handle EINTRs caused by
 * the handler.
 *
 * Revision 1.7  1996/03/01 23:54:11  scotty
 * Fixed call to setsockopt() -- the wrong size was being passed and
 * some compilers want a cast for &one.
 *
 * __isascii is now only defined if it's not already defined
 * (Linux defines it).
 *
 * Revision 1.6  1996/02/26 22:51:30  scotty
 * Added support for braindead (Ultrix) syslog()s.
 *
 * Revision 1.5  1996/02/26 21:05:57  scotty
 * Moved idxpath[] into here from corphash.c since it's only every
 * used in this file.
 *
 * Revision 1.4  1996/02/26 19:35:25  scotty
 * Most exit()s now print a message explaing why the server has quit.
 *
 * Revision 1.3  1996/02/15 01:26:38  scotty
 * Now writes the server PID in ACCPATH/sarad.pid.
 *
 * Revision 1.2  1996/02/14 22:01:04  scotty
 * Some fixes by Tony.
 *
 * Revision 1.1  1996/01/30 20:51:18  scotty
 * Initial revision
 *
 */
#include "Corpinc.h"

#ifndef HAVE_SOCKLEN_T
typedef int socklen_t;
#endif

extern "C" {


#undef isspace

#define SERVER 1
#define MAX_USER 20
#define MAX_PW 20
#define MAX_NAME 80
#define MAX_CQL 2000
#define MAX_REGEXP 20
#define MAX_FILTER 20
#define MAX_SOL (MAXSENTENCE+40)
#define MAXFILTERN 100

struct ran {
   int       min;
   int       size;
};


union uarg {
   char        user[MAX_USER];
   char        pw[MAX_PW];
   char        name[MAX_NAME];
   xairo_uchar		namew[MAX_NAME];
   int         num;
   sara_long   lnum;
   char        cql[MAX_CQL];
   xairo_uchar        cqlw[MAX_CQL];
   char        regexp[MAX_REGEXP];
   int         (*table)[];
   char        filter[MAX_FILTER];
   struct ran  range;
};


#define MAX_COMMAND 200
#define MAX_OPS 6

struct pcmd {
   int       i;
   union uarg arg[MAX_OPS];
};


struct cmd {
   char     *name;
   int       op[MAX_OPS];
};

#define NFILES 40

struct qt {
   BOOL      bUse;
   BOOL      bSave;
   char      name[12];
   char      path[_MAX_PATH];
} qtable[NFILES];

#define OP_NULL 0
#define OP_USER    1
#define OP_PW      2
#define OP_NAME    3
#define OP_NUM     4
#define OP_CQL     5
#define OP_REGEXP  6
#define OP_TABLE   7
#define OP_FILTER  8
#define OP_RANGE   9
#define OP_LNUM   10
#define OP_CQLW   11
#define OP_NAMEW  12


struct cmd ctable[] = {
   { "LOGOUT",       { 0,        0,          0,          0        }},
   { "LOG",          { OP_USER,  OP_PW,      0,          0        }},
   { "SAVE",         { OP_NUM,   OP_NAME,    0,          0        }},
   { "DIR",          { OP_NAME,  0,          0,          0        }},
   { "REMOVE",       { OP_NAME,  0,          0,          0        }},
   { "PURGE",        { OP_NUM,   0,          0,          0        }},
   { "SOLVE",        { OP_NAME,  OP_CQLW,     0,          0        }},
   { "SUBCORPUS",    { OP_NAME,  OP_CQL,     0,          0        }},
   { "WORDLIST",     { OP_NAME,  OP_REGEXP,  0,          0        }},
   { "SQTABLE",      { OP_NAME,  OP_NAME,    OP_TABLE,   0,       }},
   { "FILTER",       { OP_NAME,  OP_FILTER,  0,          0        }},
   { "SORTFILTER",   { OP_NAME,  OP_NUM,     OP_FILTER,  0        }},
   { "SORT",         { OP_NAME,  OP_NUM,     0,          0        }},
   { "GET",          { OP_NAME,  OP_NUM,     OP_NUM,     0        }},
   { "CUT",          { 0,        0,          0,          0        }},
   { "CHAR",         { OP_NUM,   OP_NAME,    0,          0        }},
   { "GETPOS",       { OP_NAMEW,  OP_NUM,          0,          0        }},
   { "GETDTD",       { OP_NUM,   0,          0,          0        }},
   { "MAXLENGTH",    { OP_NUM,   0,          0,          0        }},
   { "GETSC",        { OP_NUM,   0,          0,          0        }},
   { "GETHEAD",      { OP_NAME,  OP_LNUM,    OP_NUM,     0        }},
   { "LOOKUP",       { OP_CQLW,  0,          0,          0        }},
   { "DMATCH",       { OP_NUM,   0,          0,          0        }},
   { "QNAME",        { 0,        0,          0,          0        }},
   { "OPEN",         { OP_NAME,  0,          0,          0        }},
   { "CSCORE",       { OP_NAME,  OP_NUM,     OP_CQLW,     0        }},
   { "LOC",          { OP_NAME,  OP_NUM,     0,          0        }},
   { "GETHEAD2",     { OP_NAME,  OP_LNUM,    OP_LNUM,    OP_NUM   }},
   { "GETCHEAD",     { OP_NUM,   0,          0,          0        }},
   { "GET1SOL",      { OP_NAME,  OP_NAMEW,    OP_NAMEW,    OP_NAMEW        }},
   { "MOTD",         { 0,        0,          0,          0        }},
   { "PWD",          { OP_PW,    OP_PW,      0,          0        }},
   { "INFO",         { OP_NAME,  0,          0,          0        }},
   { "RLOOKUP",      { OP_CQLW,   0,          0,          0        }},
   { "RGET",         { OP_NUM,   0,          0,          0        }},
   { "BIB",          { OP_NAME,  0,          0,          0        }},
   { "BIBITEM",      { OP_NAME,  OP_NUM,     0,          0        }},
   { "TRACE",        { OP_NUM,   0,          0,          0        }},
   { "DOWNLOAD",     { OP_NAME,  0,          0,          0        }},
   { "LINE",         { 0,        0,          0,          0        }},
   { "THIN",         { OP_NAME,  OP_NAME,    OP_NUM,     OP_NUM   ,OP_NUM}},
   { "WEB" ,         { 0,        0,          0,          0,       0}},
   { "SCDEF",			{OP_NAME,OP_CQL,			0,			0,		0}},
   { "SCADD",			{OP_NAME, OP_TABLE,0,0,0}},
   { "SCSET",			{OP_NAME,OP_NUM}},
   { "FTAB", {OP_NUM,OP_NUM,OP_NUM,OP_CQLW}},
   { "FENTRY", {OP_NUM}},
   { "FFREE",{0}},
   { "SCSAVE",{0}},
   { "SCLIST",{OP_NUM}},
   { "SCENUM",{OP_NAME, OP_NUM, OP_NUM}},
   { "CTAB",{OP_NAME,OP_NUM,OP_NUM,OP_NUM,OP_CQLW}},
   { "CTABENTRY", {OP_NUM}},
   { "ACSCORE", {OP_NAME,OP_NUM,OP_NUM,OP_NUM,OP_CQLW}},
   { "CTABTHIN",{OP_NAME,OP_NAME,OP_NUM,OP_NUM,OP_NUM,OP_CQLW}},
   { "CTABOPTIONS",{OP_NUM,OP_NUM,OP_NAME,OP_NUM}},
   { "FORM",{OP_NUM,OP_CQLW}},
   { "LEMMINFO", {0}},
   { "LEMMALL", {0}},
   { "LEMMSEL", {OP_NAME}},
   { "GETSOL",          { OP_NAME,  OP_NUM,     OP_NUM,     0        }},
   { "CTABFREE",{0}},
   { "LEMMDESC",{OP_NAME}},
   { "LEMMTEST",{OP_NAME, OP_NAME, OP_NAME, OP_NUM}},
   { "SOLVEX",        { OP_NAME,  OP_CQLW,     0,          0        }},
   { "LOOKUPA", {OP_NAME, OP_NAME, 0, 0 }},
   { "ADICT",   {OP_NUM, 0, 0, 0}},
   { "SCCOUNT", {OP_NAME,OP_NAME,OP_NUM}},
   { "SCWORDS", {OP_NAME,OP_NUM}},
   { "RFREE",{0}},
   { "SCNAMES", {OP_NAME,OP_CQL}},
   { "SCENUMNAMES", {OP_NAME}},
   { "COUNTCONTENT",{OP_NAME,OP_NAME,OP_NUM,OP_NAME}},
   { "SUBFORM",{OP_NUM,OP_NUM,OP_CQLW}},
   { "GETPATH",{OP_NAME,OP_NUM}},
   { "CTABX",{OP_NAME,OP_NAME,OP_NUM,OP_CQLW}},
   { "CTABTHINX",{OP_NAME,OP_NAME,OP_NAME,OP_NUM,OP_CQLW}},
   { "AKSTAT",{OP_NAMEW,OP_NAMEW}},
   { "AKFILTER",{OP_NAMEW,OP_NAMEW}},
   { "AKCLEAR",{OP_NAMEW}}

};


struct cmd ftable[] = {
   { "NOSGMLX",   { 0,        0, 0, 0  }},
   { "NOPOSX",    { 0,        0, 0, 0  }},
   { "NORMSPACE", { 0,        0, 0, 0  }},
/*   { "LEFTL",     { OP_NUM,   0, 0, 0  }},
   { "ADDSLEN",   { OP_NUM,   0, 0, 0  }}, */
   { "CMAP",      { 0,        0, 0, 0  }},
/*   { "ADDPOS",    { OP_NUM,   0, 0, 0  }},
   { "ADDTEXT",   { OP_NUM,   0, 0, 0  }}, */
   { "NORMCR0",   { 0,        0, 0, 0  }},
   { "ADJSGM",   { 0,        0, 0, 0  }},
   { "ADJPOS",    { 0,        0, 0, 0  }},
   { "WTOPOS",    { 0,        0, 0, 0  }},
};


enum filtype {
   FIL_SGMLX, FIL_NOPOSX, FIL_NORMSPACE,
   /* FIL_LEFTL, FIL_ADDSLEN, */ FIL_CMAP, /* FIL_ADDPOS,
   FIL_ADDTEXT, */FIL_NORMCR0, FIL_ADJSGML, FIL_ADJPOS,
   FIL_WTOPOS
};


char        idxpath[_MAX_PATH];

int         nPOSx;

BOOL        bSLen = FALSE;
BOOL        bPOS = FALSE;
BOOL        bText = FALSE;

BOOL bNoBib=TRUE;

int         iq = 0;

FILE       *fSysLog;  
#if !defined(_WIN32)
FILE *fAdmLog;
#endif

char        calluser[MAX_USER];
char        callloc[30] = "local";

BOOL        bInt;

BOOL bSkipTests=TRUE;
/*
 * LOGOUT
 * LOG user password
 * SAVE name
 * DIR
 * DIR q
 * REMOVE q
 * PURGE n
 * SOLVE name query
 * SUBCORPUS name query
 * WORDLIST name pattern
 * SQTABLE name name data
 * FILTER name filter
 * SORTFILTER name n filter
 * SORT name n
 * GET name n,m
 * CUT
 */


#define NCOMMAND (sizeof(ctable)/sizeof(struct cmd))

xairo_uchar        psent[MAX_SOL];

void        GetCString(char *);
BOOL        ParseString(xairo_uchar *, struct pcmd *, struct cmd[]);
void        ExecString(struct pcmd *);
void        SetString(char *);
BOOL        LookUpCmd(char *, int *, struct cmd[]);
BOOL        GetArg(int, xairo_uchar **, union uarg *);
BOOL        GetBTS(xairo_uchar **, char *,int);
BOOL        GetBTSW(xairo_uchar **, xairo_uchar *,int);
BOOL        GetRest(xairo_uchar **, char *,int);
BOOL        GetRestW(xairo_uchar **, xairo_uchar *,int);
BOOL        GetTable(xairo_uchar **, int *);
BOOL        GetRange(xairo_uchar **, struct ran *);
void        parg(int, union uarg *);
void        Log(char *, char *);
void        Logout(void);
BOOL        NewQName(char *);
char       *GetQPath(char *);
void        FreeQPath(char *);
xairo_uchar       *GetRef(sara_ulong, sara_long, WORDNO, int,int, int *, int *, int*,
                   int scope, char *,char*);
xairo_uchar       *GetSol(char *, int, int *, int *, sara_long *, int, char *,char*);
void        BoundElement(char *, WORDNO, sara_ulong, sara_long *,
                         sara_long *, int *, sara_long *, sara_long,
                         sara_long);
void        BoundElement2(char *, WORDNO, sara_ulong, sara_long *,
                          sara_long *, int *, sara_long *, sara_long,
                          sara_long);
void        AddFilter(char *, struct pcmd *);
void        ClearFilters(char *);
/* void        ApplyFilter(struct pcmd *, char *, int *, int *); */
void        ApplyFilter(enum filtype, char *, int *, int *);
void        StripAllSGML(char *s, int *, int *);
void        DelSGML(char *, int *, int *);
int         IsSGML(char *);
void        StripAllPOS(char *s, int *, int *);
int         DelPOS(char *, int *, int *);
int         IsPOS(char *, int *);
BOOL        strtest(char *, char *);
void        NormSpace(char *, int *, int *);
void        DelNL(char *, int *);
void        ConsolSpaces(char *, int *);
//void        LeftL(char *, int, int *, int *);
void        AddChar(char *, int);
void        ClearChars(void);
void        TranslateChars(char *, int *, int *);
int         IsChar(char *, int *);
void        SQTable(char *,char *, int, int *);
void        Thin(char *, char *, int, int, int, int *, int *);
void        ThinInit(char *, char *, int, int *);
void        ThinRand(char *, char *, int, int *);
void        OnePerText(char *, char *, int *);
void        NormCR0(char *, int *, int *);
BOOL        Empty(char *);
sara_long   FindHeader(FILE *);
BOOL        Ignore(char *);
BOOL        AllBlank(char *);
void        ETag(FILE *, sara_long *);
void        Trim(xairo_uchar *);
BOOL Startup(char*,BOOL,const char*,const char*);
#if defined(SERVER)
int         readrec(FILE *, char *, char *, char *);
BOOL        readst(FILE *, char *);
#endif /* ! SERVER */

BOOL        GetLims(FILE *, sara_long, sara_long, const xairo_uchar *,
                    sara_long *, sara_long *);
void        GetDictEntry(FILE *, sara_long, xairo_uchar *, int, sara_long *, sara_long*);
BOOL NewSubCorpus(const char*,const char*);
BOOL SetSubCorpus(const char*);
BOOL SetSubCorpusPart(const char*,int);
BOOL AddSubCorpus(const char*,int, int*);
BOOL SetSubCorpusNames(const char*, char*);
void SCEnumNames(const char*,char*);
BOOL SCSave(char*);
char* SCList(int);
char* SCEnum(const char*,int,int);
int SCCount();
void SCReadAll();
BOOL SCRegistered();
BOOL AKExists();
int FTable(int,int,int,const xairo_uchar*);
void FEntry(int,xairo_uchar*,int*,int*);
void FFree();
int QuickFrequency(const xairo_uchar*,int);
BOOL QuickFrequencyG0(const xairo_uchar*);
int QuickForms(const xairo_uchar*);
int DicFrequency(const xairo_uchar*);
void GetFTable(char*);
void CTab(int,xairo_uchar**,int*, double*);
int MakeCTab(char*,int,int,int,xairo_uchar*);
int MakeCTabX(char*,char*,int,xairo_uchar*);
const xairo_uchar* GetContName(int);
BOOL ACScore(char*,int,int, int,xairo_uchar*,long*,double*);
void CTabOptions(int,int,char*,int);
BOOL Form(xairo_uchar*,int,xairo_uchar*,int*);
BOOL SlowForm(xairo_uchar*,int,xairo_uchar*,int*);
BOOL SubForm(xairo_uchar*,int,int,xairo_uchar*);
BOOL SlowSubForm(xairo_uchar*,int,int,xairo_uchar*);
char* LemmInfo();
long SCCountWords(char*,int);
BOOL NewStartup(const char* f,BOOL);
int getFirstFtab();
BOOL getNextFtab(int* i,xairo_uchar* s,int* f,int* ff);
int CountDocuments();
int LocToDoc(sara_long,sara_ulong);
BOOL decode(int code,const char* input,int k,xairo_uchar* output,int* pos,int*);

int         dcmp(const xairo_uchar*, const xairo_uchar *);

int         OpenQPath(char *);
int         InUse(char *);
BOOL        Exists(char *);
void        Dummy(char *name);
void        GetCName(char *, char *);
void        WToPos(char *, int *, int *);
int         IsSpec(char *, char **);
void        TranslateSpec(int, char **);
BOOL        wtag(char *);
BOOL        DTDelt(int, char *);
void CleanUp();
BOOL GetAttrib(sara_ulong,sara_ulong,char*,char*,char*);
void CTabThin(char*,char*,int,int,int,xairo_uchar*,int*,int*,BOOL,char*);
void CTabFree();
int LookupAtt(char*,char*);
void AttDict(int,char*,int*);
const char* GetTextPath(int i);
void syslogw(int level, const xairo_uchar *format,...);

#if defined(_WIN32)
int         getpid(void);
#else
void        ALogEvent(char *,...);
#endif /* __unix__ */

BOOL        GetBibHeader(char *, int *, int *);
BOOL        GetBibItem(char *, int, char *);
int GetDocBib(int,xairo_uchar*,int);
void DocToLoc(int k,sara_long* pos,sara_ulong* code);
int Outer(int);
extern int GetPosA(xairo_uchar*,int,xairo_uchar**,int);
extern const char* xml_err;

BOOL(*interrupt_function) (void);

void        LoadChars(char *);

#if !defined(_WIN32)
void        detach(void);
#endif /* __unix__ */

int         nLog;
int         ntry;

#define MAXTRY 3

BOOL        bServer = FALSE;
int         max_sent = MAXSENTENCE;
sara_long   dbase, abase;

FILE       *fPack;
FILE       *fBib;
FILE       *fBibIndex;
FILE       *fDownload;

BOOL        bDTD;
int         nDTD;

struct pihead idxhead;
char      param_file[_MAX_PATH];           
BOOL bInit;
#define DEFPARAM "corpus.prm"


#if !defined(_WIN32)

void      Test();
void      StartServer();
void      XServer(xairo_uchar *, xairo_uchar **);
void      Session(int);
void      sig_alarm(int);
void      sig_urgent(int);


#if 0
main(int argc, char **argv)
{
   char          *param = PARAM;
   int            opt;
   int            errflag = 0;
   int            debug = 0;
   extern int     optind;
   extern char   *optarg;
   char          *progname;
   
   optarg = NULL;

   if (progname = strrchr(*argv, '/'))
      progname++;
   else
      progname = *argv;
   
   while (!errflag && (opt = getopt(argc, argv, "dnp:")) != -1)
      switch (opt) {
       case 'd':  debug = 1;        break;
       case 'p':  param = optarg;   break;
		case 'n': bSkipTests=TRUE; break;
       default:   errflag++;        break;
      }
   
   if (errflag) {
      fprintf(stderr, "Usage: %s [-d] [-n] [-p config_file]\n", progname);
      exit(1);
   }
   
   if (!Startup(param)) {
   	printf("There were errors detected in startup tests. See testlog.txt.\n");
   	exit(1);
   }	
#if defined(LOG_MAIL)
   openlog(progname, LOG_PID|LOG_NDELAY, LOG_FACILITY);
#else
   openlog(progname, LOG_PID,LOG_FACILITY);
#endif
   
   if (debug)
      Test();
   else 
      StartServer();
}
#endif

#if 0
void Test()
{
   char      s[1000];
   char      ss[1000];
   char      ap[_MAX_PATH];
   int       n;

   printf("Type end to leave test mode\n");
   syslog(LOG_NOTICE, "Started server in test mode");

   sprintf(ap, "%sadmlog", accpath);
   fAdmLog = fopen(ap, "a");
   ALogEvent("Started server in test mode");


   n = AllocBuffers(nbuff);

   syslog(LOG_INFO, "Allocated %d+%d buffers", n - nbuff, nbuff);

   while (1) {
      printf("> ");
      scanf("%[^\n]", s);
      if (!strcmp(s, "end"))
         break;

      getchar();
      XServer(s, ss);
      printf("%s\n", ss);
   }
}
#endif

struct sockaddr_in   passive,
                     inaddr;


void tryname(struct in_addr, char *);

int child;

void backup(char *path, int i)
{
   char      p1[_MAX_PATH],
             p2[_MAX_PATH];

   if (i == 0)
      strcpy(p1, path);
   else
      sprintf(p1, "%s.%d", path, i);
   sprintf(p2, "%s.%d", path, i + 1);
   remove(p2);
   rename(p1, p2);
}

void catch_child(int sig)
{
   int   status;
   pid_t pid;
   
   if (sig == SIGCHLD) {
      while (   (pid = waitpid(-1, &status, WNOHANG)) == -1
             && (errno == EINTR))
         ;
      if (WIFSIGNALED(status))
         syslog(LOG_DEBUG, "Child pid=%d was killed with signal %d",
                pid, WTERMSIG(status));
      else if (WIFEXITED(status))
         syslog(LOG_DEBUG, "Child pid=%d exited with status %d",
                pid, WEXITSTATUS(status));
   }
}

void StartServer()
{
   int               s;
   int               m;
   socklen_t               inal;
   char              us[50];
   char              us1[50];
   int               one = 1;
   char              ap[_MAX_PATH];
   int               i;
   struct servent   *sp;
   uid_t             uid;
   struct passwd    *pw;
   FILE             *fpPID;
   struct sigaction  saChild;

   saChild.sa_handler = catch_child;
   sigemptyset(&saChild.sa_mask);
   saChild.sa_flags = SA_NOCLDSTOP;

   sigaction(SIGCHLD, &saChild, 0);
                             
   GetPIndex();
   GetDict();

   sprintf(ap, "%ssyslog", accpath);
   for (i = nback - 1; i >= 0; i--)
      backup(ap, i);
   sprintf(ap, "%sadmlog", accpath);
   fAdmLog = fopen(ap, "a");

   uid = getuid();
   pw = getpwuid(uid);

   syslog(LOG_NOTICE, "Started server (%s)", pw->pw_name);
   fprintf(stdout, "Started server\n");
   fflush(stdout);

   detach();
   sprintf(ap, "%ssarad.pid", accpath);
   if ((fpPID = fopen(ap, "w")) == NULL) {
      syslog(LOG_WARNING, "Couldn't create ap", ap);
   } else {
      fprintf(fpPID, "%d\n", getpid());
      fclose(fpPID);
   }

   s = socket(AF_INET, SOCK_STREAM, 0);

   if (s < 0) {
      printf("Error opening socket\n");
      exit(1);
   }

   passive.sin_family = AF_INET;
   passive.sin_addr.s_addr = INADDR_ANY;
   passive.sin_port = htons(port);
/*
   passive.sin_port=sp->s_port;
*/

   if (bind(s, (struct sockaddr *)&passive,
            sizeof(struct sockaddr_in)) < 0) {
      fprintf(stderr, "Error binding socket [%d]", errno);
      perror("bind()");
      syslog(LOG_ERR, "Error binding socket [%d]", errno);
      exit(1);
   }

   setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *)&one, sizeof(one));
   listen(s, 1);
   while (1) {
      inal = sizeof(inaddr);

      /*
       * Server used to fall over if accept() failed
       *
       * Since it now catches children, we just log errors
       * and keep going (ignoring EINTR errors)
       */
      while ((m = accept(s, (struct sockaddr *)&inaddr, &inal)) == -1) {
         if (errno != EINTR)  {
            syslog(LOG_ERR, "Error accepting connection [%d]", errno);
         }
      }

      tryname(inaddr.sin_addr, callloc);

      syslog(LOG_INFO, "Connect from %s", callloc);
      ALogEvent("Connect from %s", callloc);

      child = fork();

      if (child == -1) {
         syslog(LOG_ERR, "Cannot fork [%d]", errno);
         close(m);
         continue;
      }
      if (child) {
         syslog(LOG_DEBUG, "Forked process %i", child);
         close(m);
      } else {
         close(s);
         Session(m);

         if (fSysLog != NULL)
            fflush(fSysLog);
         fflush(fAdmLog);

         exit(0);
      }
   }
}


void tryname(struct in_addr a, char *m)
{
   struct hostent *h;

   h = gethostbyaddr((char *) &a, sizeof(a), AF_INET);

   if (h)
      strcpy(m, h->h_name);
   else
      strcpy(m, inet_ntoa(a));
}

char      bigbuff[6000];
char      bb2[6000];

#define TIMEOUT 600

int       urgsock;

void Timer(int t)
{
#if defined(DONT_CARE_ABOUT_POSIX_COMPLIANCE)
   struct itimerval tv;

   tv.it_value.tv_sec = timeout;
   tv.it_value.tv_usec = 0;
   tv.it_interval.tv_sec = 0;
   tv.it_interval.tv_usec = 0;
   setitimer(ITIMER_REAL, &tv, NULL);
#else /* ! DONT_CARE_ABOUT_POSIX_COMPLIANCE */
   alarm(t);
#endif /* DONT_CARE_ABOUT_POSIX_COMPLIANCE */
}


void Session(int s)
{
   int       k,
             n;

   struct sigaction sa;

   nLog = 0;


   sa.sa_handler = sig_alarm;
   sigemptyset(&sa.sa_mask);
   sa.sa_flags = 0;
   sigaction(SIGALRM, &sa, NULL);

   sa.sa_handler = sig_urgent;
   sigemptyset(&sa.sa_mask);
   sa.sa_flags = 0;
   sigaction(SIGURG, &sa, NULL);

   urgsock = s;
   fcntl(s, F_SETOWN, getpid());
   n = AllocBuffers(nbuff);

   while (1) {
   		int bp=0;
      if (timeout)
         Timer(timeout);
      		while (1) {
      		k = read(s, bigbuff+bp, sizeof(bigbuff)-bp);
      		bp+=k;
      if (k == 0) {
         Logout();
         ALogEvent("Client closed socket");
         syslog(LOG_NOTICE, "Client closed socket");
         exit(1);
      }
      		if (bigbuff[bp-1]==0) break;          
      		}
      		
/*
 * Now we should clear the timer and prepare for interrupts
 */

      if (timeout)
         Timer(0);

//      XServer(bigbuff, bb2);


      if (nLog == -1)
         break;

      write(s, bb2, strlen(bb2) + 1);
   }

   close(s);
}


void sig_alarm(int sig)
{
   syslog(LOG_WARNING, "Timed out");
   ALogEvent("Timed out");
   Logout();
   exit(1);
}

void sig_urgent(int sig)
{
   char      buffer[100];

   recv(urgsock, buffer, sizeof(buffer), 0);
   syslog(LOG_NOTICE, "Client sent interrupt");
   bInt = TRUE;
}

#else /* ! __unix__ */


# if !defined(WIN32)
int main()
{
   _fullpath(param_file, DEFPARAM, sizeof(param_file));
   return 1;
}

# else /* ! WIN32 */
#  define DLL_PROCESS_ATTACH 1
BOOL __stdcall DllMain(void *hInst,
		                 long unsigned ul_reason_for_call,
		                 void *lpReserved)
{
   switch (ul_reason_for_call) {
    case DLL_PROCESS_ATTACH:
//      _fullpath(param_file, DEFPARAM, sizeof(param_file));
      break;

    default:
      break;
   }
   return TRUE;
}

# endif /* WIN32 */


#endif /* __unix__ */

# if defined(_WIN32)
__declspec(dllexport) int InitServer(const char* p,BOOL(*f) (void),BOOL bLog,const char* temp,const char* bl)
# else /* ! WIN32 */
int InitServer(const char* p,BOOL(* f) (void),BOOL bLog,const char* temp,const char* bl)
# endif /* WIN32 */
{
   strcpy(param_file,p);

   interrupt_function = f;
	if (!Startup(param_file,bLog,temp,bl)) return 0;
   nLog = 1;
   return 1;
}

int nprmerrs;

FILE* fBoot;

BOOL Startup(char* file,BOOL bLog,const char* temp,const char* bl)
{                   
	char dpath[_MAX_PATH]; 
//	char* p=strrchr(file,'\\'); 
	int bt;
	nprmerrs=0;;
	fBoot=fopen(bl,"wt");
   _getcwd(dpath,_MAX_PATH);
   strcpy(tmppath,temp);
	bt=_chdir(tmppath);
	_chdir(dpath);
	if (bt) {
		bootlog("The temporary directory does not exist\n");
		fclose(fBoot);
		return FALSE;
	}
	char fp[_MAX_PATH];
	_fullpath(fp,file,_MAX_PATH);
	if (!NewStartup(fp,bLog)) {
		fclose(fBoot);
		return FALSE;
	}
   strcpy(dpath,etcpath);
   strcat(dpath,corpname);
   strcat(dpath,".dsc");
//   if (!ReadDescFile(dpath)) {
//	   fclose(fBoot);
//	   return FALSE;
  // }
	syslog(LOG_NOTICE,"Read .DSC file (why???)");
 

   //f (!bSkipTests && !Tests()) return 0;
   if (!GetAccels()) {
	   bootlog("No accelerator file found\n");
	fclose(fBoot);
	   return FALSE;
   }

		fclose(fBoot);
   syslog(LOG_NOTICE,"Initialised accelerators");
   GetDict();                                                      
   syslog(LOG_NOTICE,"Opened dictionary");
   SCReadAll();
   syslog(LOG_NOTICE,"Read subcorpus register");
//   initialise_lexrules(NULL);
   return TRUE;
}

#if defined(SERVER)

xairo_uchar      ans[MAX_SOL + 10];

struct pcmd pc;


void

# if defined(_WIN32)
#  if defined(WIN32)
__declspec(dllexport) 
#  else /* ! WIN32 */
__far __pascal __export
#  endif /* WIN32 */
# endif /* __unix__ */


XServer(xairo_uchar *c, xairo_uchar **cc)
{

# if defined(_WIN32)
   _fmode = _O_BINARY;
# endif /* __unix__ */
   syslogw(LOG_DEBUG, L("Client sent string %s"), c);
   if (ParseString(c, &pc, ctable))
		ExecString(&pc);
   *cc=ans;   
   syslogw(LOG_DEBUG, L("Sent reply %.500s"), ans);
   return;
}

#else /* ! SERVER */

void Server(void)
{
   char      s[MAX_COMMAND];
   struct pcmd pc;

   ntry = nLog = 0;

   do {
      GetCString(s);
      if (!ParseString(s, &pc, ctable))
         continue;
      ExecString(&pc);
   }

   while (nLog != -1);
}


void GetCString(char *c)
{
   printf("server> ");
   scanf("%[^\n]", c);
   while (getchar() != 10)
      ;
   return;
}

#endif /* SERVER */

#if !defined(_WIN32)
BOOL TestInterrupt(void)
{
   return bInt;
}

#else /* ! __unix__ */

BOOL TestInterrupt(void)
{
   bInt = bInt || (*interrupt_function) ();
   return bInt;
}

#endif /* __unix__ */

BOOL ParseString(xairo_uchar *s, struct pcmd *pc, struct cmd table[])
{
   char      cd0[64];
   char     *cd = cd0;
   int       i;
   xairo_uchar      c1;

   while (*s == 32)
      s++;

   while (*s != 0 && *s != 32) {
      c1 = *(s++);
      *(cd++) = toupper((char) c1);
   }

   *cd = 0;

   if (!LookUpCmd(cd0, &(pc->i), table)) {
      SetString("NO COMMAND");
      return FALSE;
   }

   for (i = 0; i < MAX_OPS; i++)
	   if (!GetArg(table[pc->i].op[i], &s, &(pc->arg[i]))) {
		   SetString("NO OVERFLOW");
		   return FALSE;
	   }


   return TRUE;
}


#if defined(SERVER)

void SetString(char *s)
{
   mbstowcs(ans, s,sizeof ans);
   return;
}

void SetStringW(xairo_uchar *s)
{
   wcscpy(ans, s);
   return;
}

#else /* ! SERVER */

void SetString(char *s)
{
   printf("%s\n", s);
   return;
}

#endif /* SERVER */


BOOL LookUpCmd(char *cd, int *i, struct cmd table[])
{
   int       j;
	
   for (j = 0; j < NCOMMAND; j++)
      if (!strcmp(cd, table[j].name))
	 break;

   *i = j;

   return j < NCOMMAND;
}


BOOL GetArg(int op, xairo_uchar **s, union uarg * arg)
{
   char      t[100];
   switch (op) {
	   case OP_NULL:
		   return TRUE;
    case OP_USER:
      return GetBTS(s, arg->user,MAX_USER);
 
    case OP_PW:
      return GetBTS(s, arg->pw,MAX_PW);

    case OP_NAME:
      return GetBTS(s, arg->name,MAX_NAME);
 
    case OP_NAMEW:
      return GetBTSW(s, arg->namew,MAX_NAME);
 
    case OP_NUM:
      if (!GetBTS(s, t,100)) return FALSE;
      arg->num = atoi(t);
      return TRUE;

    case OP_LNUM:
      if (!GetBTS(s, t, 100)) return FALSE;
      arg->lnum = atol(t);
      return TRUE;

    case OP_CQL:
      return GetRest(s, arg->cql, MAX_CQL);

    case OP_CQLW:
      return GetRestW(s, arg->cqlw, MAX_CQL);

    case OP_REGEXP:
      return GetBTS(s, arg->regexp,MAX_REGEXP);
  
    case OP_TABLE:
      return GetTable(s, &(arg->num));
 
    case OP_FILTER:
      return GetRest(s, arg->filter,MAX_FILTER);
 
    case OP_RANGE:
      return GetRange(s, &(arg->range));
  
    default:
      return FALSE;
   }
}


BOOL GetBTS(xairo_uchar **s, char *c,int max)
{
	char* c0=c;
   while (isspace(**s))
      (*s)++;

   if (**s == '\"') {           /* " for emacs */
      (*s)++;
      while (**s != 0 && **s != '\"') /* " for emacs */
	  {
		  *(c++) = (char) *((*s)++);
		  if (c-c0 >= max) return FALSE;

	  }

      *c = 0;

      if (**s != 0)
         (*s)++;

      return TRUE;
   }

	  while (**s != 0 && !isspace(**s)) {
      *(c++) = (char) *((*s)++);
	  if (c-c0>=max) return FALSE;
	  }

   *c = 0;

   return TRUE;
}

BOOL GetBTSW(xairo_uchar **s, xairo_uchar *c,int max)
{
	xairo_uchar* c0=c;
   while (isspace(**s))
      (*s)++;

   if (**s == '\"') {           /* " for emacs */
      (*s)++;
	  while (**s != 0 && **s != '\"') /* " for emacs */ {
         *(c++) = (char) *((*s)++);
		if (c-c0 >= max) return FALSE;	
	  }

      *c = 0;

      if (**s != 0)
         (*s)++;

      return TRUE;
   }

	  while (**s != 0 && !isspace(**s)) {
      *(c++) = (char) *((*s)++);
	  if (c-c0>=max) return FALSE;
	  }

   *c = 0;

   return TRUE;
}


BOOL GetRest(xairo_uchar **s, char *c,int max)
{
   while (**s > 0 && isspace(**s))
      (*s)++;

   while (**s != 0) {
      *(c++) = (char) *((*s)++);
      max--;
      if (max <=0) return FALSE;
   }   

   *c = 0;

   return TRUE;
}

BOOL GetRestW(xairo_uchar **s, xairo_uchar *c,int max)
{
   while (**s > 0 && (**s==' '))
      (*s)++;

   while (**s != 0) {
      *(c++) = *((*s)++);
      max--;
      if (max <=0) return FALSE;
   }   

   *c = 0;

   return TRUE;
}


BOOL GetRange(xairo_uchar **s, struct ran *r) // unchecked
{
   char      c0[10];
   char     *c = c0;

   while (isspace(**s))
      (*s)++;

   while (**s != 0 && !isspace(**s) && **s != ';')
      *(c++) = (char) *((*s)++);

   r->min = atoi(c);
   r->size = 1;

   if (**s != ';')
      return TRUE;

   c = c0;

   while (**s != 0 && !isspace(**s))
      *(c++) = (char) *((*s)++);

   r->size = atoi(c);

   return TRUE;
}

#define MAXSELTAB 1000

static int       seltab[MAXSELTAB];

BOOL GetTable(xairo_uchar **s, int *i0) // unchecked
{
   int       i = 0;
   int       j = 0;
   BOOL      bInt = FALSE;

   while (**s != 0) {
      if (**s == ' ' && bInt && i<MAXSELTAB) {
         seltab[i++] = j;
         bInt = FALSE;

         j = 0;

      } else if (isdigit(**s)) {
         j = 10 * j + (**s - '0');
         bInt = TRUE;
      }
      (*s)++;
   }

   if (bInt && i<MAXSELTAB)
      seltab[i++] = j;

   *i0 = i;

   return TRUE;
}

struct pcmd pf;


void ExecString(struct pcmd *pc)
{
   int         i;
//   int         n;
   sara_long   ln;
	BOOL bXML=FALSE;
    char xmlerr[200]; 

#if !defined(SERVER)
   printf("%i\n", pc->i);
#endif /* SERVER */

   for (i = 0; i < MAX_OPS; i++)
      parg(ctable[pc->i].op[i], &(pc->arg[i]));

   if (nLog != 1 && pc->i > 1 && pc->i != 32) {
      SetString("NO LOGIN");
      return;
   }

   switch (pc->i) {
    case 0:				 /* LOGOUT */
      Logout();
      SetString("");        
      CleanUp();

      break;

    case 1:				 /* LOG */
      Log(pc->arg[0].user, pc->arg[1].pw);
      break;

    case 2:				 /* SAVE */
		assert(FALSE);
    case 3:				 /* DIR */
		assert(FALSE);
   
    case 4:				 /* REMOVE */
		break;
    case 5:				 
		assert(FALSE);

    case 6:				 /* SOLVE */
solve:
		if (pc->arg[1].cqlw[0]==0 || (bXML && wcscmp(pc->arg[1].cqlw,L("<cql><this/></cql>"))==0)) {
			char s[40];
			ln=Solve0(pc->arg[0].name,&i);
			sprintf(s, "OK %ld %d", ln, i);
			SetString(s);
			break;
		}

      if (!NewQName(pc->arg[0].name)) {
         SetString("NO FILES");
         break;
      }
		ln = SolveX(pc->arg[0].name, pc->arg[1].cqlw, &i,xmlerr);

      if (ln <= 0)
         remove(GetQTPath(pc->arg[0].name));
      else {
         remove(GetQPath(pc->arg[0].name));
         rename(GetQTPath(pc->arg[0].name), GetQPath(pc->arg[0].name));
      }

      if (ln == -4)
         SetString("NO STREAMS");
      else if (ln == -3)
         SetString("NO ABORT");

      else if (ln == -2)
         SetString("NO SPACE");

      else if (ln == -1)
	  {
		  char s[2000];
		  strcpy(s,"NO SYNTAX ");
		  strcat(s,xmlerr);
		  SetString(s);
	  }

      else if (ln == 0)
         SetString("NO 0");

      else {
         char      s[40];

         sprintf(s, "OK %ld %d", ln, i);
         SetString(s);
      }

      break;
      
    case 7:				 /* SUBCORPUS */
      {
         char      s[200]; double m, sd; long nw, nu; int k;

//        if (strcmp(pc->arg[0].name, corpname)) {
//			 SetString("NO");
//            break;
//		 }
			m=0.0;
			sd=0.0; nw=nu=0;
			nw=dicthead.ndix;
			nu=dicthead.nu;

			k=CountDocuments();
         sprintf(s, "OK %d %f %f %ld %ld", k , m, sd, nw, nu);
         SetString(s);
         break;
      }

   case 8:				 /* WORDLIST */
		assert(FALSE);
   case 9:				 /* SQTABLE */
      {
         int       i = pc->arg[2].num;
         char      s[10];
         int       j;

         if (!NewQName(pc->arg[1].name)) {
            SetString("NO FILES");
            break;
         }
         SQTable(pc->arg[0].name,pc->arg[1].name, i, &j);

         if (i > 0)
            sprintf(s, "OK %d %d", i, j);
         else
            strcpy(s, "NO 0");

         SetString(s);
         break;
      }

    case 10:				 /* FILTER */
		assert(FALSE);

    case 11:				 /* SORTFILTER */
		assert(FALSE);
      
    case 12:				 /* SORT */
		assert(FALSE);

    case 13:				 /* GET */
      {
         xairo_uchar       *s;
         xairo_uchar        s1[MAX_SOL];
         int         pos;
	      int         len;
         sara_long   code;

         char         sno[20];
		 char loc[20];
         char        id[_MAX_PATH];

         bPOS = bSLen = bText = FALSE;
		 strcpy(sno,"?");
         s = GetSol(pc->arg[0].name, pc->arg[1].num, &pos, &len,
                    &code, pc->arg[2].num, sno,loc);
         
         if (code==-1) {
         	SetString("NO SOL");
         	break;
         }	

         if (s == NULL) {
         	swprintf(s1,L("OK -1 -1 0 0 HDR Text not available"),id);
            SetStringW(s1);
            break;
         }
//         Reformat(pc->arg[0].name, s, &pos, &len);
		 {
			 char path[_MAX_PATH];
			 GetDocPath(code,id,path);
		 }
         if (len > (int) wcslen(s) - pos && (int) wcslen(s) - pos >= 0)
            len = wcslen(s) - pos;

         swprintf(s1, L("OK %S %S %d %d %S %s"), id, sno, pos, len,
			 "HDR", s);
         SetStringW(s1);
         break;
      }

    case 14:				 /* CUT */
		assert(FALSE);
    case 15:				 /* CHAR */
		assert(FALSE);

    case 16:				 /* GETPOS */
      {
		xairo_uchar s[1000];   
		  xairo_uchar*       sig[200];
		 int n=GetPosA(pc->arg[0].namew,pc->arg[1].num,sig,200);
		int i;
         swprintf(s, L("OK %d"), n);

         for (i = 0; i < n; i++) {
            wcscat(s, L(" "));
            wcscat(s, sig[i]); 
         }
         SetStringW(s);
         break;
      }

    case 17:				 /* GETDTD */
      break;

    case 18:				 /* MAXLENGTH */
      {
         char      s[20];
         int       n = pc->arg[0].num;

         if (n > MAXSENTENCE)
            n = MAXSENTENCE;

         max_sent = n;
         sprintf(s, "OK %d", n);
         SetString(s);
         break;
      }

   case 19:				 /* GETSC */
      {
         char      s[200],
                   s1[200];
         char      path[_MAX_PATH] = "";
//         FILE* f; 
		 BOOL b=TRUE;
         
//         if (strcmp(pc->arg[0].name, corpname))
//            {
//            SetString("NO");
//            break;
//         }   

         GetDocName(pc->arg[0].num, s);
//         f=fopen(path,"r");
//         b=(f!=NULL);
//         if (b) fclose(f);
         sprintf(s1, "OK %s %d", s,b);
         SetString(s1);
         break;
      }

   case 20:				 /* GETHEAD */
		assert(FALSE);
   case 21:				 /* LOOKUP */
      {
         sara_long   x1; 
         sara_long   x2;
         char        s[40]; int k;
		 xairo_uchar w[1000];
		 wcscpy(w,pc->arg[0].cqlw);
		 MkLC(w);
		 MkNorm(w);
         if (GetLims(fDict, 0, dicthead.ndix - 1,w, &x1, &x2)) {
            dbase = x1;
			if (!SCRegistered()) {
				int i; xairo_uchar s1[80]; int freq, forms; k=0; 
				for (i=0;i<x2-x1+1;i++) {
					GetDictEntry(fDict,i+dbase,s1,80,&freq,&forms);
					if (QuickFrequencyG0(s1)) k++;
				}
			} else k=x2-x1+1;

            sprintf(s, "OK %d", k);
         } else
            strcpy(s, "NO 0");

         SetString(s);
         break;
      }

   case 22:				 /* DMATCH */
      {
         xairo_uchar        s1[80];
         xairo_uchar        s2[80];
         xairo_uchar        s[200];    /* used to be 100, but 80 + 80 > 100 */
         sara_long   freq, forms;
         int         i = 0;
         int         j = 0;
         int         n = pc->arg[0].num;
		 int db;
         sara_long   k;
		 s1[0]=s2[0]=0;
		 db=dbase;

//xx:
         k = n + db;

         GetDictEntry(fDict, k, s1, sizeof(s1)/sizeof(xairo_uchar), &freq, &forms);
         wcscpy(s2, s1);
//         TranslateChars(s2, &i, &j); 
		 if (s2[0]==0) wcscat(s2, L(" "));
		 if (!SCRegistered()||AKExists()) {
			 freq=QuickFrequency(s1,0);
//			 if (freq==0) {db++; goto xx;}
			 forms=QuickForms(s1);
		 }

         swprintf(s, L("OK %li %li %s"), freq, forms, s1);

         SetStringW(s);
         break;
      }

   case 23:				 /* QNAME */
      {
         char      s[40];
         char      n[20];

         do
            sprintf(n, "hq%06d", iq++);
         while (Exists(n));

         Dummy(n);
         sprintf(s, "OK %s", n);
         SetString(s);
         break;
      }

   case 24:				 /* OPEN */
      {
         int       k = OpenQPath(pc->arg[0].name);
         char      s[40];

         if (k > 0)
            sprintf(s, "OK %d", k);
         else
            strcpy(s, "NO NOFILE");

         SetString(s);
         break;
      }

   case 25:				 /* CSCORE */
      {
         char        s[40]; int t;
         sara_long   l = colloc_score(pc->arg[0].name, pc->arg[2].cqlw,
                                      pc->arg[1].num,pc->arg[1].num,&t);

         if (l < 0)
            strcpy(s, "NO SYNTAX");
         else
            sprintf(s, "OK %ld %d", l, t);

         SetString(s);
         break;
      }

   case 26:				 /* LOC */
      {
         char              s[40];
         FILE             *f;
         struct locent1    l;
         char             *name;

         name = GetQPath(pc->arg[0].name);
         if ((f = fopen(name, "r")) == NULL) {
            syslog(LOG_ERR, "Can't open %s\n", name);
            strcpy(s, "NO FILE");
            break;
         } else {
            fseek(f, (pc->arg[1].num * sizeof(struct locent1)),
                  SEEK_SET);
            fread(&l, sizeof(l), 1, f);
            fclose(f);

            sprintf(s, "OK %ld %ld %d", l.lcode, l.pos, l.lw);
         }
         SetString(s);
         break;
      }

   case 27:				 /* GETHEAD2 */
	   assert(FALSE);
    case 28:				 /* GETCHEAD */
		assert(FALSE);
      
    case 29:				 /* GET1SOL */
      {
         char      text[100];
         xairo_uchar      element[100];
         xairo_uchar      att[100];
         xairo_uchar      s[1000];
		 xairo_uchar pfx[100];

         wcscpy(s, L("OK "));
         strcpy(text, pc->arg[0].name);
         wcscpy(element, pc->arg[1].namew);
         wcscpy(att, pc->arg[2].namew);
		 wcscpy(pfx,pc->arg[3].namew);

         if (Get1Sol(FindDoc(text), element, att, pfx, s + 3,990)) 
         SetStringW(s);
		 else SetString("NO");

         break;
      }
      
    case 30:				 /* MOTD */
		assert(FALSE);
      
    case 31:				 /* PWD */
      {      
#if !defined(_WIN32)      
         if (SetPassword(calluser, pc->arg[0].pw, pc->arg[1].pw))
            SetString("OK");
         else
            SetString("NO");
         break;    
#endif         
      }
      
    case 32:				 /* INFO */
      /*
       * Format is as follows:
       * 
       * Client sends
       *    INFO cs
       * where
       *    cs is the character set ID
       * 
       * Server sends
       *    OK t hdv sv cv nm k
       * where
       *    t is the timeout
       *    hdv is the header version
       *    sv is the server version
       *    cv is the lowest acceptable client version 
       *    nm is the corpus name
	   *    k is the number of registered subcorpora (excluding all)
       *
       */
      {
         char      s[300];
         char      t[300];

         strcpy(t, pc->arg[0].name);
//         LoadChars(t);
         sprintf(s, "OK %d %d %d 931 %d %s", timeout, hdv, VER,SCCount()-1,corpname);

         SetString(s);
         break;
      }
      
    case 33:				 /* RLOOKUP */
      {
         int       n;
         char      s[30];

         n = WoL1(pc->arg[0].cqlw);
         if (n == -2){
			 sprintf(s,"NO TOOMANY %d",MAXRE);
			 SetString(s);
		 }
         else if (n == 0)
            SetString("NO 0");
         else {
            sprintf(s, "OK %i", n);
            SetString(s);
         }

         break;
      }
      
    case 34:				 /* RGET */
      {
         int         n;
         xairo_uchar        s[80];
         sara_long   freq = 0;
		 sara_long forms=0;
         xairo_uchar        wd[80];
         xairo_uchar        wd1[80];
         int         i = 0;
         int         j = 0;
		 wd[0]=wd1[0]=0;

         n = pc->arg[0].num;

         WoL2(n, wd, sizeof(wd)/sizeof (xairo_uchar), &freq, &forms);
		    	syslog(LOG_DEBUG,"In GetDictEntry w=%S",wd);

				if (!SCRegistered()) {
					freq=QuickFrequency(wd,0);
					forms=QuickForms(wd);
				}

         wcscpy(wd1, wd);
//         TranslateChars(wd1, &i, &j); 
		 if (wd1[0]==0) wcscpy(wd1,L(" "));

         swprintf(s, L("OK %li %li %s"), freq, forms, wd );
         SetStringW(s);
         break;
      }

    case 35:				 /* BIB */
      {
         xairo_uchar      s[12010];
		s[0]='O'; s[1]='K'; s[2]=' '; s[3]=0;
         if (GetDocBib(FindDoc(pc->arg[0].name),s+3,12000) > 0) {
            SetStringW(s);
         } else
            SetString("NO BIB");
         break;
      }
      
    case 36:				 /* BIBITEM */
      {
         char      s[1000];

         strcpy(s, "OK ");
         if (GetBibItem(pc->arg[0].name, pc->arg[1].num, s + 3))
            SetString(s);
         else
            SetString("NO BIB");
         break;
      }
      
    case 37:				 /* TRACE */
      bTrace = pc->arg[0].num;
      SetString("OK");
      break;
      
    case 38:				 /* DOWNLOAD */
      {
         char      path[_MAX_PATH] = "";

         if (!strcmp(pc->arg[0].name, "header")) {
         	char buffer[_MAX_PATH];
			strcpy(buffer,idxpath);
			strcat(buffer,"dsc.txt");
 			fDownload=fopen(buffer,"r");
         } else {
            strcpy(path, etcpath);
            strcat(path, pc->arg[0].name);
            strcat(path, ".txt");
            if ((fDownload = fopen(path, "r")) == NULL) {
               syslog(LOG_ERR, "Can't open %s for download\n", path);
               SetString("NO FILE");
               return; 
               }
         }
         SetString("OK");
         return;
      }

    case 39:				 /* LINE */
      {
   		int i; char *s, *s1;
		char ln[MAXSENTENCE+3];
   		if (feof(fDownload))  {
   			fclose(fDownload);
   			SetString("NO MORE");
   			return;
   		}	
   		strcpy(ln,"OK ");
   		i=fread(ln+3,1,MAXSENTENCE,fDownload);
   		ln[i+3]=0;
		s=s1=ln+3;
		while (*s) {
			if (*s!='\r') *s1++=*s; 
			s++;
		}
		*s1=0;
		SetString(ln);
		return;
   		
  
      }            
   
    case 40:				 /* THIN */
      {
         char      s[40];
         int       i1 = 0;
         int       i2 = 0;

         if (!NewQName(pc->arg[1].name)) {
            SetString("NO FILES");
            break;
         }
         Thin(pc->arg[0].name, pc->arg[1].name, pc->arg[2].num,
              pc->arg[3].num, pc->arg[4].num, &i1, &i2);
         sprintf(s, "OK %d %d", i1, i2);
         SetString(s);
         break;
      }
	case 41:     /* WEB */
		{
		char s[_MAX_PATH+10];
		if (website[0]==0) SetString("NO"); else {
			sprintf(s,"OK %s",website);
			SetString(s);
		}
		break;
		}
	case 42:     /* SCDEF */
		{
			char s[100];
			if (NewSubCorpus(pc->arg[0].name,pc->arg[1].cql))
				sprintf(s,"OK %d",MAXSELTAB); 
			else strcpy(s,"NO");
				SetString(s);
			break;
		}
	case 43:	 /* SCADD */
		{
			AddSubCorpus(pc->arg[0].name,pc->arg[1].num,seltab);
			SetString("OK");
			break;
		}
	case 44:	 /* SCSET */
		{
			if (SetSubCorpusPart(pc->arg[0].name,pc->arg[1].num))
				SetString("OK");
			else
				SetString("NO");
			break;
		}
	case 45: /* FTAB */
		{
			char s[100];
			int n;
			if (!SCRegistered()) {
				strcpy(s,"NO FTAB");
			}
			else
			{
				n=FTable(pc->arg[0].num,pc->arg[1].num,pc->arg[2].num,pc->arg[3].cqlw);
				sprintf(s,"OK %d",n);
			}
			SetString(s);
			break;
		}
	case 46: /* FENTRY */
		{
			xairo_uchar s[1000]; xairo_uchar t[MAXWORD]; int k, k1;
			FEntry(pc->arg[0].num,t,&k, &k1);
			swprintf(s,L("OK %d %d %s"),k, k1,t);
			SetStringW(s);
			break;
		}
	case 47: /* FFREE */
		FFree();
		SetString("OK");
		break;
	case 48: /* SCSAVE */
		{ char p[_MAX_PATH];
			char s[_MAX_PATH+3];
			if (SCSave(p))
			   sprintf(s,"OK %s",p);
			else sprintf(s,"NO");
			   SetString(s);
			break;
		}
	case 49:  /* SCLIST */
		{
			char s[1000];
			sprintf(s,"OK %s",SCList(pc->arg[0].num));
			SetString(s);
			break;
		}
	case 50:  /* SCENUM */
		{
			char s[1000];
			sprintf(s,"OK %s",SCEnum(pc->arg[0].name,pc->arg[1].num,pc->arg[2].num));
			SetString(s);
			break;
		}
	case 51: /* CTAB */
		{
			char s[1000];
			int n=MakeCTab(pc->arg[0].name,pc->arg[1].num,pc->arg[2].num,pc->arg[3].num,pc->arg[4].cqlw);
			if (n >= 0) sprintf(s,"OK %d",n); else strcpy(s,"NO");
			SetString(s);
			break;
		}
	case 52: /* CTABENTRY */
		{
			xairo_uchar s[1000]; xairo_uchar* w; int f; double sc;
			CTab(pc->arg[0].num,&w,&f,&sc);
			swprintf(s,L("OK {%s} %d %f"),w,f,sc);
			SetStringW(s);
			break;
		}
	case 53: /* ACSCORE */
		{
			char s[1000]; double sc; long d;
			BOOL b=ACScore(pc->arg[0].name,pc->arg[1].num,pc->arg[2].num, pc->arg[3].num, pc->arg[4].cqlw,
				&d,&sc);
			if (b) {
				sprintf(s,"OK %d %f",d,sc);
				SetString(s);
			} else SetString("NO");
			break;
		}
	case 54: /* CTABTHIN */
		{
			char s[1000]; int i1=0; int i2=0;
			if (!NewQName(pc->arg[1].name)) {
				SetString("NO FILES");
				break;
			}
			CTabThin(pc->arg[0].name,pc->arg[1].name,
				pc->arg[2].num,pc->arg[3].num,pc->arg[4].num,pc->arg[5].cqlw,
				&i1,&i2,FALSE,"");
			sprintf(s,"OK %d %d",i1, i2);
			SetString(s);
			break;
		}
	case 55: /* CTABOPTIONS */
		CTabOptions(pc->arg[0].num,pc->arg[1].num,pc->arg[2].name,pc->arg[3].num);
		SetString("OK");
		break;
	case 56: /* FORM */
		{
			int i=pc->arg[0].num; xairo_uchar s[1000]; int f;
			xairo_uchar form[MAXWORD];
			if (!SCRegistered()) {
				if (SlowForm(pc->arg[1].cqlw,i,form,&f))
					swprintf(s,L("OK {%s} %d"),form,f);
				else
					wcscpy(s,L("NO"));
			} else {

			if (Form(pc->arg[1].cqlw,i,form,&f))
				swprintf(s,L("OK {%s} %d"),form,f);
			else 
				wcscpy(s,L("NO"));
			}
			SetStringW(s);
			break;
		}
	case 57: /* LEMMINFO */
		{
			char s[1000];
			sprintf(s,"OK %s",LemmInfo());
			SetString(s);
			break;
		}
	case 58: /* LEMMALL */
		{
			char s[1000]; int i;
			strcpy(s,"OK ");
			for (i=0;i<LemmaNSchemes();i++) {
				strcat(s,LemmaName(i));
				strcat(s," ");
			}
			SetString(s);
			break;
		}
	case 59: /* LEMMSEL */
		{
			int i;
			for (i=0;i<LemmaNSchemes();i++) {
				if (!strcmp(pc->arg[0].name,LemmaName(i))) break;
			}
			if (i==LemmaNSchemes())
				SetString("NO");
			else {
				LemmaSetScheme(i,TRUE);
				SetString("OK");
			}
			break;
		}
    case 60:				 /* GETSOL */
      {
         xairo_uchar       *s;
         xairo_uchar        s1[MAX_SOL];
         int         pos;
	      int         len;
         sara_long   code;

         char         sno[200];
		char loc[20];
         bPOS = bSLen = bText = FALSE;
		 strcpy(sno,"?");
         s = GetSol(pc->arg[0].name, pc->arg[1].num, &pos, &len,
                    &code, pc->arg[2].num, sno, loc);
         
         if (code==-1) {
         	SetString("NO SOL");
         	break;
         }	

         if (s == NULL) {
         	swprintf(s1,L("OK -1 -1 0 0 HDR Text not available"));
            SetStringW(s1);
            break;
         }
//         Reformat(pc->arg[0].name, s, &pos, &len);
         if (len > (int) wcslen(s) - pos && (int) wcslen(s) - pos >= 0)
            len = wcslen(s) - pos;
         swprintf(s1, L("OK %d %S %d %d %S %s"), code, sno, pos, len,
			 loc, s);
         SetStringW(s1);
         break;
      }
	case 61: /* CTABFREE */
		CTabFree();
		SetString("OK");
		break;
	case 62: /* LEMMDESC */
		{
			int i;
			char s[1000];
			for (i=0;i<LemmaNSchemes();i++) {
				if (!strcmp(pc->arg[0].name,LemmaName(i))) break;
			}
			if (i==LemmaNSchemes())
				SetString("NO");
			else {
				char* d;
				LemmaDescribe(i,&d);
				sprintf(s,"OK %s",d);
				SetString(s);
			}
			break;
		}
	case 63: /* LEMMTEST */
		assert(FALSE);
	case 64:  /* SOLVEX */
		bXML=TRUE;
		goto solve;

	case 65:  /* LOOKUPA */
		{
			char s[100];
			int i=LookupAtt(pc->arg[0].name,pc->arg[1].name);
			if (i > 0) 
				sprintf(s,"OK %d",i);
			else 
				strcpy(s,"NO");
			SetString(s);
			break;
		}
	case 66: /* ADICT */
		{
			char s[1000]; char ss[1000]; int f;
			strcpy(s,"OK ");
			AttDict(pc->arg[0].num,ss,&f);
			sprintf(s,"OK {%s} %d",ss,f);
			SetString(s);
			break;
		}
	case 67: /* SCCOUNT */
		{
			char s[100];
			int n, nt;
			CountQuery(pc->arg[0].name,pc->arg[1].name,pc->arg[2].num,&n,&nt);
			sprintf(s,"OK %d %d",n,nt);
			SetString(s);
			break;
		}

	case 68: /* SCWORDS */
		{
			char s[100];
			long n=SCCountWords(pc->arg[0].name,pc->arg[1].num);
			sprintf(s,"OK %d",n);
			SetString(s);
			break;
		}

	case 69: /* RFREE */
		break;

	case 70: /* SCNAMES */
		{
			if (SetSubCorpusNames(pc->arg[0].name,pc->arg[1].cql))
				SetString("OK");
			else
				SetString("NO");
			break;
		}
	case 71: /* SCENUMNAMES */
		{ 
			char s[1000];
			strcpy(s,"OK ");
			SCEnumNames(pc->arg[0].name,s+3);
			SetString(s);
			break;
		}
	case 72: /* COUNTCONTENT */
		{
			char s[100];
			int n, nt;
			CountContent(pc->arg[0].name,pc->arg[1].name,pc->arg[2].num,pc->arg[3].name,&n,&nt);
			sprintf(s,"OK %d %d",n,nt);
			SetString(s);
			break;
		}
	case 73: /* SUBFORM */
		{
			int i=pc->arg[0].num; int j=pc->arg[1].num; xairo_uchar s[1000];
			xairo_uchar form[MAXWORD];
			if (SCRegistered()) {
				if (SubForm(pc->arg[2].cqlw,i,j,form))
					swprintf(s,L("OK {%s}"),form);
				else 
					wcscpy(s,L("NO"));
			} else {
				if (SlowSubForm(pc->arg[2].cqlw,i,j,form))
					swprintf(s,L("OK {%s}"),form);
				else 
					wcscpy(s,L("NO"));
			}
			SetStringW(s);
			break;
		}
	case 74: /* GETPATH */
		      {
         char      s[200],
                   s1[200+_MAX_PATH];
         char      path[_MAX_PATH] = "";
		 int pos; sara_ulong it;
//         FILE* f; 
		 BOOL b=TRUE;
         if (strcmp(pc->arg[0].name, corpname))
            {
            SetString("NO");
            break;
         }   
		DocToLoc(pc->arg[1].num,&pos,&it);
		it=Outer(it);
         GetDocPath(it, s,path);
         sprintf(s1, "OK %s", path);
         SetString(s1);
         break;
      }
	case 75: /* CTABX */
		{
			char s[1000];
			int n=MakeCTabX(pc->arg[0].name,pc->arg[1].name,pc->arg[2].num,pc->arg[3].cqlw);
			if (n >= 0) sprintf(s,"OK %d",n); else strcpy(s,"NO");
			SetString(s);
			break;
		}
	case 76: /* CTABTHINX */
		{
			char s[1000]; int i1=0; int i2=0;
			if (!NewQName(pc->arg[1].name)) {
				SetString("NO FILES");
				break;
			}
			CTabThin(pc->arg[0].name,pc->arg[1].name,0,0,
				pc->arg[3].num,pc->arg[4].cqlw,
				&i1,&i2,TRUE,pc->arg[2].name);
			sprintf(s,"OK %d %d",i1, i2);
			SetString(s);
			break;
		}
	case 77: /* AKCOUNT */
		{
			char s[1000]; int t;
			int n=AKStat(pc->arg[0].namew,pc->arg[1].namew,&t);
			sprintf(s,"OK %d %d",n,t);
			SetString(s);
			break;
		}
   case 78: /* AKFILTER */
	   {
			AKFilter(pc->arg[0].namew,pc->arg[1].namew);
			SetString("OK");
			break;
	   }
   case 79: /* AKCLEAR */
	   {
			AKClear(pc->arg[0].namew);
			SetString("OK");
			break;
	   }
	   }
	   return;
}

BOOL AllBlank(char *s)
{
   while (*s != 0) {
      if (!isspace(*s))
         return FALSE;
      s++;
   }

   return TRUE;
}


void parg(int op, union uarg * a)
{
   switch (op) {
    case OP_USER:               /* FALLTHROUGH */
    case OP_PW:                 /* FALLTHROUGH */
    case OP_NAME:               /* FALLTHROUGH */
    case OP_CQL:                /* FALLTHROUGH */
    case OP_REGEXP:             /* FALLTHROUGH */
    case OP_FILTER:
      return;

    case OP_NUM:
      return;

    case OP_TABLE:
      return;

    case OP_RANGE:
      return;

    default:
      return;
   }
}

void trace(char *ty, char *format,...)
{
   time_t   t;
   char     tt[30];
   char     buf[4096];

   va_list(args);
   time(&t);
   strcpy(tt, ctime(&t));
   tt[strlen(tt) - 1] = 0;
   va_start(args, format);
   vsprintf(buf, format, args);
   syslog(LOG_NOTICE, "%s", buf);
}

static char *logID = 0;
#define LOGHOSTLEN   128
static char logHost[LOGHOSTLEN+1];

void opensyslog(char *id, int flags, int facility,BOOL bL)
{
   char  ap[_MAX_PATH];
   
      
#if !defined(_WIN32)
   if (logID == 0) {
      sprintf(ap, "%ssyslog", tmppath);
      fSysLog = fopen(ap, "w");
//      logID = strdup(logID);
//      gethostname(logHost, LOGHOSTLEN);
//      logHost[LOGHOSTLEN] = '\0';
#else   /* __unix__ */
	if (bSysLog||bL) {
		sprintf(ap,"%ssyslog.txt",tmppath);
		fSysLog=fopen(ap,"wb");
		fprintf(fSysLog,"\0xff\0xfe");
		if (fSysLog==NULL) bSysLog=FALSE;
#endif	/* __unix__ */  	      
	  logHost[0] = '\0';
//      if ((pch = strchr(logHost, '.')) != 0)
//         *pch = '\0';
   }
}

void syslog(int level, char *format,...)
{
   time_t    t;
   char      tt[30]; char buffer[1000];
   va_list(args);
   if (fSysLog == NULL)
      return;
   time(&t);
  	strcpy(tt,ctime(&t));
	tt[strlen(tt)-1]=0;
   fprintf(fSysLog, "%s %s[%d]: ", tt, logHost, getpid());
   va_start(args, format);
   vsprintf(buffer,format,args);
   fprintf(fSysLog,"%s\r\n",buffer);
   fflush(fSysLog);
}

void bootlog(char *format,...)
{
   va_list(args);
   if (fBoot == NULL)
      return;
   va_start(args, format);
   vfprintf(fBoot,format,args);
   fflush(fBoot);
}

void syslogw(int level, const xairo_uchar *format,...)
{
#if defined(_WIN32)
   time_t    t;
   char      tt[30];
   va_list(args);
   if (fSysLog == NULL)
      return;
   time(&t);
  	strcpy(tt,ctime(&t));
	tt[strlen(tt)-1]=0;
	fwprintf(fSysLog, L("%S %S[%d]: "), tt, logHost, getpid());
   va_start(args, format);
   vfwprintf(fSysLog, format, args);
   fwprintf(fSysLog, L("\r\n"));
   fflush(fSysLog);
#else
	return;
#endif
}

#if !defined(_WIN32)
void ALogEvent(char *format,...)
{
   time_t    t;
   char      tt[30];

   va_list(args);
   if (fAdmLog == NULL)
      return;
   time(&t);
  	strcpy(tt,ctime(&t));
	tt[strlen(tt)-1]=0;
   fprintf(fAdmLog, "%s local[%d]: ", tt, getpid());
   va_start(args, format);
   vfprintf(fAdmLog, format, args);
   fprintf(fAdmLog, "\n");
   fflush(fAdmLog);
}
#endif

void Log(char *user, char *pw)
{              
#if !defined(_WIN32)
   if (!LogUser(user, pw)) {
      SetString("NO BADLOG");
      /*
       * Took out the bits that printed the user name and password.
       *
       * People often accidentally type their password when prompted for
       * their username so it's best not to log them.
       */
 
      ALogEvent("Unsuccessful attempt at log from %s (attempt %d)",
                callloc, ntry);
      syslog(LOG_WARNING, "Unsuccessful attempt at log from %s (attempt %d)",
               callloc, ntry);
      ntry++;

      if (ntry >= MAXTRY)
         nLog = -1;
   } else {
      SetString("OK (C) BNC Consortium, 1995");
      ALogEvent("%s logged on from %s", user, callloc);
      syslog(LOG_NOTICE, "%s logged on from %s", user, callloc);

      strcpy(calluser, user);
      nLog = 1;
      bServer = TRUE;
   }

   return;  
#endif   
}


void Logout(void)
{
   int       i;

   bServer = FALSE;

   for (i = 0; i < NFILES; i++)
      if (qtable[i].bUse && !qtable[i].bSave)
         remove(qtable[i].path);

   if (nLog == 1) {
      syslog(LOG_NOTICE, "%s logged out", calluser);
#if !defined(_WIN32)
      ALogEvent("%s logged out", calluser);
#endif      
   }
   nLog = -1;

   if (fSysLog != NULL)
      fflush(fSysLog);

   return;
}


xairo_uchar *GetSol(char *n, int k, int *pos, int *len, sara_long *code,
             int scope, char *sno, char* loc)
{
   FILE             *f;
   int               jj;
   struct locent1    l;
   char             *name;
	*code=-1;
   name = GetQPath(n);
   if ((f = fopen(name, "r")) == NULL) {
      syslog(LOG_ERR, "Couldn't open %s\n", name);
      return NULL;
   }
   fseek(f, (k * sizeof(struct locent1)), SEEK_SET);
   jj = fread(&l, sizeof(l), 1, f);

   /*
    * locent1s are written by the server, so we don't need to
    * fix them!
    */
   /* fix_locent1(&l, 1); */
   fclose(f);

   if (jj == 0)
      return NULL;

   return GetRef(l.lcode, l.pos, l.word, l.lw, l.lw0, pos, len, code, scope, sno, loc);
}


void SQTable(char *n, char* nn, int nsols, int *nt)
{
   FILE             *f;
   FILE             *f1;
   int               i;
   sara_ulong        lc = 0xf000000;
   struct locent1    l;
   char             *name;

   *nt = 0;
   name = GetQPath(n);
   if ((f = fopen(name, "r")) == NULL) {
      syslog(LOG_ERR, "Can't open %s\n", name);
      return;
   }
   name=GetQPath(nn);

   if ((f1 = fopen(name, "w")) == NULL) {
      syslog(LOG_ERR, "Can't open %s\n", name);
      return;
   }

   for (i = 0; i < nsols; i++) {
      fseek(f, (seltab[i] * sizeof(struct locent1)), SEEK_SET);
      fread(&l, sizeof(l), 1, f);
      fwrite(&l, sizeof(l), 1, f1);
      if (lc != l.lcode) {
         lc = l.lcode;
         (*nt)++;
      }
   }
   fclose(f);
   fclose(f1);
   return;
}

void Thin(char *n, char *n1, int how, int ns, int seed, int *k, int *nt)
{
   switch (how) {
    case 0:
      ThinInit(n, n1, ns, nt);
      *k = ns;
      return;

    case 1:
      srand(seed);
      ThinRand(n, n1, ns, nt);
      *k = ns;
      return;

    case 2:
      OnePerText(n, n1, nt);
      *k = *nt;
      return;
   }
}


BOOL isin(sara_long i, sara_long *sel, int nsel)
{
   int       j;

   for (j = 0; j < nsel; j++)
      if (sel[j] == i)
         return TRUE;
   return FALSE;
}

#if !defined(RAND_MAX)
# define RAND_MAX 0x80000000
#endif /* ! RAND_MAX */


sara_ulong r(sara_long n)
{
   return (sara_ulong) ((((double) n) * rand()) / RAND_MAX);
}


void ThinRand(char *n, char *n1, int ns, int *nt)
{
   sara_long        *sel;
   FILE             *f;
   FILE             *f1;
   sara_ulong        lc = 0xf000000;
   struct locent1    l;
   sara_long         ntot;
   int               nsel;
   int               i;
   char             *name;

   name = GetQPath(n);
   if ((f = fopen(name, "r")) == NULL) {
      syslog(LOG_ERR, "Can't open %s\n", name);
      return;
   }
   
   fseek(f, 0, SEEK_END);
   ntot = ftell(f) / sizeof(l);
   rewind(f);
   if ((sel = (sara_long*) malloc(ns * sizeof(sara_long))) == NULL) {
      syslog(LOG_ERR, "malloc() failed");
#if !defined(_WIN32)
      exit(1);       
#else
      abort();
#endif	        
   }
   nsel = 0;

   for (i = 0; i < ns; i++) {
      sara_long      k;
      do
         k = r(ntot);
      while (isin(k, sel, nsel));
      sel[nsel++] = k;
   }
   *nt = 0;

   name = GetQPath(n1);
   
   if ((f1 = fopen(name, "w")) == NULL) {
      syslog(LOG_ERR, "Can't open %s\n", name);
   } else {
      for (i = 0; i < ntot; i++) {
         fread(&l, sizeof(l), 1, f);
         if (!isin(i, sel, nsel))
            continue;
         fwrite(&l, sizeof(l), 1, f1);
         if (lc != l.lcode)
            (*nt)++;
         lc = l.lcode;
      }
      fclose(f1);
   }
   fclose(f);
   free(sel);
   return;
}

void ThinInit(char *n, char *n1, int ns, int *nt)
{
   FILE             *f;
   FILE             *f1;
   int               i;
   sara_ulong        lc = 0xf000000;
   struct locent1    l;
   char             *name;
   *nt = 0;

   name = GetQPath(n);
   if ((f = fopen(name, "r")) == NULL) {
      syslog(LOG_ERR, "Can't open %s\n", name);
      return;
   }
   
   name = GetQPath(n1);
   if ((f1 = fopen(name, "w")) == NULL) {
      syslog(LOG_ERR, "Can't open %s\n", name);
   } else {
      for (i = 0; i < ns; i++) {
         fread(&l, sizeof(l), 1, f);
         fwrite(&l, sizeof(l), 1, f1);
         if (lc != l.lcode)
            (*nt)++;
         lc = l.lcode;
      }
      fclose(f1);
   }
   fclose(f);
   return;
}

void OnePerText(char *n, char *n1, int *nt)
{
   FILE             *f;
   FILE             *f1;
   sara_ulong        lc = 0xf000000;
   struct locent1    l;
   char             *name;

   *nt = 0;

   name = GetQPath(n);
   if ((f = fopen(name, "r")) == NULL) {
      syslog(LOG_ERR, "Can't open %s\n");
      return;
   }

   name = GetQPath(n1);
   if ((f1 = fopen(name, "w")) == NULL) {
      syslog(LOG_ERR, "Can't open %s\n");
   } else {
      while (fread(&l, sizeof(l), 1, f) == 1) {
         if (lc == l.lcode)
            continue;
         fwrite(&l, sizeof(l), 1, f1);
         lc = l.lcode;
         (*nt)++;
      }
      fclose(f1);
   }
   fclose(f);
   return;
}

#define MAXTAG 1000


int MeasureTag(char* s)
{
	char* s0=s; char c; BOOL bQ=FALSE;
	while (c=*s++) {
		if (c=='>' && !bQ) break; 
		else if (c=='\"') bQ=!bQ;
	}
	return s-s0;
}

int MeasureChar(char* s,int *x)
{
	int c; char ss[100]; char* sss=ss;
	int k=IsChar(s,&c);
	if (k==0) {k=IsSpec(s,&sss);c=-1;}
	*x=k;
	if (s[k-1] != ';') (*x)++;
	if (c>0 && c<256 && c != '&') *x=1;
	if (c==0) *x=0;
	return k;
}

void AdjustLength(char* text,int* i0,int n0,int* l0)
{
	int i=*i0+n0;
	int l=*l0-n0;
	if (text[i]=='<') return;
	while (l > 0) {
		char c=text[i];
		if (c=='<') { 
			int n=MeasureTag(text+i);
			i+=n;

		} else if (c=='&') {
			int x;
			int n=MeasureChar(text+i,&x);
			i+=n;
			l-=x;
		} else 
		{
			i++;
			l--;
		}
	}
// Should also cope with ignorable characters and entities
	*l0=i-*i0;
}



xairo_uchar *GetRef(sara_ulong code, sara_long pos, WORDNO lw, int nw, int nw0,
             int *ipos, int *len, int *doc, int scope, char *sno, char* loc)
{
	int p0; int p1; int p2; int p3; int pl; 
	char* pt;
	const char* p; FILE* f; char* s;int x; char* lo;
	p0=p1=p2=pos; p2+=nw+1; p3=pos;
	*doc=LocToDoc(pos,code);
	s=LastSGMLv(&p0,code,"unit",it_global_attribute_value);
	if (s) {
	while (*s) {
		*sno=*s;
		if (*sno==' ') *sno='_';
		sno++; s++;
	}}
	else *sno++='?';
	*sno=0;
	LastSGMLv(&p1,code,"scope",it_global_attribute_value);
	NextSGMLv(&p2,code,"scope",it_global_attribute_value);
	while (scope > 0) {
		p1--; p2++;
		if (LastSGMLv(&p1,code,"scope",it_global_attribute_value)==NULL) p1++;
		if (NextSGMLv(&p2,code,"scope",it_global_attribute_value)==NULL) p2--;
		scope--;
	}
	
	p=GetTextPath(code);
	f=fopen(p,"rb");
	assert(f != NULL);
	if (p2-p1>MAX_SOL) {
		p1=pos-MAX_SOL/3;
		if (p1<0) p1=0;
		p2=pos+MAX_SOL/3;
	}
	pl=p1;
	lo=LastSGMLv(&pl,code,"locale",it_global_attribute_value);
	if (lo && lo[0]) strcpy(loc,lo); else strcpy(loc,"-");
	fseek(f,p1,SEEK_SET);
	pt=(char*) malloc(p2-p1+1);
	x=fread(pt,1,p2-p1,f);
	pt[x]=0;
	fclose(f);
	*ipos=pos-p1; *len=nw;
	decode(code,pt,x,psent,ipos,len);
	free(pt);
	return psent;
}


void TextExtract(sara_ulong text, sara_long l1, sara_long l2, xairo_uchar *sol)
{
   char      id[_MAX_PATH]; char* pt;
   char      path[_MAX_PATH] = "";
   FILE     *fText;
   int       j = 0;
   int ipos=0; int ilen=0;

   GetDocPath(text, id, path);

   if ((fText = fopen(path, "r")) == NULL) {
      syslog(LOG_ERR, "Can't open %s", path);
      return;
   }
   pt=(char*) malloc(l2-l1+1);
   fseek(fText, l1, SEEK_SET);
   fread(pt, 1, (size_t) (l2 - l1), fText);
   fclose(fText);
   pt[l2 - l1] = 0;
   	decode(text,pt,l2-l1,psent,&ipos,&ilen);

//   Trim(psent);

   while (psent[j] != '>')
      j++;

   wcscpy(sol, psent + j + 1);
   free(pt);
}                   

#define MAXPOS 10000000
#define MAXEE 200

void BoundElement(char *el, WORDNO nw, sara_ulong code, sara_long *l1,
                  sara_long *l2, int *it, sara_long *ls, sara_long pos,
                  sara_long len)
{
   int         i;
   sara_long   ltt = -1;
   sara_long   lt;
   int         ik;
   char       *s;
   int         ns = 0;
   char       *w = el;
   char        el0[20];
   char sent[2];
   strcpy(sent,"s");


   el0[0] = '\0';
   if (isdigit(*w)) {
      ns = atoi(w);
      *l1 = LastSGML(nw - ns, code, defscope, ID_TAGO);
      *ls = LastSGML(nw, code, defscope, ID_TAGO);
      *l2 = NextSGML(nw + 1, code, defscope, ID_TAGO);
      *it = LG_SENT;
      return;
   }
   while (s = strtok(w, ",")) {
      w = NULL;
/*
      if (i==nelts)
         continue;
 */
      lt = LastSGMLp(pos, code, s, ID_TAGO);
      if (lt > ltt) {
         strcpy(el0, s);
         ltt = lt;
      }
   }
   if (!el0[0]) {
 /*     strcpy(el0, DOCTYPE); */
/*      ltt = LastSGMLp(pos, code, el0, ID_TAGO); */
		ltt=0;
   }
   for (i = 0; i < nelts; i++)
      if (SGMLComp(eltab[i].name, el0))
         break;
   *it = ik = (i < nelts) ? eltab[i].eltype : LG_SENT;   /* LG_EMPTY was LG_SENT */

   switch (ik) {
    case LG_SENT:  
    	*ls=ltt;
      *l1 = ltt;
//    panic:
      *l2 = NextSGMLp(pos + len, code, sent, ID_TAGO);
      return;

    case LG_EMPTY:
      *l1 = ltt;
/*      *l2 = NextSGML(nw + 1, code, DOCTYPE, ID_TAGE); */
	*l2=MAXEE+ltt;
      *ls = LastSGML(nw, code, defscope, ID_TAGO);
      return;

    case LG_ELEMENT:
      *l1 = ltt;
      *l2 = NextSGMLp(*l1, code, el0, ID_TAGE)+strlen(el0)+3;
      *ls = LastSGMLp(pos, code, defscope, ID_TAGO);
      if (*l2 < pos + len) {
      	*l2 = NextSGMLp(pos + len, code, defscope, ID_TAGO);
		*it=LG_SENT;
		}		
      break;
   }
}


#if defined(SERVER)

int readrec(FILE *f, char *a, char *b, char *c)
{
   if (!readst(f, a))
      return 0;

   readst(f, b);
   readst(f, c);
   return 3;
}

BOOL readst(FILE *f, char *n)
{
   char  c;

   while (isspace(c = fgetc(f)))
      ;

   if (c == EOF)
      return FALSE;

   do {
      *(n++) = c;
      c = fgetc(f);
   } while (!isspace(c) && c != EOF)
      ;

   *n = 0;

   return TRUE;
}

#endif /* SERVER */







BOOL strtest(char *test, char *string)
{
   char *t = test;
   
   while (*t != 0)
      if (*(t++) != *(string++))
         return FALSE;
	if (isalnum(*string)) return FALSE;
   return TRUE;
}



void LeftL(char *s, int n, int *pos, int *len)
{
   int   i;

   if (*pos > n) {
      int   k = *pos - n;

      for (i = 0; i < MAX_SOL - k; i++)
         s[i] = s[i + k];

      for (i = MAX_SOL - k; i < MAX_SOL; i++)
         s[i] = ' ';
   } else {
      int   k = n - *pos;

      for (i = MAX_SOL - 1; i >= k; i--)
         s[i] = s[i - k];

      for (i = k - 1; i >= 0; i--)
         s[i] = ' ';
   }

   *pos = n;
}


int chtab[MAXCENTS];

void ClearChars(void)
{
   int   i;

   for (i = 0; i < MAXCENTS; i++)
      chtab[i] = -1;
}


void AddChar(char *n, int k)
{
   int   i;

   for (i = 0; i < nctab; i++)
      if (!strcmp(n, ctab[i].cname))
         break;

   if (i == nctab)
      return;

   chtab[i] = k;
}


void TranslateChars(char *line, int *pos, int *len)
{
   char *s0 = (char*) malloc(strlen(line)+1); char* s=s0;
   char *s1 = line;
   int   k;
   int  c;
   int   pose = *pos + *len;

   strcpy(s,line);

   while (*s != 0) {
      if (s == s0 + *pos)
         *pos = s1 - line;

      if (s == s0 + pose)
         pose = s1 - line;

      if (k = IsSpec(s, &s1))
         s += k;

      else if ((k = IsChar(s, &c)) && (c != -1)) {
         s += k;
		 if (c) {
			*(s1++) = '\x15';
			*(s1++) = '0';
			*(s1++) = '0';
			*(s1++) = (c/16 > 9) ? (c/16-9+64):(c/16+'0');
			*(s1++) = (c%16 > 9) ? (c%16-9+64):(c%16+'0');
		 }
      } else
         *(s1++) = *(s++);
   }
      if (s == s0 + *pos)
         *pos = s1 - line;

      if (s == s0 + pose)
         pose = s1 - line;

   *s1 = 0;
   *len = pose - *pos;
	free(s0);
   return;
}


int IsSpec(char *s, char **s1)
{
   int   i;
   int   k;

   if (*s != '&')
      return 0;

   for (i = 0; i < nspec; i++)
      if (strtest(specs[i].name, s + 1))
         break;

   if (i == nspec)
      return 0;

   k = strlen(specs[i].name) + 1;
   if (*(s + k) == ';')
      k++;
   TranslateSpec(i, s1);

   return k;
}


void TranslateSpec(int i, char **s)
{
	char* ss=specs[i].disp;
	while (*ss)
		*((*s)++)=*(ss++);
}


int IsChar(char *s, int *c)
{
   int   i;
//   int   j;
   int   k;

   if (*s != '&')
      return 0;

   for (i = nctab - 1; i >= 0; i--)
      if (strtest(ctab[i].cname, s + 1))
         break;

   if (i < 0)
      return 0;

   k = strlen(ctab[i].cname) + 1;
   if (*(s + k) == ';')
      k++;
/*   *c = 0;
   j = ctab[i].val;
   if (j != -1)
      *c = j;
*/
	*c=ctab[i].val;

   return k;
}


void NormCR0(char *s, int *p, int *l)
{
   char     *s1 = s;
   char      sx[MAX_SOL] = "";
   char     *s2 = sx;
   int       p1 = *p + *l;
   int       np0,
             np1;
   int       i;

   while (*s1 != 0 && s2 - sx < MAX_SOL - 10) {
      if (s1 == s + *p)
         np0 = s2 - sx;

      if (s1 == s + p1)
         np1 = s2 - sx;

      if (*s1 == 13 && *(s1 + 1) == 10) {
         *(s2++) = *(s1++);
         *(s2++) = *(s1++);
      } else if (*s1 == 10) {
         *(s2++) = 13;
         *(s2++) = *(s1++);
      } else
         *(s2++) = *(s1++);
   }

   *p = np0;
   *s2 = 0;
   *l = np1 - *p;

   for (i = 0; i < MAX_SOL - 1; i++) {
      if (sx[i] == 0)
         break;
      s[i] = sx[i];
   }

   s[i++] = 0;
   return;
}


void corperr(const char *s, ...)
{
   return;
}


sara_long MatchTags(sara_long pos, int ifx, char* nm)
{
	int k=1;
	sara_long ls, le;
	while (k>0) {
		le=NextSGMLp(pos+1, ifx, nm, ID_TAGE);
		ls=NextSGMLp(pos+1, ifx, nm, ID_TAGO);
		if (ls < le) {
			k++;
			pos=ls;
		} else if (le < ls) {
		k--;
		pos=le;
		} else break;
	}
	return pos;
}





BOOL namechar(char c)
{
	if (isalnum(c)) return TRUE;
	if (c=='-' || c=='.') return TRUE;
	return FALSE;
}




BOOL Empty(char *nm)
{
   int   i;

   for (i = 0; i < nelts; i++)
      if (SGMLComp(eltab[i].name, nm))
         break;

   return eltab[i].eltype == LG_EMPTY || eltab[i].eltype == LG_SENT;
}


sara_long FindHeader(FILE *f)
{
   int   bx = 0;
   char  b[MAXSENTENCE];
   char *b1;

 y:
   fread(b, 1, MAXSENTENCE, f);

   b1 = strstr(b, "<" DOCTYPE);
   if (b1 == NULL) {
      bx += MAXSENTENCE / 2;
      fseek(f,bx, SEEK_SET);
      goto y;
   }
   return bx + b1 - b;
}


BOOL Ignore(char *s)
{
   return FALSE;
}


void ETag(FILE *f, sara_long *l)
{
   int       i;
   char      bu[100];

   fseek(f, *l, SEEK_SET);
   fread(bu, sizeof(bu), 1, f);

   for (i = 0; i < sizeof(bu); i++)
      if (bu[i] == '>')
         break;
	i++;
	while (isspace(bu[i])) i++;
   *l += i;
}


void Trim(xairo_uchar *s)
{
}

/*
void GetDictEntry(FILE *f, sara_long k, char *wd, int maxwd, sara_long *freq, sara_long *forms)
{
   struct di rec[2];
   off_t     kk = (off_t)k * sizeof(rec[0]) + sizeof(struct dhead);

   fseek(f, kk, SEEK_SET);
   fread(rec, sizeof(rec[0]), 2, f);
   fix_di(rec, 2);

   *freq = rec[0].freq;
   strcpy(wd,GetWordList(rec[0].w));
   *forms=rec[1].str-rec[0].str;

//   fseek(f, (off_t)rec.w, SEEK_SET);
//   fread(wd, 1, maxwd, f);
}
*/

char        db[1000];
int         sb = 1000;
int         kb;
int         nb;
sara_long   lb;
int xbx, xby;

/*
int OldGetFirstDictEntry(char *s)
{
   struct di rec;
   struct di recl;
   sara_long x1;
   sara_long x2;
   off_t     kk = sizeof(struct dhead);

//   GetLims(fDict, 0, ndix - 1, s, &x1, &x2);

   fseek(fDict, (off_t)(kk + x1 * sizeof(rec)), SEEK_SET);
   fread(&rec, sizeof(rec), 1, fDict);
   fix_di(&rec, 1);
   fseek(fDict, (off_t)(kk + x2 * sizeof(rec)), SEEK_SET);
   fread(&recl, sizeof(recl), 1, fDict);
   fix_di(&recl, 1);
   fseek(fDict, (off_t)rec.w, SEEK_SET);

   kb = nb = sb;
   lb = (recl.w - rec.w) + kb;
	xbx=x1; xby=x2;
   return (int) x1;
}
*/


xairo_uchar *GetNextDictEntry()
{
	static xairo_uchar w[MAXWORD]; sara_long f, f0;
	if (xbx > xby) return NULL;
	GetDictEntry(fDict, xbx, w, MAXWORD, &f, &f0);
	xbx++;
	return w;
}

/*

char *GetNextDictEntry()
{
   int   i;
   int   j;

   if (kb > lb)
      return NULL;

   for (i = kb; i < nb; i++)
      if (db[i] == 0)
         break;

   if (i < nb) {
      int   x = kb;

      kb = i + 1;

      return db + x;
   }

   if (nb < sb)
      return NULL;

   for (j = kb; j < nb; j++)
      db[j - kb] = db[j];

   nb = fread(db + nb - kb, 1, kb, fDict) + j - kb;
   lb -= kb;
   kb = 0;

   return GetNextDictEntry();
}
*/


BOOL GetLims(FILE *f, sara_long i1, sara_long i2, const xairo_uchar *patt,
             sara_long *x1, sara_long *x2)
{
   sara_long   mid;
   xairo_uchar        s[100];
   int         c;
   sara_long   fr, fo;

   s[0]=0;
   if (i1 > i2)
      return FALSE;

   mid = (i1 + i2) / 2;

   GetDictEntry(f, mid, s, sizeof(s)/sizeof (xairo_uchar), &fr, &fo);

   c = dcmp(s, patt);
   if (c < 0)
      return GetLims(f, mid + 1, i2, patt, x1, x2);
   else if (c == 0) {
      if (i1 == i2) {
         if (x1)
            *x1 = i1;

         if (x2)
            *x2 = i1;

         return TRUE;
      } else if (i1 == mid) {
         if (x1)
            *x1 = i1;

         if (x2)
            *x2 = GetLims(f, i2, i2, patt, NULL, NULL) ? i2 : i1;

         return TRUE;
      } else {
         if (x1)
            if (!GetLims(f, i1, mid, patt, x1, NULL))
				*x1=mid;

         if (x2)
            if (!GetLims(f, mid, i2, patt, NULL, x2))
				*x2=mid;

         return TRUE;
      }
   } else
      return GetLims(f, i1, mid - 1, patt, x1, x2);
}


int ccmp(unsigned char c1,unsigned char c2)
{
//	unsigned int a1=c1;
//	unsigned int a2=c2;
//	if (a1 != a2) return a1-a2;
	return c1-c2;
}
// this seems to be what DOS sort does

/*
int dcmp(const xairo_uchar *s1, const xairo_uchar *s2)
{
	int k;
	while (*s2 != 0) {
	   k=*s1-*s2;
      if (k<0)
         return -1;
      else if (k>0)
         return 1;
      else {
         s1++;
         s2++;
	}}

   return 0;
}
*/

BOOL NewQName(char *name)
{
   int   i;

   for (i = 0; i < NFILES; i++)
      if (qtable[i].bUse && !strcmp(qtable[i].name, name))
         return TRUE;

   for (i = 0; i < NFILES; i++)
      if (!qtable[i].bUse)
         break;

   if (i == NFILES)
      return FALSE;

   strcpy(qtable[i].name, name);
   sprintf(qtable[i].path, "%s%s.sqy", tmppath, name);

   qtable[i].bUse = TRUE;
   qtable[i].bSave = FALSE;

   return TRUE;
}


char *GetQPath(char *name)
{
   int   i;

   for (i = 0; i < NFILES; i++)
      if (qtable[i].bUse && !strcmp(qtable[i].name, name))
         break;

   if (i == NFILES)
      return NULL;

   return qtable[i].path;
}


void FreeQPath(char *name)
{
   int   i;

   for (i = 0; i < NFILES; i++)
      if (qtable[i].bUse && !strcmp(qtable[i].name, name))
         break;

   if (i == NFILES)
      return;
   qtable[i].bUse = FALSE;
   remove(qtable[i].path);
}


char      tpath[_MAX_PATH];

char *GetQTPath(char *name)
{
   strcpy(tpath, GetQPath(name));
   tpath[strlen(tpath) - 1] = 't';
   return tpath;
}


int OpenQPath(char *name)
{
   int   i;
   int   k = InUse(name);

   if (k == 0)
      return 0;

   for (i = 0; i < NFILES; i++)
      if (!qtable[i].bUse)
         break;

   if (i == NFILES)
      return 0;

   strcpy(qtable[i].name, name);
   sprintf(qtable[i].path, "%s%s.sqy", tmppath, name);
   qtable[i].bUse = TRUE;
   qtable[i].bSave = TRUE;

   return k;
}


BOOL Exists(char *name)
{
   char  path[_MAX_PATH] = "";
   FILE *f;

   sprintf(path, "%s%s.sqy", tmppath, name);

   f = fopen(path, "r");
   if (f != NULL)
      fclose(f);

   return (f != NULL);
}


void Dummy(char *name)
{
   char  path[_MAX_PATH];
   FILE *f;

   sprintf(path, "%s%s.sqy", tmppath, name);

   f = fopen(path, "w");
   fclose(f);
}


int InUse(char *name)
{
   char        path[_MAX_PATH] = "";
   FILE       *f;
   sara_long   l;

   sprintf(path, "%s%s.sqy", tmppath, name);

   f = fopen(path, "r");

   if (f == NULL)
      return 0;

   fseek(f, 0, SEEK_END);
   l = ftell(f);
   fclose(f);

   return (int) l / sizeof(struct locent1);
}




char      text[MAXTEXTS][MAX_CODE+1];

int OldFindDoc(char *name)
{
   int   i;

/*
   if (!strcmp(name,"000"))
      return 0;
 */
   for (i = 0; i < ntexts; i++)
      if (!strcmp(text[i], name))
         break;
   return i;
}


void GetPIndex(void)
{
   char  idxpath[_MAX_PATH];
   int   i;

#if defined(_WIN32)
   _fmode = _O_BINARY;
#endif /* ! __unix__ */

   strcpy(idxpath, etcpath);
   strcat(idxpath, PINDEXNAME);
   fPack = fopen(idxpath, "r");
	if (fPack==NULL) return;
   fread(&idxhead, sizeof(idxhead), 1, fPack);
   fix_pihead(&idxhead, 1);
   ntexts = idxhead.n;

   for (i = 0; i < ntexts; i++) {
      int         j;
      sara_long   x;
      struct pii  ie;
      char        path[_MAX_PATH];

      fread(&ie, sizeof(ie), 1, fPack);
      fix_pii(&ie, 1);
      x = ftell(fPack);
      fseek(fPack, ie.path, SEEK_SET);
      fread(path, _MAX_PATH, 1, fPack);
      j = strlen(path);
      if (i == 0)
         strcpy(text[i], "000");
      else {
		  int k;
		  for (k=j-1;k>=0;k--)
			  if (path[k]=='/'||path[k]=='\\') break;
         strcpy(text[i], path+k+1);
	  }
      fseek(fPack,x, SEEK_SET);
   }
}

void GetBIndex()
{
	char idxpath[_MAX_PATH]; int k;
	bNoBib=FALSE;
	k=strlen(bibname);
	if (k>0 && bibname[k-1]==SEPCHAR) {
		strcpy(idxpath,bibname);
		strcat(idxpath,BIBNAME);
	} else {
		strcpy(idxpath, etcpath);
		strcat(idxpath, bibname);
	}
   if ((fBib = fopen(idxpath, "r")) == NULL) {   
   		bNoBib=TRUE;
		return;
   }            
   

   strcpy(idxpath, etcpath);
   strcat(idxpath, PBIBNAME);
   if ((fBibIndex = fopen(idxpath, "r")) == NULL) {
   		bNoBib=TRUE;
   }

   return;
}


void GetDocPathOld(sara_ulong k, char *name, char *path)
{
   struct pii  ie;
   int         i;

   fseek(fPack, (sizeof(idxhead) + k * sizeof(ie)), SEEK_SET);
   fread(&ie, sizeof(ie), 1, fPack);
   fix_pii(&ie, 1);
   fseek(fPack, ie.path, SEEK_SET);
	strcpy(path,txtpath);
   i = strlen(path);
   fread(path + i, 1, _MAX_PATH - i, fPack);

   i = strlen(path);
   if (k == 0)
      strcpy(name, "000");
   else {
	   int j;
	   for (j=i-1;j >= 0;j--)
		   if (path[j]=='\\' || path[j]=='/') break;
	  strcpy(name,path+j+1);
   }
}


#if defined(_WIN32)
int getpid(void)
{
   return 0;
}
#endif /* ! __unix__ */


BOOL GetBibHeader(char *code, int *ty, int *nr)
{
   long          k;
   struct bibhdr  b;
   int            l = 0;
   int            p = 0;
   int            i = FindDoc(code);
   
   if (bNoBib) return FALSE;
   if (i == 0) {
      *ty = 0;
      *nr = 2;
      return TRUE;
   }
   k = i * sizeof(b);
   fseek(fBibIndex, k, SEEK_SET);
   fread(&b, sizeof(b), 1, fBibIndex);
   fix_bibhdr(&b, 1);
   if (b.type == -1)
      return FALSE;
   *ty = b.type;
   *nr = b.nrows;
   return TRUE;
}

#define bncbib1 "The British National Corpus"
#define bncbib2 "Corpus header"

BOOL GetBibItem(char *code, int n, char *buffer)
{
   long             k;
   struct bibhdr     b;
   struct bibentry   bi;
   int               l = 0;
   int               p = 0;
   int               i = FindDoc(code);

	if (bNoBib) return FALSE;
   if (i == 0) {
      if (n == 0)
         strcpy(buffer, bncbib1);
      else
         strcpy(buffer, bncbib2);
      return TRUE;
   }
   k = i * sizeof(b);
   fseek(fBibIndex, k, SEEK_SET);
   fread(&b, sizeof(b), 1, fBibIndex);
   fix_bibhdr(&b, 1);
   if (n >= b.nrows)
      return FALSE;
   fseek(fBibIndex, (b.seek + n * sizeof(bi)), SEEK_SET);
   fread(&bi, sizeof(bi), 1, fBibIndex);
   fix_bibentry(&bi, 1);
   fseek(fBib,bi.name, SEEK_SET);
   fread(buffer, 1, bi.namelen, fBib);
   buffer[bi.namelen] = 0;
   TranslateChars(buffer, &l, &p);
//   NormSpace(buffer, &l, &p);
   return TRUE;
}


void LoadChars(char *file)
{
   char  path[_MAX_PATH];
   FILE *f;
   char  buffer[100];

   strcpy(path, etcpath);
   strcat(path, file);
   strcat(path, ".cs");
   if ((f = fopen(path, "r")) == NULL) {
      syslog(LOG_ERR, "Can't open %s", path);
      return;
   }
   ClearChars();

   while (fgets(buffer, sizeof(buffer), f)) {
      char *s = strtok(buffer, " \n\t\r");
      int   k;

      if (s == NULL)
         continue;
      k = atoi(strtok(NULL, " \n\t\r"));
      AddChar(s, k);
   }
   fclose(f);
}


BOOL DTDelt(int n, char *s)
{
   if (n < npos)
      sprintf(s, "POS %s", postab[n].name);
   else if (n < npos + nelts)
      sprintf(s, "ELT %s %d", eltab[n - npos].name,
              eltab[n - npos].eltype);
   else if (n < npos + nelts + natts)
      sprintf(s, "ATT %s %s %d %s", atttab[n - npos - nelts].name,
              atttab[n - npos - nelts].element,
              atttab[n - npos - nelts].attv,
              atttab[n - npos - nelts].attdesc);
   return n < npos + nelts + natts;
}


void ScratchFile(char *s)
{
   int   ifile = 0;
   FILE *f;

   do {
      sprintf(s, "%swl%06d", tmppath, ifile++);
   } while (Exists(s));

   f = fopen(s, "w");
   fclose(f);
}

#if !defined(_WIN32)
/*
 * Detach from our controlling tty
 */

void detach(void)
{
   pid_t pgrp;

   /*
    * if our parent is init (pid 1) we don't have a controlling
    * terminal
    */
   if (getppid() != 1) {
      switch(fork()) {
       case -1:
//         syslog(LOG_ERR, "fork() failed [%d]", errno);
//         perror("fork() failed");
         exit(1);
         break;

       case 0:
#if !defined(NO_SETSID)
         if ((pgrp = setsid()) == -1) {
//            syslog(LOG_ERR, "setsid() failed [%d]", errno);
//            perror("setsid() failed");
            exit(1);
         }
#else
         if ((pgrp = setpgrp(getpid(), 0)) == -1) {
//           syslog(LOG_ERR, "setpgrp() failed [%d]", errno);
//            perror("setpgrp");
            exit(1);
         }
#endif
         break;

       default:
         exit(0);
      }
   }
}                    
#endif /* __unix__ */       

void DelAll(char*,char*);                        

void CleanUp()
{
	DelAll(tmppath,"*.sqy");
	DelAll(tmppath,"*.wl");
}

void DelAll(char* path,char* ext)
{                     
#if defined(_WIN32)
#ifndef WIN32
	char p[_MAX_PATH];                 
	struct _find_t fi;
	sprintf(p,"%s%s",path,ext);
	if (_dos_findfirst(p,_A_NORMAL,&fi)) return;
	sprintf(p,"%s%s",path,fi.name);
	remove(p);
	while (_dos_findnext(&fi)==0) {
		sprintf(p,"%s%s",path,fi.name);
		remove(p);
	}
#endif
#endif
}		
	
struct _sds {
	xairo_uchar w[MAXWORD];
	int f, f1;
};

struct _sds* thedict;

int FTable(int k0,int k1,int k2,const xairo_uchar* patt)
{
	struct regexpnew* re; int i0;
	int t=0; int tt; 
	xairo_uchar w[MAXWORD]; int f, f1;
	if (k0==-1) k0=100000000;
	if (k1==-1) k1=0;
	if (k2==-1) k2=100000000;
	re = ParseRegExp(patt,FALSE);
	if (re==NULL) return 0;
	i0=getFirstFtab();
	while (getNextFtab(&i0,w,&f,&f1))
	{
		if (f > k2) continue; 
		if (f < k1) break;
		t++;
		if (t==k0) break;
	}
	thedict=(struct _sds*) malloc(t*sizeof (struct _sds));
	if (thedict==NULL) {
		return 0; 
	}
	tt=0; i0=getFirstFtab();
	while (getNextFtab(&i0,w,&f,&f1) && tt < t)
	{
		if (f > k2) continue;
		if (f < k1) break;
		if (!pMatch(w,re)) continue; 
		wcscpy(thedict[tt].w,w);
		thedict[tt].f=f;
		thedict[tt].f1=f1;
		tt++;
		if (tt==k0) break;
	}
	return tt;
}

void FEntry(int k,xairo_uchar* buffer,int* freq,int* forms)
{
	wcscpy(buffer,thedict[k].w);
	*freq=thedict[k].f;
	*forms=thedict[k].f1;
}

void FFree()
{
	free(thedict);
}

int QuickFrequency(const xairo_uchar* w,int nl)
{
	struct stream* s=NULL; int k;
	if (SCRegistered() && !AKExists()) 
		return DicFrequency(w);
	if (nl) 
		s=MakeAllPosStream(w,GetContName(nl-1));
	else 
		s=MakeLemmaStream(w);  

	if (s==NULL) return 0;
	k=XCountStream(s);
	FreeStreams();
	return k;
}

BOOL QuickFrequencyG0(const xairo_uchar* w)
{
	struct stream* s=NULL; BOOL b;
//	if (SCRegistered()) 
//		return DicFrequency(w);
	s=MakeLemmaStream(w);  
	if (s==NULL) return 0;
	b=XAdvanceStream(s);
	FreeStreams();
	return b;
}




int DicSB(const xairo_uchar* w)
{
    sara_long   x1;
    sara_long   x2;
    if (GetLims(fDict, 0, ndix - 1, w, &x1, &x2)) {
		struct di rec;
		long kk = x1 * sizeof(rec) + sizeof(struct dhead);
		fseek(fDict, kk, SEEK_SET);
		fread(&rec, sizeof(rec), 1, fDict);
		fix_di(&rec, 1);
		return rec.str;
    } else
		return 0;
}


sara_long      Solve0(char *name, int *nt)
{
   FILE             *f;
   sara_ulong        lc = 0xf000000;
   struct locent1    l;
   char             *n;
   sara_long k=0;
   *nt = 0;

   n = GetQPath(name);
   if ((f = fopen(n, "r")) == NULL) {
      syslog(LOG_ERR, "Can't open %s\n", n);
      return 0;
   }
   while (fread(&l,sizeof l,1,f)) {
       if (lc != l.lcode)
            (*nt)++;
	   lc=l.lcode;
	   k++;
   }
   fclose(f);
   return k;
}
		

}