/*
 * External Type definitions
 *
 * These are the structures used in the SARA data files
 *
 * $Id: Exttypes.h,v 1.2 2005-10-04 09:05:29 tonydodd Exp $
 *
 * $Log: not supported by cvs2svn $
 * Revision 1.3  1996/05/12 23:36:15  scotty
 * v927 fixes from Tony.
 *
 * Revision 1.2  1996/02/14 22:01:04  scotty
 * Ported back to DOS by Tony.
 *
 * Revision 1.1  1996/01/30 20:48:12  scotty
 * Initial revision
 *
 */
#ifdef __cplusplus
extern "C" {
#endif


#ifndef EXTTYPES_H
# define EXTTYPES_H

/*
 * On Unix systems:
 *
 *    sara_int    4 bytes
 *    sara_long   4 bytes
 *    sara_ulong  4 bytes
 *    sara_double 8 bytes
 *
 * On DOS/Windows
 *
 *    sara_int    2 bytes
 *    sara_long   4 bytes
 *    sara_ulong  4 bytes
 *    sara_double ? bytes
 */

//#  pragma pack(2)

typedef int           sara_int;
typedef int            sara_long;
typedef unsigned int   sara_ulong;
//typedef double          sara_double;
//typedef long            off_t;

typedef sara_long WORDNO;

/*
 * The header for a .pdx file
 */
struct pihead {
	sara_int n;
};

/*
 * A path entry in a .pdx file
 *
 * The name member used to be 10 chars, but on SunOS it was padded to
 * 12, so it's now 12 in case we find a system that doesn't pad it
 * out.
 * Not on the PC - Tony
 */
struct pii {
   char        name[10];
   sara_long   path;            /* offset to path string */
};

/*
 * An entry in a .acc file
 */
struct accentry {
   sara_int    x;
   sara_int    y;
};

/*
 * The header of an index file (idxNNN/NNN/...)
 *
 * The file structure is
 *   1            struct ihdr
 *   ihdr.ndict   struct dictent
 *   ihdr.nill    struct locent
 *   ihdr.nbytes  char
 */
struct ihdr {
   sara_int    version;         /* Version of Indexer that was used  */
   sara_int    ndict;           /* number of dicent structures       */
   sara_int    nill;            /* number of locent structure        */
   sara_int    nbytes;          /* number of data bytes              */
   sara_int    bOS;
};

/*
 * A dictionary entry in an index file
 */
struct dicent {
   sara_int    dwd;
   sara_int    type;
   sara_int    extra;
   sara_int    first;
};

struct locent {
//   sara_int    bNew;
//   sara_ulong  lcode;
//   WORDNO      word;
//   sara_long   pos;
//   sara_long	len;
   int  lcode;
   int  word;
   int  pos;
   int	len;
//   int sb, pre, nxt, seq;
   int sb, seq;
   int par;
};

struct locent1 {
//   sara_int    bNew;
   sara_ulong  lcode;
   WORDNO      word;
   sara_long   pos;
   sara_int    lw, lw0, wc;
//   int sb, pre, nxt;
   int sb, nxt;
};

/*
 * A header entry in a bibliography file
 */
struct bibhdr {
   sara_int    type;
   sara_long   seek;
   sara_int    nrows;
};

/*
 * A bibliography entry
 */
struct bibentry {
   sara_long   name;
   sara_int    namelen;
};

// Changed 24/1/00
struct di {
   sara_long   w;
   sara_long   freq;
   sara_long   extra;
   sara_long   str;
};

struct adi {
	sara_long w;
	sara_long freq;
};

// Changed 24/1/00
struct dhead {
	sara_long v0, v1;
	sara_long v2;
   sara_long nw;
   sara_long nu;
};

// Added 24/1/00
struct strb {
	sara_int bLast;
	sara_long word;
	sara_int ihash;
	sara_int type;
	sara_int extra;
	sara_int first, ff;
	sara_int last, lf;
	sara_int frobpos, frobno;  
	sara_long wc;
};

struct index_sig {
	char name[101];
	sara_int nelts;
	sara_int natts;
	sara_int npos;
	sara_ulong ntexts;
	sara_int nhash, nhashsig;
	sara_int hashshift, hashlast;
	sara_int maxiloc;
};		
	

//#  pragma pack()

#endif /* EXTTYPES_H */

#ifdef __cplusplus
}
#endif