#include "Corpinc.h"
#define SYNSIZE 99
#define WORDSIZE 99
#define MAXLLINE 80

int nscheme;
void loadfrobtags();
char* TagFrob(char*);
int ClassifyPOS(char*);
void LemmaLoad();
void LoadLemmaList();
void LoadTypeList();

typedef int q_semtag;

typedef struct                  /* main sentence structure */
        {
        int     decision,       /* lemma decision code */
                lexrule_no;     /* index of appropriate lexeme rule */
        char    word[WORDSIZE], /* lower case version of word */
                stem[WORDSIZE], /* stem of word */
                stem2[WORDSIZE],        /* stem of word used in SDF's Lemmatization stuff */
                suffix[WORDSIZE],       /* remainder of word - after stem2 has b
een subtracted */
                lexeme[WORDSIZE], /* lexeme as output in extra column */
                syntag[SYNSIZE],        /* syntax tag */
                *orig;          /* original form of word */
        } unit;

#define ENDOFSTR 0


#define LEMWORD_FN      "lemwordlist."
#define LEMSUFFIX_FN    "lemsuffixlist."
#define LEMTAG_FN       "lemtaglist."
#define LEMFROB_FN      "fmul."

BOOL is_consonant(char);
BOOL is_vowel(char);
BOOL open_resource_file(FILE**,char*);
BOOL is_punc(char*);
BOOL is_idiom(int);

BOOL initialise_lexrules(char*);
void find_lexeme();

extern unit sent;

struct _p {
	char* item;
}
	punc[99];

#pragma warning(disable:4018)

#define MAX_NO_WORD_RULES  4000  /* max no. of word rules per tag */
#define MAX_NO_SUFF_RULES   40  /* max no. of suffix rules per tag */
#define MAX_NO_TAGS     300     /* max no. of basic Claws2 tags */

typedef struct {
	char tag[SYNSIZE];
	char newtag[SYNSIZE];
	int  no_wrules;
	int  no_srules;
	char *word_rules[2][MAX_NO_WORD_RULES];   
	char *suff_rules[5][MAX_NO_SUFF_RULES];   
        char *default_rule[2];
        } tag_unit;

tag_unit rule_set[MAX_NO_TAGS];

char    *nulstr 	= "";           /* null string */
char    *nulltag 	= "@";       /* Lexeme tag for null fields */
char    *nullsyn 	= "NULL";       /* Syntax tag for null fields  */

static int	infile_empty    = FALSE;        /* flags */

int	no_null_tags    = 0,
	no_tag_units    = 0,            /* no. tag units after each set of lex rules has been read in */
        no_ambiguous_lexemes = 0,        /* count of tokens marked AMBIGUOUS */
	no_d_rules_used = 0,		/* counters for default rule uses */
	no_i_rules_used = 0,		/* counters for idiom rule uses */
	no_s_rules_used = 0,		/* counters for suffix rule uses */
	no_p_rules_used = 0,		/* counters for punctuation rule uses */
	no_w_rules_used = 0;		/* counters for word rule uses */

/* prototypes */
void apply_irule();
void apply_prule();
void apply_drule();
void derive_lexeme(char*,char*,char*,char*,int);
void get_stem(char*,char*,char*);
int apply_srule();
int apply_wrule();
int find_lexrule();
void find_lexeme();
void read_line(FILE*,char*);
void initialise_drules(FILE*,FILE*);
void initialise_srules(FILE*,FILE*);
void initialise_wrules(FILE*,FILE*);
BOOL initialise_lexrules(char*);
int is_rule(char [MAXLLINE]);
int is_tagline(char [MAXLLINE]);
void check_overflow(int,int,char*,char*);
char *string_it(char*,char*);
char *suffmatch(char*,char*,char*);
void upper(char*);
int compare(char*,char*);

/**********************************************************************************/
/* Function to dynamically allocate memory to hold a copy of string. If insufficient
   memory available, terminates the program with an error message.  */

char *string_it(char* in_string, char* err_mes)

//char *in_string;	/* String to copy into dynamically allocated store */
//char *err_mes;		/* Detailed error message to print on failure */

{
	char *ret_val;

	if ((ret_val = (char *)malloc((unsigned)strlen(in_string)+1)) == NULL)
	{
		fprintf(stdout,"NOT ENOUGH MEMORY: data=%s: %s", in_string, err_mes);
		exit(1);
	}
	strcpy(ret_val,in_string);
	return(ret_val);
}

void upper(char* s)
/* convert all characters in a string to upper case */
{
        while(*s)
        {
                if((*s <= 'z') && (*s >= 'a'))
                {
                        *s = (*s - 32);
                }
                s++;
        }
}/* end of upper() */

int compare(char* s1,char* s2)
/* case insensitive Strcmp */

{
        while(*s1 || *s2)
        {
                if((*s1 != *s2) && (*s1 != (*s2 + 32)) && (*s1 != (*s2 - 32)))
                        return(FALSE);
                else
                {
                s1++;
                s2++;
                }
        }
        return(TRUE);
}

void check_overflow(int i,int l,char* c1,char* c2)
/* Rgg function for generalised checking   */

{
if (i == l)
        {
        fprintf(stdout,"overflow on array '%s' at '%s'\n", c1, c2) ;
        exit(1) ;
        }
} /*end of check_overflow  */


int is_tagline(char aaa[MAXLLINE])
/* check whether input line contains a tag */
{
	return((isalpha(aaa[0])) || (aaa[0] == '$') || (aaa[0] == '&'));
}       /* end of is_tagline  */

int is_rule(char aaa[MAXLLINE])
/* check whether input line contains a lex rule */ /* seems like a pretty grotty check */
{
	return(isspace(aaa[0]));
}       /* end of is_rule  */

char *suffmatch(char* s1,char* s2,char* s3)
/* reverse - case sensitive - comparision */
{
	char *e1, *e2;
	int con = FALSE;   /* records whether last element was consonant variable    */
	char var = '\0';         /* records its value */

	if(strlen(s1) >= strlen(s2))
	{
		e1 = s1+strlen(s1);             /* Word   */
		e2 = s2+strlen(s2);             /* Suffix template */

		while(e2 >= s2)
		{
			if(*e2 != *e1)
			{
				if((*e2 == 'C') && is_consonant(*e1))          /* check the wildcards      */
				{
					if((con  == TRUE) && (*e1 != var))
						return(NULL);
					else if((con == TRUE) && (*e1 ==  var))
					{
						var = '\0';
						con = FALSE;
					}
					else
					{
						var = *e1;
						con = TRUE;
					}
					e2--;
					e1--;
				}
				else if((*e2 == 'V') && is_vowel(*e1))
				{
					con = FALSE;
					e2--;
					e1--;
				}
				else return(NULL);
			}
			else
			{
				con = FALSE;
				e2--;
				e1--;
			}
		}
                /* add in suffix length test here before returning position - if it fails, simply returns NULL */
                if(*s3 == '\0')
                {
                        /* fprintf(stdout, "no length constraint\n"); */
                	return(e1+1);
                }
                else if((*s3 == '>') && ((strlen(s1) - strlen(s2)) > atoi(s3+1))) 
                {
                       /*  fprintf(stdout, "length of word: %d, length of suffix: %d, length of root: %d, constraint %d, Word: %s, Suffix: %s\n", strlen(s1), strlen(s2), (strlen(s1) - strlen(s2)), atoi(s3+1), s1, s2); */
		        return(e1+1);
                }
                else if((*s3 == '<') && ((strlen(s1) - strlen(s2)) < atoi(s3+1)))
                {
                       /*  fprintf(stdout, "length of word: %d, length of suffix: %d, length of root: %d, constraint %d, Word: %s, Suffix: %s\n", strlen(s1), strlen(s2), (strlen(s1) - strlen(s2)), atoi(s3+1), s1, s2); */
		        return(e1+1);
                }
	}
	return(NULL);
}


BOOL initialise_lexrules(char* tagset)
{
	FILE	*rfp1, *rfp2, *rfp3;      /*Rule File Pointers
                                            wfile = Newlexwordlist
                                            sfile = Newlexsufflist
                                            dfile = Newlexdeflist
                                          */
	char	rulename1[40];
	char	rulename2[40];
	char	rulename3[40];

	strcpy(rulename1, LEMWORD_FN);
	strcat(rulename1,tagset);
	strcpy(rulename2, LEMSUFFIX_FN);
	strcat(rulename2,tagset);
	strcpy(rulename3, LEMTAG_FN);
	strcat(rulename3,tagset);
        
        if (!open_resource_file(&rfp1, rulename1)) return FALSE;
        if (!open_resource_file(&rfp2, rulename2)) return FALSE;
        if (!open_resource_file(&rfp3, rulename3)) return FALSE;

    initialise_wrules(NULL, rfp1);
	initialise_srules(NULL, rfp2);
	initialise_drules(NULL, rfp3);
	loadfrobtags(tagset);
	fclose(rfp1);
	fclose(rfp2);
	fclose(rfp3);
	/* quicksort_lexrule_table(rule_set, no_tag_units);*/
	return TRUE;
}  /* end of initialise_lexrules  */


void initialise_wrules(FILE* repfp, FILE*ifp)
/* A function to read in Word-specific rules */
{
	int	i,k, more, no_tags_found = 0, no_rules_found =0;
	char	*p, *q, *r;
	char	line[MAXLLINE];
	char	*err_mes = "reading word rules";

//	fprintf(repfp, "Lemmings Word rules ... ");

	while(!infile_empty)
	{
		read_line(ifp,line);
		for(i=0; i < MAX_NO_TAGS; i++)
		{
			if(is_tagline(line))
			{
				no_tags_found++;
				p = line;
				q = p;
				while(!isspace(*q)) /*parse tag out of line and add null character */
					q++;
				*q = '\0';
				strcpy(rule_set[i].tag,p);
		  		/* fprintf(stdout, "\nTag:  %s\n", rule_set[i].tag); */
				p = q;

				read_line(ifp,line);
				for(k=0; k< MAX_NO_WORD_RULES; k++)
				{
					if(is_rule(line))
					{
						/*fprintf(testptr,"%s\n",line);*/
						no_rules_found++;
 						p = line;
                               			q = p;
                               			while(isspace(*q)|| *q == '*')
							q++; /* skip empty space at beginning of rule line */
                               			p = q;
                               			while((!isspace(*q)) && (*q != ENDOFSTR)) /*parse rule head out of line and add null character */
                                       			q++;
						more = (*q == ENDOFSTR)? FALSE: TRUE;	/* reset flag if already at end of string */
                               			*q = '\0';
                               			rule_set[i].word_rules[0][k] = string_it(p,err_mes);
						r = p;
						if (more && *(p+1))
						{
                                                	q++;
                                                	while(isspace(*q))
                                                        	q++;	/* skip empty space before 2nd element of rule */
							if (*q != ENDOFSTR) 		/* carry on only if not at end of string */
							{
                                                		p = q;    /* set pointer p to start of Rule 2nd element  */
                                                		while((!isspace(*q)) && (*q != ENDOFSTR)) /*parse rule 2nd element out of line and add null character */
                                                       	 		q++;
                                                		*q = '\0';
                                                		rule_set[i].word_rules[1][k] = string_it(p,err_mes);
							}
						}
						else
							rule_set[i].word_rules[1][k] = string_it(r,err_mes);

                         /*       			fprintf(stdout, "  Verb Form: %s\n", rule_set[i].word_rules[0][k]);
                                                fprintf(stdout, "     Lexeme: %s\n", rule_set[i].word_rules[1][k]); */
						read_line(ifp,line);
                        		}
					else
						break;
				}
				rule_set[i].no_wrules = k;
			}
		}
        }
	no_tag_units = no_tags_found;
//	fprintf(repfp, "Number of tags: %d; ", no_tags_found);
//	fprintf(repfp, "Number of rules: %d\n", no_rules_found);
/*	for(i=0;i<no_tags_found;i++)
		fprintf(stdout, "TAG:- %s\n", rule_set[i].tag); */
}    /* end of initialise_wrules */


void initialise_srules(FILE* repfp,FILE* ifp)
/* A function to read in Suffix rules and add them to appropriate place in rules table */
{
	int	i,k, more, no_tags_found = 0, no_rules_found =0, x,found,res;
	char	*p, *q;
	char	line[MAXLLINE];
	char	*err_mes = "reading suffix rules";
	char	current_tag[SYNSIZE];

//	fprintf(repfp, "Lemmings Suffix rules ... ");

	infile_empty=FALSE;
	for(i=0; i< no_tag_units; i++)
		rule_set[i].no_srules = 0; /* initialise suffix rules count for tags already in table to zero  */
	while(!infile_empty)
        {
                read_line(ifp,line);
                for(i=0; i < MAX_NO_TAGS; i++)
                {
			if(is_tagline(line))
			{
				no_tags_found++;
                                p = line;
                                q = p;
                                while(!isspace(*q)) /*parse tag out of line and add null character */
                                        q++;
                                *q = '\0';
                                strcpy(current_tag,p);
                                /* fprintf(stdout, "\nSuffrules Tag:  %s\n", current_tag); */
                                p = q;
				for(x = 0; x < no_tag_units; x++)  /* search for current tag in rules table */
				{
					found = FALSE;
					res = strcmp(current_tag, rule_set[x].tag);
					if(res == 0)
					{
						found = TRUE;
						/* fprintf(stdout, "Tag %s found in rules table\n", current_tag); */
						break;
					}
				}
				if(!found)
				{
					/* fprintf(stdout, "WARNING: Tag %s NOT found in rules table\n", current_tag); */
					/* fprintf(stdout, "Inserting tag %s in rules table\n", current_tag); */
					strcpy(rule_set[no_tag_units].tag, current_tag);
					rule_set[no_tag_units].no_wrules = 0;   /* initilise word rules count to zero for items found first in suffix rule list */
					x = no_tag_units;
					no_tag_units++; 
				}
				/* fprintf(stdout, "Test: Current Tag is: %s, x = %d\n", current_tag, x);*/

                                read_line(ifp,line);
				for(k=0;k< MAX_NO_SUFF_RULES; k++)
				{ 
					if(is_rule(line))
					{
						no_rules_found++;
						p = line;
						q = p;
						while(isspace(*q))
							q++;  /* skip empty space at beginning of rule line */
						p = q;
						while((!isspace(*q)) && (*q != ENDOFSTR)) /* parse first thing out of rule line and add null character */
							q++;
						more = (*q == ENDOFSTR) ? FALSE: TRUE; /* reset flag if already at ENDOFSTR */
						*q = '\0';
						if((*p == '<') || (*p == '>'))  /* if this is a stem condition store it and get 1st rule element */
						{
							rule_set[x].suff_rules[0][k] = string_it(p, err_mes);
						/*	fprintf(stdout, "Stem length condition found: %s %s\n", rule_set[x].suff_rules[0][k], rule_set[x].tag); */
							if(more && *(p+1))
							{
								q++;
								while(isspace(*q))
									q++;  /* skip spaces before first rule element */
								if(*q != ENDOFSTR)
								{
									p = q; /* set pointer to start of 1st element of rule*/
									while((!isspace(*q)) && (*q != ENDOFSTR)) /* parse out first element and add null character */
										q++;
									more = (*q == ENDOFSTR) ? FALSE: TRUE; /* reset flag if already at ENDOFSTR */
									*q = '\0';
								}
							}
						}
						else
							rule_set[x].suff_rules[0][k] = nulstr;  /* initialise null length constraints */

						/* now copy first element of rule to table */
						rule_set[x].suff_rules[1][k] = string_it(p, err_mes);
						if(more && *(p+1))
						{
							q++;
							while(isspace(*q))
								q++;
							if(*q != ENDOFSTR)
							{
								p = q;
								while((!isspace(*q))&& (*q != ENDOFSTR)) /* get 2nd element */
									q++;
								*q = '\0';
								rule_set[x].suff_rules[2][k] = string_it(p,err_mes);
							}
						}
						else
							rule_set[x].suff_rules[2][k] = nulstr;

				/*		fprintf(stdout, "Rule elements: %s, %s, %s\n", rule_set[x].suff_rules[0][k],rule_set[x].suff_rules[1][k], rule_set[x].suff_rules[2][k]); */
						read_line(ifp,line);
					}
					else
						break;
				}
				rule_set[x].no_srules = k;
				/* fprintf(stdout, "TAG: %s - No. of Suffix Rules: %d\n", rule_set[x].tag, rule_set[x].no_srules);*/
			}
		}
	}
//	fprintf(repfp, "Number of tags: %d ", no_tags_found);
//	fprintf(repfp, "Number of rules: %d\n", no_rules_found);
/* 	for(i=0;i<no_tag_units;i++)*/
		/* fprintf(stdout, "TAG:- %s\n", rule_set[i].tag); */

} /* end of initialise_srules */


void initialise_drules(FILE* repfp, FILE* ifp)
/* A function to read in Default rules and add them to appropriate place in rules table */
{
	int	i, no_tags_found = 0, x,found,res;
	char	line[MAXLLINE];
	char	*err_mes = "reading suffix rules";
	char	current_tag[SYNSIZE];
	char	new_tag[SYNSIZE];

//	fprintf(repfp, "Lemmings Default rules ... ");

	infile_empty=FALSE;
	while(!infile_empty)
        {
                read_line(ifp,line);
                for(i=0; i < MAX_NO_TAGS; i++)
		{
			if(is_tagline(line))
			{
				int t;
				no_tags_found++;
				sscanf(line,"%s %n",current_tag,&t);
				strcpy(new_tag,line+t);
//                                p = line;
//                                q = p;
//                                while(!isspace(*q)) /*parse tag out of line and add null character */
//                                        q++;
//                                *q = '\0';
//                                strcpy(current_tag,p);
//                                p = q;
				found=FALSE;
				for(x = 0; x < no_tag_units; x++)  /* search for current tag in rules table */
				{
					res = strcmp(current_tag, rule_set[x].tag);
					if(res == 0)
					{
						found = TRUE;
						/* fprintf(stdout, "Tag %s found in rules table\n", current_tag); */
						break;
					}
				}
				if(!found)
				{
					/* fprintf(stdout, "WARNING: Tag %s NOT found in rules table\n", current_tag); */
					/* fprintf(stdout, "Inserting tag %s in rules table\n", current_tag);*/
					strcpy(rule_set[no_tag_units].tag, current_tag);


					rule_set[no_tag_units].no_wrules = 0;   /* initilise word rules count to zero for items found first in suffix rule list */
					rule_set[no_tag_units].no_srules = 0;   /* initilise suffix rules count to zero for items found first in suffix rule list */
					x = no_tag_units;
					no_tag_units++; 
				}
				/* fprintf(stdout, "Test: Current Tag is: %s, x = %d\n", current_tag, x);*/
				rule_set[x].default_rule[0] = string_it(rule_set[x].tag,err_mes);
				rule_set[x].default_rule[1] = string_it("UNMATCHED",err_mes);
				strcpy(rule_set[x].newtag, new_tag);

                                read_line(ifp,line);
			}
		}
	}
//	fprintf(repfp, "Number of Default rules: %d\n\n", no_tags_found);
	/* fprintf(stdout, "Number of tag units: %d\n", no_tag_units); */
/*	for(i=0;i<no_tag_units;i++)
	{
		fprintf(stdout, "TAG:- %s\n", rule_set[i].tag);
		fprintf(stdout, "DEFAULT RULE:-  %s %s\n", rule_set[i].default_rule[0], rule_set[i].default_rule[1]);
	}
*/
} /* end of initialise_drules */


void read_line(FILE* ifp,char* line)
/* reads a line from input file and appends a terminating character */
{
int	z, c;

	if(infile_empty)
		return;
        z = 0 ;
	c = fgetc(ifp);
        while ((c != '\n') && (c != EOF))
        {
        	check_overflow(z, MAXLLINE, "line", line) ;
		line[z] = c ;
		z++ ;
               	c = fgetc(ifp) ;
	}
	check_overflow(z, MAXLLINE, "line", line) ;
	while (z>0) {
		if (isspace(line[z-1])) z--;
		else break;
	}
	line[z] = '\0' ;
	if(c == EOF)
		infile_empty = TRUE;
}

void find_lexeme()
/* a function to match a word with entries in the lexrules table and assign the appropriate 
lexeme tag */
{
		if(strcmp(sent.syntag, nullsyn) == 0)
		{
			strcpy(sent.lexeme, nulltag);
			no_null_tags++;
			sent.decision = 0;
		}
		else if(is_punc(sent.syntag))	/* check if syntag is punctuation */
		{
			apply_prule();
		}
		else
		{
			sent.lexrule_no = find_lexrule();
//			fprintf(stdout, "Word: %s, Tag: %s, Use Rule: %s  (Rule No. %d)\n", sent.orig, sent.syntag, rule_set[sent.lexrule_no].tag, sent.lexrule_no);
			if(sent.lexrule_no == no_tag_units)  /* i.e. if it is a rule 174 job */
				strcpy(sent.lexeme,sent.word);
			else if(apply_wrule() == TRUE);
//				fprintf(stdout,"Lexeme Found: %s\n\n", sent.lexeme);
			else if(apply_srule() == TRUE) ;
//				 fprintf(stdout,"Lexeme Derived: %s\n\n", sent.lexeme) ;
			else
			{
				apply_drule();
//				fprintf(stdout,"Lexeme Defaulted: %s\n\n",sent.lexeme);
			}
		}
}



int find_lexrule()
/* function to find the appropriate rule by reference to word-tag */
{
	int x = 0;
	while(x < no_tag_units && strcmp(sent.syntag, rule_set[x].tag))
		x++;
	return x;
}
/* end of find_lexrule(tag)*/


int apply_wrule()
/* function to apply lexeme word rule */
{
	int x, found;

	if(rule_set[sent.lexrule_no].no_wrules == 0)
		return(FALSE);
	else
	{
		for(x = 0; x < rule_set[sent.lexrule_no].no_wrules; x++)
		{
			found = compare(rule_set[sent.lexrule_no].word_rules[0][x], sent.orig);
			if(found)
			{
				strcpy(sent.lexeme, rule_set[sent.lexrule_no].word_rules[1][x]);
/*				upper(sent[cur].lexeme);*/
/*				fprintf(stdout, "WORD RULE APPLIED\n\n");*/
				no_w_rules_used ++;
				sent.decision = 1;
				if(strncmp("**", sent.lexeme, 2) == 0)
				{
					no_ambiguous_lexemes++;
		/*			strcpy(sent[cur].lexeme, ambiguoustag);   */
					sent.decision = 6;
				}
				return(TRUE);
			}
		}
		return(FALSE);
	}
}


int apply_srule()
/* function to apply suffix rule */
{
	int x;
	char *suffix;

	if(rule_set[sent.lexrule_no].no_srules == 0)
		return(FALSE);
	else
	{
		for(x = 0; x < rule_set[sent.lexrule_no].no_srules; x++)
		{
                        /* fprintf(stdout, "%s, %s %s\n", rule_set[sent[cur].lexrule_no].suff_rules[1][x], rule_set[sent[cur].lexrule_no].suff_rules[2][x], rule_set[sent[cur].lexrule_no].suff_rules[0][x]); */
			if((suffix=suffmatch(sent.word, rule_set[sent.lexrule_no].suff_rules[1][x], rule_set[sent.lexrule_no].suff_rules[0][x])) != NULL)
                        /* suffmatch does all the work */
			{
				strcpy(sent.suffix, suffix);
	/*			fprintf(stdout," Just copied this suffix: %s %s\n\n", suffix, sent[cur].suffix); */
				get_stem(sent.word, suffix, sent.stem2);
/*				fprintf(stdout,"SUFFIX RULES USED - Word Stem: %s, Suffix %s\n\n", sent[cur].stem2, *//* rule_set[sent[cur].lexrule_no].suff_rules[1][x]*//* sent[cur].suffix);*/
				derive_lexeme(sent.stem2, sent.suffix, rule_set[sent.lexrule_no].suff_rules[1][x], rule_set[sent.lexrule_no].suff_rules[2][x], x);
/*				upper(sent[cur].lexeme);*/
				no_s_rules_used ++;
				sent.decision = 2;
				return(TRUE);
			}
		}
		return(FALSE);
	}
}   /* end of apply_srule */


void get_stem(char* s,char* e,char* st)
/* function to derive stem from word whose suffix has been identified and combine it with the modified suffix */
//char *s;    /* start of string     */
//char  *e;    /* end of string*/
//char  *st;    /* string to hold stem*/
{
/* 	char *test_s = st;
	char *test_st = st; */
	while(s < e)
	{
		*st = *s;
		s++;
		st++;
	}
	*st = '\0';
/* fprintf(stdout, " Testing get_stem: word: %s; suffix: %s; stem: %s\n\n", test_s, e, test_st); */
}   /* end of get_stem   */

void derive_lexeme(char* stem, char* suffix, char* oldsuffix_template, char* newsuffix_template,int x)
/* wildcard version */
{
	int i;
	
	if(newsuffix_template != nulstr)
	{
		for(i=0;i<strlen(sent.stem2);i++)
		{
			sent.lexeme[i] = *stem;
			*stem++;
		}
		while(*newsuffix_template)
		{
			if((*newsuffix_template == 'C') || (*newsuffix_template == 'V'))
				sent.lexeme[i] = *suffix;
			else
				sent.lexeme[i] = *newsuffix_template;
			i++;
			suffix++;
			newsuffix_template++;
		}
		sent.lexeme[i] = '\0';
		/* fprintf(stdout,"  Testing derive_lexeme - lexeme: %s\n\n", sent[cur].lexeme);*/
	}
	else
	{
		strcpy(sent.lexeme,stem);
		/* fprintf(stdout,"  Testing derive_lexeme - null condition - lexeme: %s\n\n", sent[cur].lexeme); */
	}
}


void apply_drule()
/* function to apply default rule to words not found in suffix or word rule lists */

{
	strcpy(sent.lexeme, sent.orig);
/*	upper(sent[cur].lexeme);*/
/*	fprintf(stdout, "DEAFULT RULE APPLIED\n\tLexeme assumed: %s\n\n", sent[cur].lexeme);*/
	no_d_rules_used ++;
	sent.decision = 9;
}/* end of apply_drule */

			
void apply_prule()
/* function to apply default rule to words not found in suffix or word rule lists */

{
	if (strcmp(sent.syntag,punc[10].item) == 0)
		strcpy(sent.lexeme, sent.syntag);
	else if (strcmp(sent.syntag,punc[3].item) == 0)
		strcpy(sent.lexeme, sent.syntag);
	else
		strcpy(sent.lexeme, sent.orig);
/*	upper(sent[cur].lexeme);*/
	no_p_rules_used ++;
	sent.decision = 3;
}/* end of apply_prule */


void apply_irule()
/* function to apply rule to idioms*/

{
	strcpy(sent.lexeme, sent.orig);
/*	upper(sent[cur].lexeme);*/
/*	fprintf(stdout, "TEMPORARY IDIOM RULE APPLIED\n\tLexeme assumed: %s\n\n", sent[cur].lexeme);*/
	no_i_rules_used ++;
	sent.decision = 5;
}/* end of apply_irule */


unit sent;

BOOL open_resource_file(FILE** f,char* path)
{
	char p[_MAX_PATH];
	strcpy(p,etcpath);
	strcat(p,path);
	*f=fopen(p,"r");
	return *f != 0;
}

BOOL is_vowel(char c)
{
	return c=='a'||c=='e'||c=='i'||c=='o'||c=='u';
}

BOOL is_consonant(char c)
{
	if (!isalpha(c)) return FALSE;
	return !is_vowel(c);
}

BOOL is_punc(char* c)
{
	return *c=='.'||*c==';'||*c==':'||*c==',';
}

BOOL is_idiom(int i)
{
	return FALSE;
}

BOOL lemmata0(char*,int,char*,BOOL);
BOOL lemmata1(char*,int,char*);


BOOL lemmata(char* word,int extra,char* headword,BOOL bFrob)
{
	switch (nscheme) {
	case 0:
		if (nlt) {
			sprintf(headword,"%s=%s=%d",word,(extra==HDRPOS)?"HDR":postab[extra%10000].name,extra/10000+1);
			return TRUE;
		}
		if (!bNoPos) {
			sprintf(headword,"%s=%s",word,(extra==HDRPOS)?"HDR":postab[extra%10000].name);
			return TRUE;
		}
// fall through
	case 1:
		strcpy(headword,word);
		return TRUE;
	case 2:
		return lemmata0(word,extra,headword,bFrob);

	}
	return TRUE;
}


BOOL lemmata0(char* word,int extra,char* headword,BOOL bFrob)
{
	BOOL bStop=FALSE; char* c;
	if (extra==HDRPOS) return FALSE;
	strcpy(sent.word,word);
	strcpy(sent.syntag,postab[extra%10000].name);
	if (c=strchr(sent.syntag,'-')) *c=0; // no portmanteaux
	sent.orig=sent.word;
	if (bFrob) {
		strcpy(sent.syntag,TagFrob(sent.word));
	}
	sent.lexeme[0]=0;
	find_lexeme();
	assert(sent.lexeme[0] != 0);
	strcpy(headword,sent.lexeme);
	_strlwr(headword);
	strcat(headword,"=");
	if (sent.lexrule_no < no_tag_units) {
		if (!strcmp(rule_set[sent.lexrule_no].newtag,"STOP")) bStop=TRUE;
		strcat(headword,rule_set[sent.lexrule_no].newtag);
	}
	else 
		strcat(headword,postab[extra%10000].name);
	return !bStop;
}

char* LemmInfo()
{
	return "0 =";
//	return lemmatab[nscheme].info;
//	switch (nscheme) {
//	case 0:
//		if (bNoPos) return "0 =";
//		else if (nlt==0) return "1 = POS";
//		else return "2 = POS=Sense";
//	case 1:
//		return "0 =";
//	case 2:
//		return "1 = TAG";
//	case 3:
//		return "0 =";
//	}
//	return "";
}

#define MAXNFROBTAG 1000

struct _frobtag {
	char w[MAXWORD];
	char t[10];
} frobtag[MAXNFROBTAG];

int nfrobtag;

#define MAXLINEF 80

void loadfrobtags(char* tagset)
{
	char path[_MAX_PATH]; FILE* f; int k=0; char line[MAXLINEF];
	strcpy(path,etcpath);
	strcat(path,LEMFROB_FN);
	strcat(path,tagset);
	f=fopen(path,"r");
	while (fgets(line,MAXLINEF,f)) {
		sscanf(line,"%s %s",frobtag[k].w,frobtag[k].t);
		k++;
	}
	nfrobtag=k;
	fclose(f);
}

char* TagFrob(char* w)
{
	int i;
	for (i=0;i<nfrobtag;i++) 
		if (!strcmp(frobtag[i].w,w)) return frobtag[i].t;
	return "UNC";
}

int nlemmatab=0;

#define MAXLSCHEMES 16

struct _lemmatab lemmatab[MAXLSCHEMES];

int LemmaNSchemes()
{
	return nlemmatab;
}

void LemmaSetScheme(int i,BOOL bRefresh)
{
	nscheme=i;
	nlemma=lemmatab[i].idx;
	if (bRefresh) LoadLemmaList();
}

int LemmaGetScheme()
{
	return nscheme;
}

#define NULLNAME "null"

const char* LemmaName(int i)
{
	return lemmatab[i].name;
}

#define MAXSCHEMES 10

BOOL bEnabled[MAXSCHEMES]={TRUE};

void LemmaEnable(int i,char* param)
{
	switch (i) {
	case 0:
		break;
	case 1:
		break;
	case 2:
		{
			char tagset[5];
			sscanf(param,"%s",tagset);
			initialise_lexrules(tagset);
			break;
		}

	}
	bEnabled[i]=TRUE;
}

BOOL LemmaIsEnabled(int i)
{
	return bEnabled[i];
}

char* ld[4]={"The null scheme maps each word to itself",
"The bnc scheme maps each word to a headword with the same spelling and no part of speech information.",
"The Lancaster scheme was developed at the University of Lancaster.",
"Lemmatisation from an inline attribute."};

void LemmaDescribe(int i,char** s1)
{
	*s1="Obsolete API Call";
}	




