#pragma once

#include "../XairoHeader.h"

XAIRO_NAMESPACE_BEGIN


class CXairoServer;


class CXairoGrammar:public CXairoIGrammar
{
public:
	CXairoGrammar(CXairoServer* s);
	~CXairoGrammar(void);

	class Element;

	XAIRO_API_EXPORT int getNamespaceCount() const;
	class Namespace : public CXairoIGrammar::Namespace
	{
	public: 
		Namespace(const xairo_uchar* url){m_url=url;}
		XAIRO_API_EXPORT const xairo_uchar* getUrl() const {return m_url;}
	private:
		const xairo_uchar* m_url;
	};
	XAIRO_API_EXPORT const Namespace* getNamespace(int) const;

	XAIRO_API_EXPORT int getAddkeyCount() const {return m_naddkeys;}

	class Codebook : public CXairoIGrammar::Codebook
	{
	public:
		Codebook(DOMElement* e,CXairoServer* s);
		Codebook(const xairo_uchar* id,CXairoServer* s);  
		XAIRO_API_EXPORT const xairo_uchar* lookup(const xairo_uchar*) const;
		XAIRO_API_EXPORT int getCount() const;
		XAIRO_API_EXPORT void getPair(int,const xairo_uchar**,const xairo_uchar**) const;
		const xairo_uchar* getId() const;
	private:
		DOMElement *m_ptr;
		CXairoServer* m_server;
	};

	class Addkey : public CXairoIGrammar::Addkey
	{
	public:
		Addkey(const xairo_uchar* id,CXairoServer* s);
		Addkey(const xairo_uchar* n,int nd,const xairo_uchar* g,bool b,DOMElement* e,CXairoServer* s):
		  m_name(n), m_dic(nd), m_gloss(g), m_bLemma(b),m_ptr(e),m_server(s) {}
		XAIRO_API_EXPORT const xairo_uchar* getName() const {return m_name;}
		XAIRO_API_EXPORT const xairo_uchar* getGloss() const {return m_gloss;}
		XAIRO_API_EXPORT bool isLemmaScheme() const {return m_bLemma;}
		XAIRO_API_EXPORT const Codebook* getCodebook() const;
		XAIRO_API_EXPORT const xairo_uchar* getAttribute() const;
		const xairo_uchar* getId() const;
		int GetDic() const {return m_dic;}
	private:
		const xairo_uchar* m_name;
		const xairo_uchar* m_gloss;
		bool m_bLemma;
		DOMElement *m_ptr;
		int m_dic;
		CXairoServer* m_server;
	};
	XAIRO_API_EXPORT const Addkey* getAddkey(int i) const {return m_addkeys[i];}
	const Addkey* getAddkeyById(const xairo_uchar*) const;

	XAIRO_API_EXPORT int getRegionCount() const;

	class Region : public CXairoIGrammar::Region

	{
	public:
		Region(DOMElement* e,CXairoServer* s):m_ptr(e),m_server(s) {}
		Region(const xairo_uchar* id,CXairoServer* s);
		XAIRO_API_EXPORT const xairo_uchar* getName() const;
		XAIRO_API_EXPORT const xairo_uchar* getGloss() const;
		XAIRO_API_EXPORT int getValueCount() const;
		XAIRO_API_EXPORT const xairo_uchar* getValue(int) const;
		XAIRO_API_EXPORT const Addkey* getAddkey() const;
		const xairo_uchar* getId() const;
	private:
		DOMElement* m_ptr;
		CXairoServer* m_server;
	};

	XAIRO_API_EXPORT const Region* getRegion(int) const;


	XAIRO_API_EXPORT int getLemmaSchemeCount() const;

	class LemmaScheme : public CXairoIGrammar::LemmaScheme
	{
	public:
		LemmaScheme(DOMElement* e,CXairoServer* s):m_ptr(e),m_server(s) {}
		LemmaScheme(const xairo_uchar* id,CXairoServer* s);
		XAIRO_API_EXPORT const xairo_uchar* getName() const;
		XAIRO_API_EXPORT const xairo_uchar* getGloss() const;
		XAIRO_API_EXPORT int getFacetCount() const;
		XAIRO_API_EXPORT const xairo_uchar* getFacet(int) const;
		const xairo_uchar* getId() const;
	private:
		DOMElement* m_ptr;
		const xairo_uchar* m_elt;
		CXairoServer* m_server;

	};
	XAIRO_API_EXPORT const LemmaScheme* getLemmaScheme(int) const;


	class Attribute : public CXairoIGrammar::Attribute
	{
	public:
		Attribute(DOMElement* e,const xairo_uchar* ee,CXairoServer* s){m_ptr=e;m_elt=ee;m_server=s;}
		Attribute(const xairo_uchar* id,const xairo_uchar* e,CXairoServer* s);
		XAIRO_API_EXPORT const xairo_uchar* getName() const ;
		XAIRO_API_EXPORT const xairo_uchar* getGloss() const;
		XAIRO_API_EXPORT const Namespace* getNamespace() const;
		XAIRO_API_EXPORT const Codebook* getCodebook() const;
		XAIRO_API_EXPORT ip getIP() const;
		XAIRO_API_EXPORT const xairo_uchar* getTarget() const;
		XAIRO_API_EXPORT bool inBib() const;
		const xairo_uchar* getElt() const {return m_elt;}
		const xairo_uchar* getId() const;
	private:
		DOMElement* m_ptr;
		const xairo_uchar* m_elt;
		CXairoServer* m_server;
	};

	XAIRO_API_EXPORT int getElementCount() const;
	
	class Element : public CXairoIGrammar::Element
	{
	public:
		Element(DOMElement* e,CXairoServer* s){m_ptr=e;m_server=s;}
		Element(const xairo_uchar* id,CXairoServer* s);
		XAIRO_API_EXPORT const xairo_uchar* getName() const;
		XAIRO_API_EXPORT const xairo_uchar* getGloss() const;
		XAIRO_API_EXPORT int getAttributeCount() const;
		XAIRO_API_EXPORT const Attribute* getAttribute(int) const;
		XAIRO_API_EXPORT const Namespace* getNamespace() const;
		XAIRO_API_EXPORT bool inBib() const;
		const xairo_uchar* getId() const;
	private:
		DOMElement* m_ptr;
		CXairoServer* m_server;
	};
	XAIRO_API_EXPORT const Element* getElement(int) const;

	XAIRO_API_EXPORT int getGlobalAttributeCount() const;
	XAIRO_API_EXPORT const Attribute* getGlobalAttribute(int) const;
	
	class Partition : public CXairoIGrammar::Partition
	{
	public:
		Partition(DOMElement*,CXairoServer*);
		Partition(const xairo_uchar* id,CXairoServer* s);
		XAIRO_API_EXPORT const xairo_uchar* getName() const;
		XAIRO_API_EXPORT const xairo_uchar* getDescription() const;
		XAIRO_API_EXPORT int getCount() const;
		XAIRO_API_EXPORT const Partition* getDivision(int) const;
		XAIRO_API_EXPORT bool hasClasses() const;
		XAIRO_API_EXPORT int getWordCount() const;
		XAIRO_API_EXPORT int getIndex() const;
		int Classify(int) const;
		const Partition* GetParent(int&) const;
		bool Test(int,int) const;
		const xairo_uchar* getId() const;
	private:
		void Load();
		DOMElement* m_ptr;
		int* m_texts;
		CXairoServer* m_server;
		volatile bool m_bLoaded;
		bool m_bHasClasses;
	};
	XAIRO_API_EXPORT const Partition* getRootPartition() const;

	class Label : public CXairoIGrammar::Label
	{
	public:
		Label(DOMElement* e,CXairoServer* s);
		Label(const xairo_uchar* id,CXairoServer* s);  
		XAIRO_API_EXPORT const xairo_uchar* getElement() const;
		XAIRO_API_EXPORT const xairo_uchar* getAttribute()const ;
		const xairo_uchar* getId() const ;
	private:
		DOMElement* m_ptr;
		CXairoServer* m_server;
	};
	XAIRO_API_EXPORT int getLabelCount() const;
	XAIRO_API_EXPORT const Label* getLabel(int) const;

	XAIRO_API_EXPORT int getCodePageCount() const;
	XAIRO_API_EXPORT const xairo_uchar* getCodePage(int) const;

	XAIRO_API_EXPORT int getCharacterEntityCount() const;
	XAIRO_API_EXPORT void getCharacterEntity(int,const xairo_uchar**,int*) const; 


	XAIRO_API_EXPORT int getWTagCount() const;
	XAIRO_API_EXPORT const xairo_uchar* getWTag(int) const;

	XAIRO_API_EXPORT const char* getDSC() const;

	bool Init();
private:
	CXairoServer* m_server;
	Addkey** m_addkeys;
	int m_naddkeys;
	DOMLSParser* m_parser;
	DOMDocument *m_doc;
	void InitAddkeys();
public:
	const xairo_uchar* gId(DOMElement*) const ;
	DOMElement* gNode(const xairo_uchar*) const;
	xairo_uchar EvalCharEntity(const xairo_uchar*) const;
};

XAIRO_NAMESPACE_END
