#include "SServer.h"
#include "Locs.h"
#include "Parameters.h"
#include "DataReader.h"

XAIRO_NAMESPACE_BEGIN

CLocs::CLocs(CXairoServer* s):m_server(s)
{
}

CLocs::~CLocs()
{
	for (int i=0;i<m_nbuffer;i++)
		delete [] m_buffer[i].data;
	delete [] m_buffer;
	delete [] m_loctab;
	delete [] m_fLocs;
}

bool CLocs::Load()
{
	FILE *fL=_wfopen(m_server->GetParam()->exppath(PARAM_INDEXPATH,L("xdblocs")),L("rb"));
	if (fL==NULL) 
	{
		m_server->BootLog(L("Locs file 0 missing\n"));
		return false;
	}
	CDataReader* d=new CDataReader(0,fL);
	d->ReadLocHdr(1,&m_loc_hdr);
	m_fLocs=new FILE* [m_loc_hdr.nf];
	m_fLocs[0]=fL;
	for (int i=1;i<m_loc_hdr.nf;i++) {
		xairo_uchar p[_MAX_PATH];
		swprintf(p,L("xdblocs%d"),i);
		m_fLocs[i]=_wfopen(m_server->GetParam()->exppath(PARAM_INDEXPATH,p),L("rb"));
		if (m_fLocs[i]==NULL) {
			m_server->BootLog(L("Locs file %d missing\n"),i);
			return false;
		}
	}
	m_loctab=new xara_loc_page[m_loc_hdr.np];
	d->ReadLocPage(m_loc_hdr.np,m_loctab);
	m_maxiloc=m_loc_hdr.ps;
	m_npages=m_loc_hdr.np;
	InitBuffers(40);
	return true;
}

// the paging system: only public entry
xara_loc* CLocs::GetLoc(int i)
{
	int p=i / m_maxiloc;
	int k=i%m_maxiloc;
	xara_loc* x=ReadIndex(p);
	return x+k; // the C memorial code line
}

xara_loc* CLocs::ReadIndex(int i)
{
	xara_loc* d=Buffered(i);
	if (d) return d;
	d=GetFreeBuffer(i);
	if (d==NULL) 
		d=FreeBuffer(i);
	ReadPage(i,d);
	return d;
}


void CLocs::InitBuffers(int n)
{
	m_nbuffer=n;
	m_buffer=new _xb [n];
	for (int i=0;i<n;i++) {
		m_buffer[i].bUsed=false;
		m_buffer[i].data=new xara_loc [m_maxiloc];
	}
	m_nt=0;
}

xara_loc* CLocs::Buffered(int p)
{
	for (int i=0;i<m_nbuffer;i++) {
		if (m_buffer[i].bUsed && m_buffer[i].page==p) {
			m_buffer[i].t=m_nt++;
			return m_buffer[i].data;
		}
	}
	return NULL;
}

xara_loc* CLocs::GetFreeBuffer(int p)
{
	for (int i=0;i<m_nbuffer;i++) {
		if (!m_buffer[i].bUsed) {
			m_buffer[i].bUsed=true;
			m_buffer[i].page=p;
			m_buffer[i].t=m_nt++;
			return m_buffer[i].data;
		}
	}
	return NULL;
}

xara_loc* CLocs::FreeBuffer(int p)
{
	int mint=m_nt; int ii=-1;
	for (int i=0;i<m_nbuffer;i++) {
		if (m_buffer[i].t<mint) {
			ii=i;
			mint=m_buffer[i].t;
		}
	}
	m_buffer[ii].page=p;
	m_buffer[ii].t=m_nt++;
	return m_buffer[ii].data;
}

// use a fixed buffer here by putting max input page in header
void CLocs::ReadPage(int n,xara_loc* b)
{
	int off=m_loctab[n].off;
	int len=m_loctab[n].len;
	int nf=m_loctab[n].fi;
	fseek(m_fLocs[nf],off,SEEK_SET);
	CDataReader d(0,m_fLocs[nf]);
	d.ReadLoc(m_loc_hdr.ps,b,len);
}


int CLocs::GetMaxLoc()
{
//	return m_loc_hdr.maxpage; should use real figure 
	return 0x7fffffff;
}

XAIRO_NAMESPACE_END
