#include "SServer.h"

#include "Parameters.h"

XAIRO_NAMESPACE_BEGIN

#ifndef _WIN32
#define SEPS L("/")
#define SEPC '/'
#else
#define SEPS L("\\")
#define SEPC '\\'
#endif

xairo_uchar* s_={0};

CParameters::CParameters(CXairoServer* s):m_server(s)
{
	for (int i=0;i<PARAM_MAX;i++)
		param_vals[i]=s_;
}

CParameters::~CParameters(void)
{
}

class PHandle : public DOMErrorHandler {
public:
	PHandle():bErr(false) {}
	bool handleError(const DOMError &);	
	bool bErr; XMLFileLoc row, col;
	const xairo_uchar* msg;
};

bool PHandle::handleError(const DOMError& e)
{
	msg=e.getMessage();
	DOMLocator* l=e.getLocation();
	row=l->getLineNumber();
	col=l->getColumnNumber();
	bErr=true;
	return false;
}



bool CParameters::Read(const char* pfile) // make this wide like everything else
{

	const char* ss=strrchr(pfile,SEPC);
	if (ss==NULL) {
		m_server->BootLog(L("Cannot set root %s\n"),pfile);
		return false;
	}
	m_root=new xairo_uchar [ss-pfile+2];
	xairo_uchar* pw=(xairo_uchar*) m_root;
	mbstowcs(pw,pfile,ss-pfile+1);
	pw[ss-pfile+1]=0;
	DOMImplementationLS* di=(DOMImplementationLS*) DOMImplementation::getImplementation();
	DOMLSParser* d=di->createLSParser(DOMImplementationLS::MODE_SYNCHRONOUS,NULL);
//	d->setFeature(XMLUni::fgDOMNamespaces,true);
//    d->setFeature(XMLUni::fgXercesSchema,true);
//	d->setFeature(XMLUni::fgDOMValidation, true);
//	d->setProperty(XMLUni::fgXercesSchemaExternalNoNameSpaceSchemaLocation,L"c:\\bnc\\corpus_parameters.xsd");
	PHandle eh;
	DOMConfiguration* dcfg=d->getDomConfig();
	dcfg->setParameter(L("error-handler"),&eh);
	DOMDocument* doc=NULL;
	doc=d->parseURI(pfile);
	if (eh.bErr) {
		m_server->BootLog(L("Parser error in parameter file at %d,%d: %s\n"),eh.row,eh.col,eh.msg);		
		return false;
	}
	DOMElement* r=doc->getDocumentElement();
	if (wcscmp(r->getTagName(),L("corpus_parameters")) != 0) {
		m_server->BootLog(L("Parameter file has wrong document element: %s\n"),r->getTagName());		
		return false;
	}
	DOMNodeList* l=r->getChildNodes();
	for (unsigned i=0;i<l->getLength();i++) {
		DOMNode* n=l->item(i);
		if (n->getNodeType() != DOMNode::ELEMENT_NODE) continue;
		DOMElement* e=(DOMElement*) n;
		if (wcscmp(e->getTagName(),L("parameter")) != 0) continue;
		const xairo_uchar* name=e->getAttribute(L("name"));
		const xairo_uchar* value=e->getTextContent();
		SetParam(name,value);
	}
	return true;
}

const xairo_uchar* CParameters::param_names[PARAM_MAX]=
{L("etcpath"),L("name"),L("textpath"),L("rootpath"),L("indexpath"),L("headerpath"),L("temppath"),
L("log"),L("sourcepath"),L("bibpath"),L("validate"),L("grammarcache")};



void CParameters::SetParam(const xairo_uchar* name,const xairo_uchar* val)
{
	int i=0;
	for (i=0;i<PARAM_MAX;i++)
	{
		if (wcscmp(name,param_names[i])==0) break;
	}
	if (i>=PARAM_MAX) return;
	if (i==PARAM_ROOTPATH) return;
	param_vals[i]=_wcsdup(val);
}

const xairo_uchar* CParameters::exppath(param pa,const xairo_uchar* s)
{
	static xairo_uchar buffer[_MAX_PATH];
	wcscpy(buffer,m_root);
	if (pa != PARAM_ROOTPATH) {
		wcscat(buffer,Param(pa));
		wcscat(buffer,SEPS);
	}
	wcscat(buffer,s);
	return buffer;
}


XAIRO_NAMESPACE_END
