#pragma once

#include "stdafx.h"

#include "xarastruc0.h"


#include "KDB.h"
#include "Grammar.h"

typedef unsigned short xairo_uchar;

XAIRO_NAMESPACE_BEGIN

enum param {PARAM_ETCPATH,PARAM_NAME,PARAM_TEXTPATH,PARAM_ROOTPATH,PARAM_INDEXPATH,PARAM_HEADER,
	PARAM_TEMPPATH,PARAM_LOG,PARAM_SOURCEPATH,PARAM_BIB,PARAM_VALIDATE,PARAM_GRAMMARCACHE,PARAM_MAX};



class CSB;
class CWordList;
class CLocs;
class CThread;
class CDictionary;
class CXairoGrammar;
class CLanguage;
class CAttributeDictionary;
class CXairoSolution;
class CStreamFactory;
class CXairoSolveError;
class CStream;
class CEntityManager;
class CXairoCorpusTextList;
class CAtomStream;
class CParameters;
class CXairoIBootMonitor;
class CXairoIInterruptHandler;
class CAccelerators;

/*
struct CXairoIParameters
{
	const char* parameter_file;
	CXairoIBootMonitor* bm;
	CXairoIInterruptHandler* inth;
	const char* temppath;
	const char* datapath;
};


class CXairoInterface
{
public:
	virtual class CXairoIWordList* getWordList(const xairo_uchar*,const xairo_uchar*)=0;
//	virtual bool Init(const char*,CXairoBootMonitor*)=0;
	virtual class CXairoIWordList* getFrequencyList(const xairo_uchar*,int,int,int,const xairo_uchar*)=0;
	virtual class CXairoIAttributeValueList* getAttributeValueList(CXairoIGrammar::Attribute*)=0;
	virtual class CXairoISolution* solve(const xairo_uchar*)=0;
//	virtual class CXairoISolution* solveInDivision(const xairo_uchar*,CXairoGrammar::Partition*)=0;
	virtual class CXairoICorpusTextList* getCorpusTextList()=0;
	virtual const CXairoIGrammar* getGrammar()=0;
	virtual int getWordCount()=0;
	virtual const xairo_uchar* getCorpusName()=0;
	virtual const xairo_uchar* getDefaultLanguage()=0;
	virtual class CXairoIExtendedForm* getUsage(unsigned int)=0;
};

*/

struct _doc {
	int seq;
	const xairo_uchar* name;
};

class CXairoServer : public CXairoInterface
{
public:
	CXairoServer(void);
	virtual ~CXairoServer(void);
	XAIRO_API_EXPORT class CXairoIWordList* getWordList(int,const xairo_uchar**,const xairo_uchar*);
	XAIRO_API_EXPORT class CXairoIWordList* getFrequencyList(int,const xairo_uchar**,int,int,int,const xairo_uchar*);
	XAIRO_API_EXPORT class CXairoIWordList* getFilteredFrequencyList(int,const xairo_uchar**,int,int,int,const xairo_uchar*,const CXairoICorpusTextFilter* f,const CXairoIAddkeyFilter* ff);	
	XAIRO_API_EXPORT class CXairoIAttributeValueList* getAttributeValueList(const CXairoIGrammar::Attribute*);
	XAIRO_API_EXPORT class CXairoIAddkeyValueList* getAddkeyValueList(const CXairoIGrammar::Addkey*);
	XAIRO_API_EXPORT bool Init(const CXairoIParameters*);
	XAIRO_API_EXPORT static CXairoServer* getServer();
	XAIRO_API_EXPORT CXairoISolution* solve(const xairo_uchar*);
	XAIRO_API_EXPORT CXairoISolution* solveInDivision(const xairo_uchar*,const CXairoGrammar::Partition*);
	XAIRO_API_EXPORT CXairoICorpusTextList* getCorpusTextList();
	XAIRO_API_EXPORT const CXairoIGrammar* getGrammar() {return m_grammar;}
	XAIRO_API_EXPORT const xairo_uchar* getVersion();
	XAIRO_API_EXPORT int getWordCount();
	XAIRO_API_EXPORT const xairo_uchar* getCorpusName();
	XAIRO_API_EXPORT const xairo_uchar* getDefaultLanguage();
	XAIRO_API_EXPORT CXairoIExtendedForm* getUsage(unsigned int);
	XAIRO_API_EXPORT CXairoIAddkeyFilter* getAddkeyFilter(const xairo_uchar* key,int nvals,const xairo_uchar** vals);

	XAIRO_API_EXPORT void* deserialize(const xairo_uchar*);
	CKDB* GetDB() {return &m_kdb;}
	CSB* GetSB() {return m_sb;}
	CWordList* GetWL() {return m_wl;}
	CDictionary* GetDict(int i) {return m_dict[i+1];}
	CAttributeDictionary* GetAttDict() {return m_attdict;}
	const CXairoGrammar* GetGrammarImpl() {return m_grammar;}
	CLanguage* GetLanguage() {return m_language;}
	CLocs* GetLocs() {return m_locs;}
	CStreamFactory* GetSF() {return m_sf;}
	const char* GetTextPath(int);
	const xairo_uchar* GetEncoding(int);
	CEntityManager* GetEntityManager() {return m_em;}
	CAtomStream* GetDocumentStream() {return m_ds;}
	CAtomStream* GetDocumentEndStream() {return m_es;}
	CAtomStream* GetUnitStream() {return m_us;}
	CAtomStream* GetScopeStream() {return m_ss;}
	CAtomStream* GetLocaleStream() {return m_sl;}
	CParameters* GetParam() {return m_param;}
	CThread* GetThread() {return m_thread;}
	CAccelerators * GetAccelerators() {return m_accels;}
	void BootLog(const xairo_uchar*,...);
	static const xairo_uchar* MakeId(const xairo_uchar*,int);
	static const xairo_uchar* MakeId(const xairo_uchar*,int,int);
	static const xairo_uchar* MakeId(const xairo_uchar*,int,int,int);
	static int ScanId(const xairo_uchar*,const xairo_uchar*);
	static void ScanId(const xairo_uchar*,const xairo_uchar*,int*);
	static void ScanId(const xairo_uchar*,const xairo_uchar*,int*,int*);
	static void ScanId(const xairo_uchar*,const xairo_uchar*,int*,int*,int*);
	const char* GetTempPath() {return m_temppath;}
	const char* GetDataPath() {return m_datapath;}
	void TestInterrupt();
	int SeqToDoc(int);
	int getPid() {return 0;}
	void Debug(const char*,...);

private:
	void MakeDocTab();
	CXairoIBootMonitor* m_bm;
	CXairoIInterruptHandler* m_inth;
	CParameters* m_param;
	CAtomStream *m_ds, *m_us, *m_ss, *m_sl, *m_es;
	CDictionary** m_dict;
	const CXairoGrammar* m_grammar;
	CAttributeDictionary* m_attdict;
	CLanguage* m_language;
	CLocs *m_locs;
	CWordList* m_wl;
	int m_maxword;
	CKDB m_kdb;
	CSB* m_sb;
	CThread* m_thread;
	CStreamFactory* m_sf;
	CEntityManager* m_em;
	CAccelerators* m_accels;
	char m_temppath[_MAX_PATH];
	char m_datapath[_MAX_PATH];
	_doc* m_doctab;
	int m_nd;
	static int fc(const void*,const void*);
	static int ffc(const void*,const void*);

};

XAIRO_NAMESPACE_END
