#include "SServer.h"
#include "Stream.h"
#include "Locs.h"
#include "Thread.h"
#include "Accelerators.h"
#include "StreamFactory.h"

XAIRO_NAMESPACE_BEGIN

CStream::CStream(CXairoServer* s):m_server(s)
{
}

CStream::~CStream()
{
}

bool CStream::gtr(int i1,int i2,bool b)
{
	if (i1==-2) return true;
	if (i1==-1) return false;
	if (i2==-2) return false;
	if (i2==-1) return true;
	if (b) 
		return i1 > i2;
	else
		return i2 > i1;
}

// advance stream to or beyond target
bool CAtomStream::Advance(int target,bool bForwards,int doc)
{
	if (bForwards)
		return AdvanceForwards(target,doc);
	else
		return AdvanceBackwards(target,doc); // hmm
}


// add here a cache of recent positions 
bool CAtomStream::AdvanceForwards(int target,int doc)
{
	xara_loc* l=NULL;
	int i1=m_first;
	int i2=m_last;
	int ilt=m_last;
	if (m_cache_b && target>=m_cache_target && target <= m_cache_seq) {
		i1=m_cache_i;
		goto d;
	}
	if (m_cache_b && target==m_cache_seq+1) {
		i1=m_cache_i+1;
		goto d;
	}
	if (doc != -1 && m_accel != -1) {
		i1=m_server->GetAccelerators()->First(m_accel,doc);
		ilt=i2=m_server->GetAccelerators()->Last(m_accel,doc);
	}
    while (i2 > i1) {
		int i=(i1+i2)/2;
		l=m_server->GetLocs()->GetLoc(i);
		if (l->seq==target) {
			i1=i;
			break;
		} else if (l->seq < target) i1=i+1;
		else i2=i;
	}
d:
	if (i1<ilt) {
		l=m_server->GetLocs()->GetLoc(i1);
		SetPos(i1);
		m_cache_i=i1;
		m_cache_seq=l->seq;
		m_cache_b=true;
		m_cache_target=target;
		if (doc != -1 && doc != m_server->SeqToDoc(l->seq)) return false;
	}
	return i1<ilt;
}

void CAtomStream::SetPos(int i)
{
	xara_loc* l=SetCursor(i-m_first);
     m_fi = l->fi;
	 m_off=l->off;
	 m_nw=l->nw;
	 m_seq=l->seq;
	 m_nseq=1;
	 m_len=l->len;
	 m_wc=(m_wtype==it_content||m_wtype==it_addkey)?1:0;
}

void CKeyStream::SetPos(xara_loc* l)
{
     m_fi = l->fi;
	 m_off=l->off;
	 m_nw=l->nw;
	 m_seq=l->seq;
	 m_nseq=1;
	 m_len=l->len;
	 m_wc=1;
}

void CMemStream::SetPos(xara_loc* l)
{
     m_fi = l->fi;
	 m_off=l->off;
	 m_nw=l->nw;
	 m_seq=l->seq;
	 m_nseq=1;
	 m_len=l->len;
	 m_wc=1;
}

void CMulStream::CopyPos(int i)
{
     m_fi=m_op[i]->m_fi;
	 m_off=m_op[i]->m_off;
	 m_nw=m_op[i]->m_nw;
	 m_seq=m_op[i]->m_seq;
	 m_nseq=m_op[i]->m_nseq;
	 m_len=m_op[i]->m_len;
	 m_wc=m_op[i]->m_wc;
}

bool CAtomStream::AdvanceBackwards(int target,int doc)
{
	xara_loc* l=NULL;
	int i1=m_first-1;
	int i2=m_last-1;
	int ift=m_first-1;
	if (m_cache_b && target<=m_cache_target && target >= m_cache_seq) {
		i2=m_cache_i;
		goto d;
	}
	if (m_cache_b && target==m_cache_seq-1) {
		i2=m_cache_i-1;
		goto d;
	}
	if (doc != -1 && m_accel != -1) {
		ift=i1=m_server->GetAccelerators()->First(m_accel,doc)-1;
		i2=m_server->GetAccelerators()->Last(m_accel,doc)-1;
	}
    while (i2 > i1) {
		int i=(i1+i2+1)/2;
		l=m_server->GetLocs()->GetLoc(i);
		if (l->seq==target) {
			i2=i;
			break;
		} else if (l->seq > target) i2=i-1;
		else i1=i;
	}
d:	if (i2>ift) { // was ge
		l=m_server->GetLocs()->GetLoc(i2);
		SetPos(i2);
		m_cache_i=i2;
		m_cache_seq=l->seq;
		m_cache_b=true;
		m_cache_target=target;
		if (doc != -1 && doc != m_server->SeqToDoc(l->seq)) return false;
	}
	return i2>ift;
}

bool COrStream::Advance(int target,bool bForwards,int doc)
{
	int best=-2; int ii=-1;
	for (int i=0;i<m_ops;i++) {
		if (m_op[i]==NULL) continue; // should be a check they are not all NULL
		if (!m_op[i]->Advance(target,bForwards,doc)) continue;
		if (gtr(best,m_op[i]->m_seq,bForwards)) {ii=i; best=m_op[i]->m_seq;}
	}
	if (ii<0) return false;
	CopyPos(ii);
	return true;
}

// This could be made a special case of a prod stream
bool CSeqStream::Advance(int target,bool bForwards,int doc)
{
x: // make this a loop
	int best=-1; int ii=-1; int i;
	for (i=0;i<m_ops;i++) {
		if (!m_op[i]->Advance(target,bForwards,doc)) return false;
		if (gtr(m_op[i]->m_seq,best,bForwards)) {
			best=m_op[i]->m_seq;
			ii=i; // must be strict
		}
	}
	if (ii < 0) return false; // can't happen
	int ip=m_op[ii]->m_nw+m_op[ii]->m_wc;
	int t=m_op[ii]->m_seq+1;
	for (i=ii+1;i<m_ops;i++) {
x2a:
		if (!m_op[i]->Advance(t,true,doc)) break;
		if (m_op[i]->m_nw != -1 && m_op[i]->m_nw<ip) 
			{t=m_op[i]->m_seq+1; goto x2a;}
		if (m_op[i]->m_nw != -1 && m_op[i]->m_nw!=ip) {
			break;
		}
		ip+=m_op[i]->m_wc;
		t=m_op[i]->m_seq+1;
	}
	if (i<m_ops || (!bForwards && m_op[m_ops-1]->m_seq > target)) {
		target=m_op[ii]->m_seq;
		if (bForwards) target++;  else target--;
		goto x;
	}
	ip=m_op[ii]->m_nw;
	t=m_op[ii]->m_seq-1;
	for (i=ii-1;i>= 0;i--) {
x2:
		if (!m_op[i]->Advance(t,false,doc)) break;
		if (m_op[i]->m_nw != -1 && m_op[i]->m_nw+m_op[i]->m_wc>ip) 
			{t=m_op[i]->m_seq-1; goto x2;}
		if (m_op[i]->m_nw != -1 && m_op[i]->m_nw+m_op[i]->m_wc!=ip) {
			break;
		}
		ip-=m_op[i]->m_wc;
		t=m_op[i]->m_seq-1;
	}
	if (i >= 0 || (bForwards && m_op[0]->m_seq < target)) { 
		target=m_op[ii]->m_seq;
		if (bForwards) target++; else target--;
		goto x;
	}
	m_fi=m_op[0]->m_fi;
	m_off=m_op[0]->m_off;
	m_nw=m_op[0]->m_nw;
	m_wc=m_op[m_ops-1]->m_nw+m_op[m_ops-1]->m_wc-m_nw;
	m_seq=m_op[0]->m_seq;
	m_len=m_op[m_ops-1]->m_off+m_op[m_ops-1]->m_len-m_op[0]->m_off;
	m_nseq=m_op[m_ops-1]->m_seq+m_op[m_ops-1]->m_nseq-m_op[0]->m_seq;
	return true;
}

int CAndStream::GetParent()
{
	return m_op[0]->GetParent();
}

bool CAndStream::Advance(int target,bool bForwards,int doc)
{
x: // make this a loop
	int best=-1; int ii=-1;
	for (int i=0;i<m_ops;i++) {
		if (m_op[i]==NULL || !m_op[i]->Advance(target,bForwards,doc)) return false;
		if (gtr(m_op[i]->m_seq,best,bForwards)) {
			best=m_op[i]->m_seq;
			ii=i; // must be strict
		}
	}
	int t=m_op[ii]->m_seq;
	int pos=m_op[ii]->m_off;
	int fi=m_op[ii]->m_fi;
	for (int i=0;i<m_ops;i++) {
		if (i==ii) continue; 
		if (m_op[i]->Advance(t,true,doc)) {
			if (m_op[i]->m_off==pos && m_op[i]->m_fi==fi) {
				if (bForwards || m_op[i]->m_seq <= target)
					continue;
			}
		}
		if (m_op[i]->Advance(t,false,doc)) {
			if (m_op[i]->m_off==pos && m_op[i]->m_fi==fi) {
				if (!bForwards || m_op[i]->m_seq >= target)
					continue;
			}
		}
		target=t;
		if (bForwards) target++; else target--;
		goto x;
	}
	m_fi=m_op[0]->m_fi;
	m_off=m_op[0]->m_off;
	m_nw=m_op[0]->m_nw;
	m_wc=m_op[0]->m_wc;
	m_seq=m_op[0]->m_seq;
	m_nseq=1; // no, order may be wrong - see bprod
	m_len=m_op[0]->m_len;
	return true;
}

bool CKeyStream::Advance(int target,bool bForwards,int doc)
{
	if (bForwards)
		return AdvanceForwards(target);
	else
		return AdvanceBackwards(target); // hmm	return false;
}

bool CKeyStream::AdvanceForwards(int target)
{
	int i1=0;
	int n=m_server->GetDB()->countrecords(m_key); 
	int i2=n;
	xara_loc l;
    while (i2 > i1) {
		int i=(i1+i2)/2;
		m_server->GetDB()->recorded(m_key,i,&l);
		if (l.seq==target) {
			i1=i;
			break;
		} else if (l.seq < target) i1=i+1;
		else i2=i;
	}
	if (i1<n) {
		m_server->GetDB()->recorded(m_key,i1,&l);
		SetPos(&l);
	}
	return i1<n;
}

bool CKeyStream::AdvanceBackwards(int target)
{
	int i1=-1;
	int n=m_server->GetDB()->countrecords(m_key); 
	int i2=n-1;
	xara_loc l;
    while (i2 > i1) {
		int i=(i1+i2+1)/2;
		m_server->GetDB()->recorded(m_key,i,&l);
		if (l.seq==target) {
			i2=i;
			break;
		} else if (l.seq > target) i2=i-1;
		else i1=i;
	}
	if (i2>=0) {
		m_server->GetDB()->recorded(m_key,i2,&l);
		SetPos(&l);
	}
	return i2>=0;
}
bool CMemStream::Advance(int target,bool bForwards,int doc)
{
	if (bForwards)
		return AdvanceForwards(target);
	else
		return AdvanceBackwards(target); // hmm	return false;
}

bool CMemStream::AdvanceForwards(int target)
{
	int i1=0;
	int n=m_count; 
	int i2=n;
	xara_loc *l;
    while (i2 > i1) {
		int i=(i1+i2)/2;
		l=&m_buffer[i];
		if (l->seq==target) {
			i1=i;
			break;
		} else if (l->seq < target) i1=i+1;
		else i2=i;
	}
	if (i1<n) {
		l=&m_buffer[i1];
		SetPos(l);
	}
	return i1<n;
}

bool CMemStream::AdvanceBackwards(int target)
{
	int i1=-1;
	int n=m_count; 
	int i2=n-1;
	xara_loc* l;
    while (i2 > i1) {
		int i=(i1+i2+1)/2;
		l=&m_buffer[i];
		if (l->seq==target) {
			i2=i;
			break;
		} else if (l->seq > target) i2=i-1;
		else i1=i;
	}
	if (i2>=0) {
		l=&m_buffer[i2];
		SetPos(l);
	}
	return i2>=0;
}


bool CAllStream::Advance(int target,bool bForwards,int doc)
{
	m_seq=target;
	m_wc=1;
	m_nw=-1;
	return true;
}

bool CNegStream::Advance(int target,bool bForwards,int doc)
{
	int w=0; // this is broken
	while (m_op->Advance(target,bForwards,doc) && m_op->m_nw==w) {
		target=m_op->m_seq+1;
		w++;
	};
	 m_nw=w;
	 m_seq=target;
	 m_wc=1;
 	return true;
}

bool CLimitStream::Advance(int target,bool bForwards,int doc)
{
x:
	if (!m_op->Advance(target,bForwards,doc)) return false;
	bool b=true;
	switch (m_limtype) {
		case LT_SGML:
			if (!m_limit1->Advance(m_op->m_seq-1,false,doc)) b=false;
			if (b) {
				int p=m_limit1->GetParent();
				int x=m_limit1->m_seq;
				bool b1=false;
				while (m_limit2->Advance(x,true,doc)) {
					if (m_limit2->GetParent()==p) {b1=true; break;}
					x=m_limit2->m_seq+1;
				}
				b &= b1;
			}
//			if (b && !m_limit2->Advance(m_limit1->m_seq,true,doc)) b=false;
// this is wrong - count starts and ends as in xpath
			if (b && m_limit2->m_seq < m_op->m_seq+m_op->m_nseq) b=false;
			break;
		case LT_XSGML:
			if (!m_limit1->Advance(m_op->m_seq,false,doc)) b=false;
			if (b && !m_limit2->Advance(m_limit1->m_seq,true,doc)) b=false;
			if (b && m_limit2->m_seq <= m_op->m_seq+m_op->m_nseq) b=false;
			break;
		case LT_INT:
			b=m_op->m_wc<=m_nlimit;
			break;
	}
	if (!b) {
		target=m_op->m_seq;
		if (bForwards) target++;
		else target--;
		goto x;
	}
     m_fi=m_op->m_fi;
	 m_off=m_op->m_off;
	 m_nw=m_op->m_nw;
	 m_seq=m_op->m_seq;
	 m_nseq=m_op->m_nseq;
	 m_len=m_op->m_len;
	 m_wc=m_op->m_wc;
	 return true;
}

bool CProdStream::Advance(int target,bool bForwards,int doc)
{
x: // make this a loop
	int best=-1; int ii=-1; int i;
	for (i=0;i<m_ops;i++) {
		if (!m_op[i]->Advance(target,bForwards,doc)) return false;
		if (gtr(m_op[i]->m_seq,best,bForwards)) {
			best=m_op[i]->m_seq;
			ii=i; // must be strict
		}
	}
	if (ii < 0) return false; // can't happen
	int ip=m_op[ii]->m_fi;
	int t=m_op[ii]->m_seq+m_op[ii]->m_nseq+1;
	for (i=ii+1;i<m_ops;i++) {
		if (!m_op[i]->Advance(t,true,doc)) break;
		if (m_op[i]->m_fi != ip) {
			break;
		}
		t=m_op[i]->m_seq+m_op[i]->m_nseq+1;
	}
	if (i<m_ops || (!bForwards && m_op[m_ops-1]->m_seq > target)) {
		target=m_op[ii]->m_seq;
		if (bForwards) target++;  else target--;
		goto x;
	}
	t=m_op[ii]->m_seq-1;
	for (i=ii-1;i>= 0;i--) {
		if (!m_op[i]->Advance(t,false,doc)) break;
		if (m_op[i]->m_fi != ip) {
			break;
		}
		t=m_op[i]->m_seq-1;
	}
	if (i >= 0 || (bForwards && m_op[0]->m_seq < target)) { 
		target=m_op[ii]->m_seq;
		if (bForwards) target++; else target--;
		goto x;
	}
	m_fi=m_op[0]->m_fi;
	m_off=m_op[0]->m_off;
	m_nw=m_op[0]->m_nw;
	m_wc=m_op[m_ops-1]->m_nw+m_op[m_ops-1]->m_wc-m_op[0]->m_nw;
	m_seq=m_op[0]->m_seq;
	m_len=m_op[m_ops-1]->m_off+m_op[m_ops-1]->m_len-m_op[0]->m_off;
	m_nseq=m_op[m_ops-1]->m_seq+m_op[m_ops-1]->m_nseq-m_op[0]->m_seq;
	return true;
}

bool CBProdStream::Advance(int target,bool bForwards,int doc)
{
x: // make this a loop
	int best=-1; int ii=m_ops-1; int i;
	if (!m_op[ii]->Advance(target,bForwards,doc)) return false;
	int ip=m_op[ii]->m_fi;
	int t=m_op[ii]->m_seq+1;
	int doc1=m_server->SeqToDoc(t);
	if (doc != -1 && doc1 != doc) return false;
	for (i=ii+1;i<m_ops;i++) {
		if (!m_op[i]->Advance(t,false,doc1)) break;
		if (m_op[i]->m_fi != ip) {
			break;
		}
		t=m_op[i]->m_seq+1;
	}
	if (i<m_ops || (!bForwards && m_op[m_ops-1]->m_seq > target)) {
		target=m_op[ii]->m_seq;
		if (bForwards) target++;  else target--;
		goto x;
	}
	t=m_op[ii]->m_seq-1;
	for (i=ii-1;i>= 0;i--) {
		if (!m_op[i]->Advance(t,false,doc1)) break;
		if (m_op[i]->m_fi != ip) {
			break;
		}
//		t=m_op[i]->m_seq-1;
	}
	if (i >= 0) { //  || (bForwards && m_op[0]->m_seq < target)) { 
		target=m_op[ii]->m_seq;
		if (bForwards) target++; else target--;
		goto x;
	}
	m_fi=m_op[ii]->m_fi;
	m_off=m_op[ii]->m_off;
	m_nw=m_op[ii]->m_nw;
	m_wc=m_op[ii]->m_wc;
	m_seq=m_op[ii]->m_seq;
	m_len=m_op[ii]->m_len;
	m_nseq=m_op[ii]->m_nseq;
	return true;
}

bool CIdStream::Advance(int target,bool bForwards,int doc)
{
	if (GetVar(m_kvar)->bSet && GetVar(m_kvar)->setter != this) {
//		do {
//			if (!m_op->Advance(target,bForwards,doc))
//				return false;
//			if (m_op->m_nw == GetVar(m_kvar)->val) break;
//			target=bForwards?m_op->m_seq+1:m_op->m_seq-1;
//			} while (true); 
			CStream* s=m_server->GetSF()->MakeAttributeIdStream(m_elt,m_url,m_att,m_aurl,GetVar(m_kvar)->val);
			if (s==NULL) return false;
			bool b=s->Advance(target,bForwards,doc);
			if (!b) {delete s; return false;}
	m_fi=s->m_fi;
	m_off=s->m_off;
	m_nw=s->m_nw;
	m_wc=s->m_wc;
	m_seq=s->m_seq;
	m_len=s->m_len;
	m_nseq=s->m_nseq;
			delete s;
			return true;
      } else {
			if (!m_op->Advance(target,bForwards,doc))
				return false;
			GetVar(m_kvar)->bSet = true;
			GetVar(m_kvar)->setter = this;
			GetVar(m_kvar)->val = m_op->m_nw;
	//		GetVar(m_kvar)->stream=;
      }
	m_fi=m_op->m_fi;
	m_off=m_op->m_off;
	m_nw=m_op->m_nw;
	m_wc=m_op->m_wc;
	m_seq=m_op->m_seq;
	m_len=m_op->m_len;
	m_nseq=m_op->m_nseq;
  
		return true;
}

bool CIDCStream::Advance(int target,bool bForwards,int doc)
{
	do {
			if (!m_op->Advance(target,bForwards,doc))
				return false;
			if (m_op->m_nw == m_kval) break;
			target=bForwards?m_op->m_seq+1:m_op->m_seq-1;
			} while (true); 
	m_fi=m_op->m_fi;
	m_off=m_op->m_off;
	m_nw=m_op->m_nw;
	m_wc=m_op->m_wc;
	m_seq=m_op->m_seq;
	m_len=m_op->m_len;
	m_nseq=m_op->m_nseq;
	return true;

}

// These aren't needed now. Indeed, seq and and streams can be seen as special limits on 
// prod and bprod respectively.

bool CProdStream::LimitStream(lim_type tp,CStream* s1,CStream* s2,int tn)
{
	m_limit1 = s1;
	m_limit2 = s2;
	m_limtype = tp;
	m_limint = tn;
	return false;
}


bool CBProdStream::LimitStream(lim_type tp,CStream* s1,CStream* s2,int tn)
{
	m_limit1 = s1;
	m_limit2 = s2;
	m_limtype = tp;
	m_limint = tn;
	return false;
}

xara_loc* CAtomStream::SetCursor(int i)
{
	m_c=i;
	if (m_first+i < m_last) 
		return m_server->GetLocs()->GetLoc(m_first+i);
	else 
		return NULL;
}

int CAtomStream::GetCursor()
{
	return m_c;
}

bool CSeqStream::Flatten() {
	if (m_op[0]->IsWildcard()) return false;
	if (m_op[m_ops-1]->IsWildcard()) return false;
	return true;
}

int CAtomStream::Count(const CXairoGrammar::Partition* p)
{
	if (p==NULL) return Count();
	int ip=0;
	const CXairoGrammar::Partition* pp=p->GetParent(ip);
	CAtomStream* ds=m_server->GetDocumentStream();
	int pos=0; int k=0; int id=-1;
	int sd=0;
	while (ds->Advance(sd,true)) {
		if (!Advance(ds->m_seq,true)) break;
		int ppos=GetCursor();
		if (pp->Test(id,ip)) 
			k+=(ppos-pos);
		id++;
		pos=ppos;
		sd=ds->m_seq+1;
	}
	if (pp->Test(id,ip)) k+=(Count()-pos);
	return k;
}



int CAtomStream::Count(int np,int* p)
{
	int ip=0;
	CAtomStream* ds=m_server->GetDocumentStream();
	int pos=0; int k=0; int id=-1;
	int sd=0;
	while (ds->Advance(sd,true)) {
		if (!Advance(ds->m_seq,true)) break;
		int ppos=GetCursor();
		if (id >= 0 && id < np && p[id]) 
			k+=(ppos-pos);
		id++;
		pos=ppos;
		sd=ds->m_seq+1;
	}
	if (id >= 0 && id < np && p[id]) k+=(Count()-pos);
	return k;
}

int CAtomStream::CountMap(int np,int* p)
{
	int ip=0;
	int pos=0; int k=0; 
	int sd=0; bool b=false;
	for (int i=0;i<np;i++) {
		if (!Advance(p[i],true)) break;
		int ppos=GetCursor();
		if (b) 
			k+=(ppos-pos);
		pos=ppos;
		b=!b;
	}
	if (b) k+=(Count()-pos);
	return k;
}


int CAtomStream::GetParent()
{
	xara_loc* l=SetCursor();
	return l->parent;
}

void CStream::InitVars()
{
	for (int i=0;i<NVARS;i++)
		GetVar(i)->bSet=false;
}

xara_loc* CStream::GetLocBySeq(int seq)
{
	if (seq < 0) return 0;
	int i=m_server->GetThread()->GetThread(seq);
	return m_server->GetLocs()->GetLoc(i);
}


XAIRO_NAMESPACE_END
