#include "SServer.h"
#include "XairoForm.h"
#include "Dictionary.h"
#include "WordList.h"
#include "SB.h"
#include "StreamFactory.h"
#include "Stream.h"
#include "XairoCorpusTextFilter.h"
#include "XairoAddkeyFilter.h"

XAIRO_NAMESPACE_BEGIN


CXairoForm::CXairoForm(int d,int p,int i,CXairoServer* s):m_d(d),m_p(p),m_i(i),m_server(s)
{
	CDictionary* di=m_server->GetDict(m_d);
	xara_lemma* l=di->LookupEntry(m_p);
	m_form=&l->pt[m_i];
}

CXairoForm::~CXairoForm()
{

}

const xairo_uchar* CXairoForm::getAdditionalKey(const xairo_uchar* n)
{
	int i=m_server->GetWL()->addkey_index(n);
	if (i < 0) return L("");
	xara_streambase *sb=m_server->GetSB()->GetSB(m_form->ptr);
	return m_server->GetWL()->GetWordW(sb->w.con[i]);
}

const xairo_uchar* CXairoForm::getNorm()
{
	xara_streambase *sb=m_server->GetSB()->GetSB(m_form->ptr);
	return m_server->GetWL()->GetWordW(sb->w.val);
}

const xairo_uchar* CXairoForm::getText()
{
	xara_streambase *sb=m_server->GetSB()->GetSB(m_form->ptr);
	return m_server->GetWL()->GetWordW(sb->w.xval);
}



#define FF L("f(%d,%d,%d)")

const xairo_uchar* CXairoForm::getId()
{
	return CXairoServer::MakeId(FF,m_d,m_p,m_i);
}

CXairoForm::CXairoForm(const xairo_uchar* id,CXairoServer* s):m_server(s)
{
	CXairoServer::ScanId(FF,id,&m_d,&m_p,&m_i);
	CDictionary* di=m_server->GetDict(m_d);
	xara_lemma* l=di->LookupEntry(m_p);
	m_form=&l->pt[m_i];
}

int CXairoForm::getFilteredFrequency(const CXairoICorpusTextFilter* f)
{
	if (f==NULL) return m_form->freq;
	int * sc;
	int nsc=((CXairoCorpusTextFilter*) f) -> getTexts(sc);
	int k=0;
	for (int i=0;i<m_form->count;i++) {
		CAtomStream* s=m_server->GetSF()->Make1Stream(m_form->ptr+i);
		if (s==NULL) return 0;
		k+=s->Count(nsc,sc);
	}
	delete [] sc;
	return k;
}

int CXairoForm::getFilteredFrequencyMap(int np,int *p)
{
	int k=0;
	for (int i=0;i<m_form->count;i++) {
		CAtomStream* s=m_server->GetSF()->Make1Stream(m_form->ptr+i);
		if (s==NULL) return 0;
		k+=s->CountMap(np,p);
	}
	return k;
}

bool CXairoForm::included(const CXairoIAddkeyFilter* f)
{
	CXairoAddkeyFilter* ff = (CXairoAddkeyFilter*) f;
	const xairo_uchar* v=getAdditionalKey(ff->getKey());
	for (int i=0;i<ff->getCount();i++) 
		if (wcscmp(v,ff->getVal(i))==0) return true;
	return false;
}



XAIRO_NAMESPACE_END
