#include "SServer.h"
#include "XairoStubServer.h"
#include "XairoWordList.h"
#include "XairoSolution.h"
#include "XairoHit.h"
#include "XairoCorpusText.h"
#include "XairoSolveError.h"
#include "XairoCollocate.h"
#include "XairoCorpusTextList.h"
#include "XairoCorpusTextFilter.h"
#include "XairoAttributeValueList.h"
#include "XairoAttributeValue.h"
#include "XairoAddkeyValueList.h"
#include "XairoAddkeyValue.h"
#include "XairoXpathValue.h"
#include "XairoTextEntity.h"
#include "XairoForm.h"
#include "XairoExtendedForm.h"
#include "XairoAddkeyFilter.h"
#include "Grammar.h"
#include "RPCUtils.h"


XAIRO_NAMESPACE_BEGIN


CXairoStubServer::CXairoStubServer(CXairoServer* s):m_server(s)
{
	XMLTransService::Codes res;
	m_utf8=XMLPlatformUtils::fgTransService->makeNewTranscoderFor("UTF-8",res,1024);
}

CXairoStubServer::~CXairoStubServer(void)
{
}

class EH : public HandlerBase
{
	void warning  (  const SAXParseException &    exc  ) {throw new CXairoError(CXairoIError::ERR_XMLRPC,exc.getMessage());}
	void error  (  const SAXParseException &    exc  ) {throw new CXairoError(CXairoIError::ERR_XMLRPC,exc.getMessage());}
	void fatalError  (  const SAXParseException &    exc  ) {throw new CXairoError(CXairoIError::ERR_XMLRPC,exc.getMessage());}
};

char* CXairoStubServer::Transport(const char* s)
{
	EH eh; 
	MemBufInputSource is((XMLByte*) s,strlen(s),L("xml-rpc")); 
	is.setEncoding(L("UTF-8"));
	m_parser.setErrorHandler(&eh);
	try {
		m_parser.parse(is);
	}
	catch (CXairoError* ex)
	{
		return (char*) Encode(RPCFault(ex->getCode(),ex->getMessage()));
	}
	DOMElement* e=m_parser.getDocument()->getDocumentElement();
	try {
		return (char*) Encode(RPCStub(e));
	}
	catch (CXairoError* err)
	{
		return (char*) Encode(RPCFault(err->getCode(),err->getMessage()));
	}
	catch (...)
	{
		return (char*) Encode(RPCFault(CXairoIError::ERR_UNKNOWN,L("Untrapped server exception")));
	}
}

const char* CXairoStubServer::Encode(const xairo_uchar* s)
{
	int bs=4*((int) wcslen(s)+1);
	char* x=new char[bs]; XMLSize_t  ce;
	int k=m_utf8->transcodeTo(s,(int) wcslen(s),(XMLByte*) x,bs,ce,XMLTranscoder::UnRep_Throw);
	x[k]=0;
	const char* y=_strdup(x);
	delete [] x;
	return y;
}


const xairo_uchar* rpcmethod[RM_MAX]=
{
	L("getWordList"),
	L("getFrequencyList"),
	L("getFilteredFrequencyList"),
	L("WordList_getCount"),
	L("WordList_getWord"),
	L("Word_getFrequency"),
	L("Word_getFilteredFrequency"),
	L("Word_getFormCount"),
	L("Word_getNorm"),
	L("Word_getForm"),
	L("Form_getFrequency"),
	L("Form_getFilteredFrequency"),
	L("Form_getAdditionalKey"),
	L("Form_getNorm"),
	L("Form_getText"),
	L("Form_included"),
	L("ExtendedForm_getAdditionalKey"),
	L("ExtendedForm_getNorm"),
	L("ExtendedForm_getText"),
	L("ExtendedForm_getType"),
	L("Grammar_getAddkeyCount"),
	L("Grammar_getAddkey"),
	L("Grammar_getNamespaceCount"),
	L("Grammar_getNamespace"),
	L("Grammar_getElementCount"),
	L("Grammar_getElement"),
	L("Grammar_getGlobalAttributeCount"),
	L("Grammar_getGlobalAttribute"),
	L("Grammar_getAttributeCount"),
	L("Grammar_getAttribute"),
	L("Grammar_getLabelCount"),
	L("Grammar_getLabel"),
	L("Grammar_getWTagCount"),
	L("Grammar_getWTag"),
	L("solve"),
	L("Solution_getCount"),
	L("Solution_getCorpusTextCount"),
	L("Solution_getHit"),
	L("Hit_getCorpusText"),
	L("Hit_getText"),
	L("CorpusText_getTitle"),
	L("CorpusText_classify"),
	L("Solution_thinInit"),
	L("Solution_thinRand"),
	L("Solution_thinOnePerText"),
	L("Solution_thinCollocation"),
	L("Solution_thinSelection"),
	L("Solution_thinCorpusTextFilter"),
	L("Solution_thinAddkeyFilter"),
	L("Solution_makeCollocationTable"),
	L("Solution_getAnalysis"),
	L("Solution_getAnalysisScope"),
	L("Solution_release"),
	L("CollocationTable_getCount"),
	L("CollocationTable_getCollocate"),
	L("getWordCount"),
	L("getCorpusName"),
	L("CorpusTextList_getCount"),
	L("CorpusTextList_getCorpusText"),
	L("getAttributeValueList"),
	L("AttributeValueList_getCount"),
	L("AttributeValueList_getAttributeValue"),
	L("AttributeValue_getFilteredFrequency"),
	L("Grammar_getRootPartition"),
	L("Grammar_Partition_getDivision"),
	L("Grammar_Partition_getWordCount"),
	L("Grammar_Partition_getIndex"),
	L("CorpusText_evalXpath"),
	L("CorpusText_getWordCount"),
	L("CorpusText_isSelected"),
	L("CorpusText_getTextEntity"),
	L("CorpusText_getIndex"),
	L("Grammar_Codebook_lookup"),
	L("Grammar_Codebook_getCount"),
	L("Grammar_Codebook_getPair"),
	L("Grammar_getCodePageCount"),
	L("Grammar_getCodePage"),
	L("Grammar_getCharacterEntityCount"),
	L("Grammar_getCharacterEntity"),
	L("Grammar_getDSC"),
	L("getDefaultLanguage"),
	L("CorpusTextList_getCorpusTextFilter"),
	L("Grammar_getLemmaScheme"),
	L("Grammar_getLemmaSchemeCount"),
	L("Grammar_LemmaScheme_getFacet"),
	L("Grammar_getRegion"),
	L("Grammar_getRegionCount"),
	L("Grammar_Region_getValue"),
	L("Solution_copy"),
	L("getAddkeyFilter"),
	L("getAddkeyValueList"),
	L("AddkeyValueList_getCount"),
	L("AddkeyValueList_getAddkeyValue"),
};

enum rm CXairoStubServer::LUM(const xairo_uchar* s)
{
	for (int i=0;i<RM_MAX;i++) 
		if (wcscmp(rpcmethod[i],s)==0)
			return (rm) i;
	return RM_MAX;
}

const xairo_uchar* CXairoStubServer::RPCStub(DOMElement* e)
{
	DOMNodeList* nl=e->getElementsByTagName(L("methodName"));
	int k=nl->getLength();
	if (k==0) return NULL;
	DOMNode* nd=nl->item(0);
	const xairo_uchar* s=nd->getTextContent();
	rm m=LUM(s);
	switch (m) {
	case RM_GETWORDLIST:
		{
			const xairo_uchar** re;
			int n=RPCVectorsParam(e,0,re);
			const xairo_uchar* ls=RPCStringParam(e,1);
			CXairoWordList* wl=(CXairoWordList*) m_server->getWordList(n,re,ls);
			if (wl)
				return RPCStringRet(wl->getId());
			else 
				return RPCStringRet(XMLRPC_NULL);
		}
	case RM_GETFREQUENCYLIST:
		{
			const xairo_uchar** re;
			int nr=RPCVectorsParam(e,0,re);
			int n=RPCIntParam(e,1);
			int m=RPCIntParam(e,2);
			int p=RPCIntParam(e,3);
			const xairo_uchar* ls=RPCStringParam(e,4);
			CXairoWordList* wl=(CXairoWordList*) m_server->getFrequencyList(nr,re,n,m,p,ls);
			if (wl)
				return RPCStringRet(wl->getId());
			else 
				return RPCStringRet(XMLRPC_NULL);
		}
	case RM_GETFILTEREDFREQUENCYLIST:
		{
			const xairo_uchar** re;
			int nr=RPCVectorsParam(e,0,re);
			int n=RPCIntParam(e,1);
			int m=RPCIntParam(e,2);
			int p=RPCIntParam(e,3);
			const xairo_uchar* ls=RPCStringParam(e,4);
			const xairo_uchar* fid=RPCStringParam(e,5);
			const xairo_uchar* ffid=RPCStringParam(e,6);
			bool b1=wcscmp(fid,L("null")) != 0;
			bool b2=wcscmp(ffid,L("null")) != 0;
			CXairoCorpusTextFilter f(fid,m_server);
			CXairoAddkeyFilter ff(ffid,m_server);
			CXairoWordList* wl=(CXairoWordList*) m_server->getFilteredFrequencyList(nr,re,n,m,p,ls,b1?&f:NULL,b2?&ff:NULL);
			if (wl)
				return RPCStringRet(wl->getId());
			else 
				return RPCStringRet(XMLRPC_NULL);
		}
	case RM_SOLVE:
		{
			const xairo_uchar* q=RPCStringParam(e,0);
			CXairoIError* e=NULL;
			CXairoSolution* sol=NULL;
			sol=(CXairoSolution*) m_server->solve(q);
			xmlrpc_arg x[1];
			x[0].t=XMLRPC_STRING;
			if (sol==NULL) x[0].s=XMLRPC_NULL;
			else x[0].s=sol->getId();
			return RPCRet(1,x);
		}

	case RM_WORDLIST_GETCOUNT:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			CXairoWordList w(id,m_server);
			return RPCIntRet(w.getCount());
		}
	case RM_WORD_GETFREQUENCY:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			CXairoWord w(id,m_server);
			return RPCIntRet(w.getFrequency());
		}
	case RM_WORD_GETFILTEREDFREQUENCY:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			const xairo_uchar* fid=RPCStringParam(e,1);
			const xairo_uchar* ffid=RPCStringParam(e,2);
			bool b1=wcscmp(fid,L("null")) != 0;
			bool b2=wcscmp(ffid,L("null")) != 0;
			CXairoCorpusTextFilter f(fid,m_server);
			CXairoAddkeyFilter ff(ffid,m_server);
			CXairoWord w(id,m_server);
			return RPCIntRet(w.getFilteredFrequency(b1?&f:NULL,b2?&ff:NULL));
		}
	case RM_WORD_GETFORMCOUNT:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			CXairoWord w(id,m_server);
			return RPCIntRet(w.getFormCount());
		}
	case RM_WORD_GETNORM:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			int i=RPCIntParam(e,1);
			CXairoWord w(id,m_server);
			return RPCStringRet(w.getNorm(i));
		}
	case RM_WORDLIST_GETWORD:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			CXairoWordList w(id,m_server);
			CXairoWord* ww=w.getWord(RPCIntParam(e,1));
			xmlrpc_arg x[4];
			x[0].t=XMLRPC_STRING; x[0].s=ww->getId(); 
			x[1].t=XMLRPC_I4; x[1].i=ww->getFrequency(); 
			x[2].t=XMLRPC_I4; x[2].i=ww->getFormCount(); 
			x[3].t=XMLRPC_STRING; x[3].s=ww->getNorm(0); 
			return RPCRet(4,x);
		}
	case RM_WORD_GETFORM:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			CXairoWord w(id,m_server);
			CXairoForm* ww=w.getForm(RPCIntParam(e,1));
			return RPCStringRet(ww->getId());
		}
	case RM_FORM_GETFREQUENCY:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			CXairoForm w(id,m_server);
			return RPCIntRet(w.getFrequency());
		}
	case RM_FORM_GETFILTEREDFREQUENCY:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			const xairo_uchar* fid=RPCStringParam(e,1);
			CXairoCorpusTextFilter f(fid,m_server);
			CXairoForm w(id,m_server);
			return RPCIntRet(w.getFilteredFrequency(&f));
		}
	case RM_FORM_GETADDITIONALKEY:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			const xairo_uchar* key=RPCStringParam(e,1);
			CXairoForm w(id,m_server);
			return RPCStringRet(w.getAdditionalKey(key));
		}
	case RM_FORM_GETNORM:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			CXairoForm w(id,m_server);
			return RPCStringRet(w.getNorm());
		}
	case RM_FORM_GETTEXT:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			CXairoForm w(id,m_server);
			return RPCStringRet(w.getText());
		}
	case RM_FORM_INCLUDED:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			const xairo_uchar* idk=RPCStringParam(e,1);
			CXairoForm w(id,m_server);
			CXairoAddkeyFilter f(idk,m_server);
			return RPCBoolRet(w.included(&f));
		}
	case RM_EXTENDEDFORM_GETADDITIONALKEY:
		{
			unsigned int i=RPCIntParam(e,0);
			const xairo_uchar* key=RPCStringParam(e,1);
			CXairoExtendedForm w(i,m_server);
			return RPCStringRet(w.getAdditionalKey(key));
		}
	case RM_EXTENDEDFORM_GETNORM:
		{
			unsigned int i=RPCIntParam(e,0);
			CXairoExtendedForm w(i,m_server);
			return RPCStringRet(w.getNorm());
		}
	case RM_EXTENDEDFORM_GETTEXT:
		{
			unsigned int i=RPCIntParam(e,0);
			CXairoExtendedForm w(i,m_server);
			return RPCStringRet(w.getText());
		}
	case RM_EXTENDEDFORM_GETTYPE:
		{
			unsigned int i=RPCIntParam(e,0);
			CXairoExtendedForm w(i,m_server);
			return RPCIntRet(w.getType());
		}
	case RM_GRAMMAR_GETADDKEYCOUNT:
		{
			return RPCIntRet(m_server->GetGrammarImpl()->getAddkeyCount());
		}
	case RM_GRAMMAR_GETADDKEY:
		{
			int i=RPCIntParam(e,0);
			const CXairoGrammar::Addkey* k=m_server->GetGrammarImpl()->getAddkey(i);
			xmlrpc_arg x[6];
			x[0].t=XMLRPC_STRING; x[0].s=k->getId(); 
			x[1].t=XMLRPC_STRING; x[1].s=k->getName(); 
			x[2].t=XMLRPC_STRING; x[2].s=k->getGloss(); 
			x[3].t=XMLRPC_BOOL; x[3].i=k->isLemmaScheme()?1:0; 
			const CXairoGrammar::Codebook* cb=k->getCodebook();
			x[4].t=XMLRPC_STRING; 
			if (cb) 
				x[4].s=cb->getId(); 
			else 
				x[4].s=XMLRPC_NULL;
			x[5].t=XMLRPC_STRING; x[5].s=k->getAttribute(); 
			return RPCRet(6,x);
		}
	case RM_GRAMMAR_GETNAMESPACECOUNT:
		{
			return RPCIntRet(m_server->GetGrammarImpl()->getNamespaceCount());
		}
	case RM_GRAMMAR_GETNAMESPACE:
		{
			int i=RPCIntParam(e,0);
			const CXairoIGrammar::Namespace* ns=m_server->GetGrammarImpl()->getNamespace(i);
			xmlrpc_arg x[1];
			x[0].t=XMLRPC_STRING; x[0].s=ns->getUrl(); 
			return RPCRet(1,x);
		}
	case RM_GRAMMAR_GETELEMENTCOUNT:
		{
			return RPCIntRet(m_server->GetGrammarImpl()->getElementCount());
		}
	case RM_GRAMMAR_GETELEMENT:
		{
			int i=RPCIntParam(e,0);
			const CXairoGrammar::Element* e=m_server->GetGrammarImpl()->getElement(i);
			xmlrpc_arg x[6];
			x[0].t=XMLRPC_STRING; x[0].s=e->getId(); 
			x[1].t=XMLRPC_STRING; x[1].s=e->getName(); 
			x[2].t=XMLRPC_STRING; x[2].s=e->getGloss(); 
			x[3].t=XMLRPC_I4; x[3].i=e->getAttributeCount(); 
			x[4].t=XMLRPC_STRING; x[4].s=e->getNamespace()->getUrl(); 
			x[5].t=XMLRPC_BOOL; x[5].i=e->inBib()?1:0;
			return RPCRet(6,x);
		}
	case RM_GRAMMAR_GETGLOBALATTRIBUTECOUNT:
		{
			return RPCIntRet(m_server->GetGrammarImpl()->getGlobalAttributeCount());
		}
	case RM_GRAMMAR_GETGLOBALATTRIBUTE:
		{
			int i=RPCIntParam(e,0);
			const CXairoGrammar::Attribute* a=m_server->GetGrammarImpl()->getGlobalAttribute(i);
			xmlrpc_arg x[8];
			x[0].t=XMLRPC_STRING; x[0].s=a->getId(); 
			x[1].t=XMLRPC_STRING; x[1].s=a->getName(); 
			x[2].t=XMLRPC_STRING; x[2].s=a->getGloss(); 
			x[3].t=XMLRPC_STRING; x[3].s=a->getNamespace()->getUrl(); 
			const CXairoGrammar::Codebook* cb=a->getCodebook();
			x[4].t=XMLRPC_STRING; 
			if (cb)
				x[4].s=cb->getId(); 
			else 
				x[4].s=XMLRPC_NULL;
			x[5].t=XMLRPC_I4; x[5].i=a->getIP();
			x[6].t=XMLRPC_STRING; x[6].s=a->getTarget();
			x[7].t=XMLRPC_BOOL; x[7].i=a->inBib()?1:0;
			return RPCRet(8,x);
		}
	case RM_GRAMMAR_GETATTRIBUTECOUNT:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			CXairoGrammar::Element elt(id,m_server);
			xmlrpc_arg x[1];
			x[0].t=XMLRPC_I4; x[1].i=elt.getAttributeCount();
			return RPCRet(1,x);
		}
	case RM_GRAMMAR_GETATTRIBUTE:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			int i=RPCIntParam(e,1);
			CXairoGrammar::Element elt(id,m_server);
			const CXairoGrammar::Attribute* a=elt.getAttribute(i);
			xmlrpc_arg x[8];
			x[0].t=XMLRPC_STRING; x[0].s=a->getId(); 
			x[1].t=XMLRPC_STRING; x[1].s=a->getName(); 
			x[2].t=XMLRPC_STRING; x[2].s=a->getGloss(); 
			x[3].t=XMLRPC_STRING; x[3].s=a->getNamespace()->getUrl(); 
			const CXairoGrammar::Codebook* cb=a->getCodebook();
			x[4].t=XMLRPC_STRING; 
			if (cb)
				x[4].s=cb->getId(); 
			else 
				x[4].s=XMLRPC_NULL;
			x[5].t=XMLRPC_I4; x[5].i=a->getIP();
			x[6].t=XMLRPC_STRING; x[6].s=a->getTarget();
			x[7].t=XMLRPC_BOOL; x[7].i=a->inBib()?1:0;
			return RPCRet(8,x);
		}
	case RM_GRAMMAR_GETLABELCOUNT:
		{
			return RPCIntRet(m_server->GetGrammarImpl()->getLabelCount());
		}
	case RM_GRAMMAR_GETLABEL:
		{
			int i=RPCIntParam(e,0);
			const CXairoIGrammar::Label* l=m_server->GetGrammarImpl()->getLabel(i);
			xmlrpc_arg x[2];
			x[0].t=XMLRPC_STRING; x[0].s=l->getElement(); 
			x[1].t=XMLRPC_STRING; x[1].s=l->getAttribute(); 
			return RPCRet(2,x);
		}
	case RM_GRAMMAR_GETWTAGCOUNT:
		{
			return RPCIntRet(m_server->GetGrammarImpl()->getWTagCount());
		}
	case RM_GRAMMAR_GETWTAG:
		{
			int i=RPCIntParam(e,0);
			const xairo_uchar* w=m_server->GetGrammarImpl()->getWTag(i);
			xmlrpc_arg x[1];
			x[0].t=XMLRPC_STRING; x[0].s=w; 
			return RPCRet(1,x);
		}
	case RM_SOLUTION_GETCOUNT:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			CXairoSolution w(id,m_server);
			return RPCIntRet(w.getCount());
		}
	case RM_SOLUTION_GETCORPUSTEXTCOUNT:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			CXairoSolution w(id,m_server);
			return RPCIntRet(w.getCorpusTextCount());
		}
	case RM_SOLUTION_GETHIT:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			int i=RPCIntParam(e,1);
			CXairoSolution w(id,m_server);
			CXairoHit* hit=w.getHit(i);
			xmlrpc_arg x[6];
			x[0].t=XMLRPC_STRING; x[0].s=hit->getUnit(); 
			x[1].t=XMLRPC_I4; x[1].i=hit->getDocNo(); 
			x[2].t=XMLRPC_I4; x[2].i=hit->getFileNo(); 
			x[3].t=XMLRPC_STRING; x[3].s=hit->getTextEntity()->getURL(); 
			x[4].t=XMLRPC_I4; x[4].i=hit->getCursorStart();
			x[5].t=XMLRPC_I4; x[5].i=hit->getCursorEnd();
			delete hit;
			return RPCRet(6,x);

		}
	case RM_CORPUSTEXTLIST_GETCORPUSTEXT:
	case RM_HIT_GETCORPUSTEXT:
		{
			int d=RPCIntParam(e,0);
			CXairoCorpusText* t=new CXairoCorpusText(d,m_server);
			return RPCStringRet(t->getId());
		}
	case RM_HIT_GETTEXT:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			int i=RPCIntParam(e,1);
			int s=RPCIntParam(e,2);
			int max=RPCIntParam(e,3);
			CXairoSolution w(id,m_server);
			CXairoHit* hit=w.getHit(i);
			const xairo_uchar* text=hit->getText(s,max);
			delete hit;
			return RPCStringRet(text);
		}
	case RM_CORPUSTEXT_GETTITLE:
		{
			int id=RPCIntParam(e,0);
			CXairoCorpusText w(id,m_server);
			return RPCStringRet(w.getTitle());
		}
	case RM_CORPUSTEXT_CLASSIFY:
		{
			int id=RPCIntParam(e,0);
			const xairo_uchar* pid=RPCStringParam(e,1);
			CXairoCorpusText w(id,m_server);
			CXairoGrammar::Partition p(pid,m_server);
			int n=w.classify(&p);
			return RPCIntRet(n);
		}
	case RM_SOLUTION_THININIT:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			int n=RPCIntParam(e,1);
			CXairoSolution w(id,m_server);
			CXairoSolution* sol=(CXairoSolution*) w.thinInit(n);
			if (sol==NULL) return RPCStringRet(L("null"));
			else return RPCStringRet(sol->getId());
		}
	case RM_SOLUTION_THINRAND:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			int n=RPCIntParam(e,1);
			int s=RPCIntParam(e,2);
			CXairoSolution w(id,m_server);
			CXairoSolution* sol=(CXairoSolution*) w.thinRand(n,s);
			if (sol==NULL) return RPCStringRet(L("null"));
			else return RPCStringRet(sol->getId());
		}
	case RM_SOLUTION_THINCOLLOCATION:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			const xairo_uchar **v=NULL;
			int n=RPCVectorsParam(e,1,v);
			const xairo_uchar* key=RPCStringParam(e,2);
			CXairoICollocationTable::_cwin win;
			win.bX=RPCBoolParam(e,3);
			win.wl=RPCIntParam(e,4);
			win.wr=RPCIntParam(e,5);
			win.s=RPCStringParam(e,6);
			win.ns=RPCStringParam(e,7);
			CXairoSolution w(id,m_server);
			CXairoSolution* sol=(CXairoSolution*) w.thinCollocation(key,win,n,v);
			if (sol==NULL) return RPCStringRet(L("null"));
			else return RPCStringRet(sol->getId());
		}
	case RM_SOLUTION_THINONEPERTEXT:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			CXairoSolution w(id,m_server);
			CXairoSolution* sol=(CXairoSolution*) w.thinOnePerText();
			if (sol==NULL) return RPCStringRet(XMLRPC_NULL);
			else return RPCStringRet(sol->getId());
		}
	case RM_SOLUTION_THINSELECTION:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			CXairoSolution w(id,m_server);
			int* v;
			int n=RPCVectorParam(e,1,v);
			CXairoSolution* sol=(CXairoSolution*) w.thinSelection(n,v);
			if (sol==NULL) return RPCStringRet(XMLRPC_NULL);
			else return RPCStringRet(sol->getId());
		}
	case RM_SOLUTION_COPY:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			CXairoSolution w(id,m_server);
			CXairoSolution* sol=(CXairoSolution*) w.copy();
			if (sol==NULL) return RPCStringRet(XMLRPC_NULL);
			else return RPCStringRet(sol->getId());
		}
	case RM_SOLUTION_THINCORPUSTEXTFILTER:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			CXairoSolution w(id,m_server);
			const xairo_uchar* fid=RPCStringParam(e,1);
			CXairoCorpusTextFilter f(fid,m_server);
			CXairoSolution* sol=(CXairoSolution*) w.thinCorpusTextFilter(&f);
			if (sol==NULL) return RPCStringRet(XMLRPC_NULL);
			else return RPCStringRet(sol->getId());
		}
	case RM_SOLUTION_THINADDKEYFILTER:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			CXairoSolution w(id,m_server);
			const xairo_uchar* fid=RPCStringParam(e,1);
			CXairoAddkeyFilter f(fid,m_server);
			CXairoSolution* sol=(CXairoSolution*) w.thinAddkeyFilter(&f);
			if (sol==NULL) return RPCStringRet(XMLRPC_NULL);
			else return RPCStringRet(sol->getId());
		}
	case RM_SOLUTION_GETCOLLOCATIONTABLE:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			CXairoSolution w(id,m_server);
			const xairo_uchar* pattern=RPCStringParam(e,1);
			const xairo_uchar* key=RPCStringParam(e,2);
			CXairoICollocationTable::cmeth m=(CXairoICollocationTable::cmeth) RPCIntParam(e,3);
			CXairoICollocationTable::_cwin win;
			win.bX=RPCBoolParam(e,4);
			win.wl=RPCIntParam(e,5);
			win.wr=RPCIntParam(e,6);
			win.s=RPCStringParam(e,7);
			win.ns=RPCStringParam(e,8);
			int maxn=RPCIntParam(e,9);
			double mins=RPCDoubleParam(e,10);
			double maxs=RPCDoubleParam(e,11);
			int lowf=RPCIntParam(e,12);
			CXairoCollocationTable* t=(CXairoCollocationTable*) w.makeCollocationTable(pattern,key,m,win,maxn,mins,maxs,lowf);
			if (t==NULL) return RPCStringRet(XMLRPC_NULL);
			else return RPCStringRet(t->getId());
		}
	case RM_SOLUTION_GETANALYSIS:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			CXairoSolution w(id,m_server);
			int *p=NULL;
			int np=RPCVectorParam(e,1,p);
			int n=RPCIntParam(e,2);
			int* v=new int [n]; int* vt=new int [n];
			w.getAnalysis(np,p,n,v,vt);
			xmlrpc_arg x[2];
			x[0].t=XMLRPC_VECTOR; x[0].nvec=n; x[0].vec=v;
			x[1].t=XMLRPC_VECTOR; x[1].nvec=n; x[1].vec=vt;
			return RPCRet(2,x);
		}
	case RM_SOLUTION_GETANALYSISSCOPE:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			CXairoSolution w(id,m_server);
			int *p=NULL;
			int np=RPCVectorParam(e,1,p);
			int n=RPCIntParam(e,2);
			int* v=new int [n]; int* vt=new int [n];
			const xairo_uchar* elt=RPCStringParam(e,3);
			w.getAnalysisScope(np,p,n,v,vt,elt);
			xmlrpc_arg x[2];
			x[0].t=XMLRPC_VECTOR; x[0].nvec=n; x[0].vec=v;
			x[1].t=XMLRPC_VECTOR; x[1].nvec=n; x[1].vec=vt;
			return RPCRet(2,x);
		}
	case RM_SOLUTION_RELEASE:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			CXairoSolution w(id,m_server);
			w.release();
			return RPCRet(0,NULL);
		}
	case RM_COLLOCATIONTABLE_GETCOUNT:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			CXairoCollocationTable w(id,m_server);
			return RPCIntRet(w.getCount());
		}
	case RM_COLLOCATIONTABLE_GETCOLLOCATE:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			CXairoCollocationTable w(id,m_server);
			int i=RPCIntParam(e,1);
			CXairoICollocate* ct=w.getCollocate(i);
			xmlrpc_arg x[3];
			x[0].t=XMLRPC_STRING; x[0].s=((CXairoWord*) ct->getWord())->getId(); 
			x[1].t=XMLRPC_I4; x[1].i=ct->getHits(); 
			x[2].t=XMLRPC_DOUBLE; x[2].d=ct->getScore(); 
			return RPCRet(3,x);
		}
	
	case RM_GETWORDCOUNT:
		{
			int k=m_server->getWordCount();
			xmlrpc_arg x[1];
			x[0].t=XMLRPC_I4; x[0].i=k;
			return RPCRet(1,x);
		}
	case RM_GETCORPUSNAME:
		{
			const xairo_uchar* n=m_server->getCorpusName();
			xmlrpc_arg x[1];
			x[0].t=XMLRPC_STRING; x[0].s=n;
			return RPCRet(1,x);
		}
	case RM_CORPUSTEXTLIST_GETCOUNT:
		{
			xmlrpc_arg x[1];
			x[0].t=XMLRPC_I4; x[0].i=m_server->getCorpusTextList()->getCount();
			return RPCRet(1,x);
		}
	case RM_CORPUSTEXTLIST_GETCORPUSTEXTFILTER:
		{
			int *p=NULL;
			int np=RPCVectorParam(e,0,p);
			CXairoICorpusTextList* tl=m_server->getCorpusTextList();
			CXairoCorpusTextFilter* f=(CXairoCorpusTextFilter*) tl->getCorpusTextFilter(np,p);
			delete tl;
			xmlrpc_arg x[1];
			x[0].t=XMLRPC_STRING;
			if (f==NULL) x[0].s=XMLRPC_NULL;
			else x[0].s=f->getId();
			return RPCRet(1,x);
		}
	case RM_GETATTRIBUTEVALUELIST:
		{
			const xairo_uchar* att=RPCStringParam(e,0);
			const xairo_uchar* elt=RPCStringParam(e,1);
			CXairoGrammar::Attribute a(att,elt,m_server);
			CXairoAttributeValueList* l=(CXairoAttributeValueList*) m_server->getAttributeValueList(&a);
			if (l==NULL) return RPCStringRet(XMLRPC_NULL);
			xmlrpc_arg x[1];
			x[0].t=XMLRPC_STRING; x[0].s=l->getId();
			return RPCRet(1,x);
		}
	case RM_ATTRIBUTEVALUELIST_GETCOUNT:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			const xairo_uchar* cbid=RPCStringParam(e,1);
			CXairoAttributeValueList l(id,cbid,m_server);
			return RPCIntRet(l.getCount());
		}
	case RM_ATTRIBUTEVALUELIST_GETATTRIBUTEVALUE:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			const xairo_uchar* cbid=RPCStringParam(e,1);
			int i=RPCIntParam(e,2);
			CXairoAttributeValueList l(id,cbid,m_server);
			CXairoAttributeValue* v=(CXairoAttributeValue*) l.getAttributeValue(i);
			xmlrpc_arg x[4];
			x[0].t=XMLRPC_STRING; x[0].s=v->getValue();
			x[1].t=XMLRPC_STRING; x[1].s=v->getGloss();
			x[2].t=XMLRPC_I4; x[2].i=v->getFrequency();
			x[3].t=XMLRPC_STRING; x[3].s=v->getId();
			return RPCRet(4,x);
		}
	case RM_ATTRIBUTEVALUE_GETFILTEREDFREQUENCY:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			const xairo_uchar* fid=RPCStringParam(e,1);
			CXairoCorpusTextFilter f(fid,m_server);
			CXairoAttributeValue av(id,m_server);
			return RPCIntRet(av.getFilteredFrequency(&f));
		}
	case RM_GRAMMAR_GETROOTPARTITION:
		{
			const CXairoGrammar::Partition* p=m_server->GetGrammarImpl()->getRootPartition();
			xmlrpc_arg x[6];
			x[0].t=XMLRPC_STRING; x[0].s=p->getId(); 
			x[1].t=XMLRPC_STRING; x[1].s=p->getName(); 
			x[2].t=XMLRPC_STRING; x[2].s=p->getDescription(); 
			x[3].t=XMLRPC_I4; x[3].i=p->getCount();
			x[4].t=XMLRPC_BOOL; x[4].i=p->hasClasses()?1:0;
			return RPCRet(5,x);
		}
	case RM_GRAMMAR_PARTITION_GETDIVISION:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			int i=RPCIntParam(e,1);
			CXairoGrammar::Partition par(id,m_server);
			const CXairoGrammar::Partition* p=par.getDivision(i);
			xmlrpc_arg x[5];
			x[0].t=XMLRPC_STRING; x[0].s=p->getId(); 
			x[1].t=XMLRPC_STRING; x[1].s=p->getName(); 
			x[2].t=XMLRPC_STRING; x[2].s=p->getDescription(); 
			x[3].t=XMLRPC_I4; x[3].i=p->getCount();
			x[4].t=XMLRPC_BOOL; x[4].i=p->hasClasses()?1:0;
			return RPCRet(5,x);
		}
	case RM_GRAMMAR_PARTITION_GETWORDCOUNT:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			CXairoGrammar::Partition par(id,m_server);
			xmlrpc_arg x[1];
			x[0].t=XMLRPC_I4; x[0].i=par.getWordCount(); 
			return RPCRet(1,x);
		}
	case RM_GRAMMAR_PARTITION_GETINDEX:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			CXairoGrammar::Partition par(id,m_server);
			xmlrpc_arg x[1];
			x[0].t=XMLRPC_I4; x[0].i=par.getIndex(); 
			return RPCRet(1,x);
		}
	case RM_CORPUSTEXT_EVALXPATH:
		{
			int id=RPCIntParam(e,0);
			const xairo_uchar* xq=RPCStringParam(e,1);
			CXairoCorpusText w(id,m_server);
			return RPCStringRet(w.evalXpath(xq)->getText());
		}
	case RM_CORPUSTEXT_GETWORDCOUNT:
		{
			int id=RPCIntParam(e,0);
			CXairoCorpusText w(id,m_server);
			return RPCIntRet(w.getWordCount());
		}
	case RM_CORPUSTEXT_GETINDEX:
		{
			int id=RPCIntParam(e,0);
			CXairoCorpusText w(id,m_server);
			return RPCIntRet(w.getIndex());
		}
	case RM_CORPUSTEXT_ISSELECTED:
		{
			int id=RPCIntParam(e,0);
			const xairo_uchar* idf=RPCStringParam(e,1);
			CXairoCorpusText w(id,m_server);
			CXairoCorpusTextFilter f(idf,m_server);
			return RPCBoolRet(w.isSelected(&f));
		}
	case RM_CORPUSTEXT_GETTEXTENTITY:
		{
			int id=RPCIntParam(e,0);
			CXairoCorpusText w(id,m_server);
			return RPCStringRet(((CXairoTextEntity*) w.getTextEntity())->getURLRemote());
		}
	case RM_GRAMMAR_CODEBOOK_LOOKUP:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			const xairo_uchar* s=RPCStringParam(e,1);
			CXairoGrammar::Codebook cb(id,m_server);
			const xairo_uchar* ss=cb.lookup(s);
			xmlrpc_arg x[1];
			x[0].t=XMLRPC_STRING; x[0].s=ss; 
			return RPCRet(1,x);
		}
	case RM_GRAMMAR_CODEBOOK_GETCOUNT:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			CXairoGrammar::Codebook cb(id,m_server);
			int k=cb.getCount();
			xmlrpc_arg x[1];
			x[0].t=XMLRPC_I4; x[0].i=k; 
			return RPCRet(1,x);
		}
	case RM_GRAMMAR_CODEBOOK_GETPAIR:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			int i=RPCIntParam(e,1);
			CXairoGrammar::Codebook cb(id,m_server);
			const xairo_uchar* n1, *n2;
			cb.getPair(i,&n1,&n2);
			xmlrpc_arg x[2];
			x[0].t=XMLRPC_STRING; x[0].s=n1; 
			x[1].t=XMLRPC_STRING; x[1].s=n2; 
			return RPCRet(2,x);
		}
	case RM_GRAMMAR_GETCODEPAGECOUNT:
		{
			return RPCIntRet(m_server->GetGrammarImpl()->getCodePageCount());
		}
	case RM_GRAMMAR_GETCODEPAGE:
		{
			int i=RPCIntParam(e,0);
			const xairo_uchar* w=m_server->GetGrammarImpl()->getCodePage(i);
			xmlrpc_arg x[1];
			x[0].t=XMLRPC_STRING; x[0].s=w; 
			return RPCRet(1,x);
		}
	case RM_GRAMMAR_GETCHARACTERENTITYCOUNT:
		{
			return RPCIntRet(m_server->GetGrammarImpl()->getCharacterEntityCount());
		}
	case RM_GRAMMAR_GETCHARACTERENTITY:
		{
			int i=RPCIntParam(e,0);
			const xairo_uchar* c; int v;
			m_server->GetGrammarImpl()->getCharacterEntity(i,&c,&v);
			xmlrpc_arg x[2];
			x[0].t=XMLRPC_STRING; x[0].s=c; 
			x[1].t=XMLRPC_I4; x[1].i=v; 
			return RPCRet(2,x);
		}
	case RM_GRAMMAR_GETDSC:
		{
			const char* c=m_server->getGrammar()->getDSC();
			xairo_uchar* w=new xairo_uchar [strlen(c)+1];
			mbstowcs(w,c,strlen(c)+1);
			return RPCStringRet(w);
		}

	case RM_GETDEFAULTLANGUAGE:
		{
			const xairo_uchar* n=m_server->getDefaultLanguage();
			xmlrpc_arg x[1];
			x[0].t=XMLRPC_STRING; x[0].s=n;
			return RPCRet(1,x);
		}
	case RM_GRAMMAR_GETLEMMASCHEME:
		{
			int i=RPCIntParam(e,0);
			const CXairoGrammar::LemmaScheme *ls=m_server->GetGrammarImpl()->getLemmaScheme(i);
			xmlrpc_arg x[4];
			x[0].t=XMLRPC_STRING; x[0].s=ls->getId(); 
			x[1].t=XMLRPC_STRING; x[1].s=ls->getName(); 
			x[2].t=XMLRPC_STRING; x[2].s=ls->getGloss(); 
			x[3].t=XMLRPC_I4; x[3].i=ls->getFacetCount(); 
			return RPCRet(4,x);
		}
	case RM_GRAMMAR_GETLEMMASCHEMECOUNT:
		{
			return RPCIntRet(m_server->GetGrammarImpl()->getLemmaSchemeCount());
		}
	case RM_GRAMMAR_LEMMASCHEME_GETFACET:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			int i=RPCIntParam(e,1);
			CXairoGrammar::LemmaScheme ls(id,m_server);
			return RPCStringRet(ls.getFacet(i));
		}
	case RM_GRAMMAR_GETREGION:
		{
			int i=RPCIntParam(e,0);
			const CXairoGrammar::Region *ls=m_server->GetGrammarImpl()->getRegion(i);
			xmlrpc_arg x[5];
			x[0].t=XMLRPC_STRING; x[0].s=ls->getId(); 
			x[1].t=XMLRPC_STRING; x[1].s=ls->getName(); 
			x[2].t=XMLRPC_STRING; x[2].s=ls->getGloss(); 
			x[3].t=XMLRPC_I4; x[3].i=ls->getValueCount(); 
			x[4].t=XMLRPC_I4; x[4].i=ls->getAddkey()->GetDic(); 
			return RPCRet(5,x);
		}
	case RM_GRAMMAR_GETREGIONCOUNT:
		{
			return RPCIntRet(m_server->GetGrammarImpl()->getRegionCount());
		}
	case RM_GRAMMAR_REGION_GETVALUE:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			int i=RPCIntParam(e,1);
			CXairoGrammar::Region ls(id,m_server);
			return RPCStringRet(ls.getValue(i));
		}
	case RM_GETADDKEYFILTER:
		{
			const xairo_uchar* key=RPCStringParam(e,0);
			const xairo_uchar **p=NULL;
			int np=RPCVectorsParam(e,1,p);
			CXairoAddkeyFilter* f=(CXairoAddkeyFilter*) m_server->getAddkeyFilter(key,np,p);
			xmlrpc_arg x[1];
			x[0].t=XMLRPC_STRING;
			if (f==NULL) x[0].s=XMLRPC_NULL;
			else x[0].s=f->getId();
			return RPCRet(1,x);
		}
	case RM_GETADDKEYVALUELIST:
		{
			const xairo_uchar* att=RPCStringParam(e,0);
			CXairoGrammar::Addkey* a=(CXairoGrammar::Addkey*) m_server->GetGrammarImpl()->getAddkeyById(att);
			CXairoAddkeyValueList* l=(CXairoAddkeyValueList*) m_server->getAddkeyValueList(a);
			if (l==NULL) return RPCStringRet(XMLRPC_NULL);
			xmlrpc_arg x[1];
			x[0].t=XMLRPC_STRING; x[0].s=l->getId();
			return RPCRet(1,x);
		}
	case RM_ADDKEYVALUELIST_GETCOUNT:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			CXairoAddkeyValueList l(id,m_server);
			return RPCIntRet(l.getCount());
		}
	case RM_ADDKEYVALUELIST_GETADDKEYVALUE:
		{
			const xairo_uchar* id=RPCStringParam(e,0);
			int i=RPCIntParam(e,1);
			CXairoAddkeyValueList l(id,m_server);
			CXairoAddkeyValue* v=(CXairoAddkeyValue*) l.getAddkeyValue(i);
			xmlrpc_arg x[2];
			x[0].t=XMLRPC_STRING; x[0].s=v->getValue();
			x[1].t=XMLRPC_I4; x[1].i=v->getFrequency();
			return RPCRet(2,x);
		}
	default:
		throw new CXairoError(CXairoIError::ERR_XMLRPC,L("RPC: unknown method"));
}
}



XAIRO_NAMESPACE_END
