#pragma once
#include "XairoProxyServer.h"

class CXairoProxyServer;

class CXairoProxyGrammar :
	public CXairoIGrammar
{
public:
	CXairoProxyGrammar();
	~CXairoProxyGrammar(void);

	int getNamespaceCount() const;
	
	class Namespace : public CXairoIGrammar::Namespace
	{
	public: 
		Namespace(const xairo_uchar* url):m_url(url) {}
		const xairo_uchar* getUrl() const {return m_url;}
	private:
		const xairo_uchar* m_url;
	};
	const Namespace* getNamespace(int) const;

	int getAddkeyCount() const;

	class Addkey : public CXairoIGrammar::Addkey
	{
	public:
		Addkey(const xairo_uchar* id,const xairo_uchar* n,const xairo_uchar* g,bool b,
			const xairo_uchar* cbid,const xairo_uchar* a,CXairoProxyServer* s):
		  m_id(id),m_name(n), m_gloss(g), m_bLemma(b), 
			  m_att(a), m_cbid(cbid), m_server(s) {}
		const xairo_uchar* getName() const {return m_name;}
		const xairo_uchar* getGloss() const {return m_gloss;}
		bool isLemmaScheme() const {return m_bLemma;}
		const xairo_uchar* getAttribute() const {return m_att;}
		const Codebook* getCodebook() const {if (wcscmp(m_cbid,L("null"))==0) return NULL; else return new Codebook(m_cbid,m_server);}
		const xairo_uchar* getId() {return m_id;}

	private:
		const xairo_uchar* m_name;
		const xairo_uchar* m_gloss;
		const xairo_uchar* m_att;
		bool m_bLemma;
		const xairo_uchar* m_cbid;
		const xairo_uchar* m_id;
		CXairoProxyServer* m_server;
	};
	const Addkey* getAddkey(int i) const;

	int getLemmaSchemeCount() const;

	class LemmaScheme : public CXairoIGrammar::LemmaScheme
	{
	public:
		LemmaScheme(const xairo_uchar* id,const xairo_uchar* n,const xairo_uchar* g,int nf,CXairoProxyServer* s):
		  m_name(n),m_gloss(g),m_nfacets(nf),m_id(id),m_server(s) {}
		const xairo_uchar* getName() const {return m_name;}
		const xairo_uchar* getGloss() const {return m_gloss;}
		int getFacetCount() const {return m_nfacets;}
		const xairo_uchar* getFacet(int) const;
	private:
		const xairo_uchar* m_id;
		const xairo_uchar* m_name;
		const xairo_uchar* m_gloss;
		int m_nfacets;
		CXairoProxyServer* m_server;
	};

	const LemmaScheme* getLemmaScheme(int) const;

	int getRegionCount() const;

	class Region : public CXairoIGrammar::Region
	{
	public:
		Region(const xairo_uchar* id,const xairo_uchar* n,const xairo_uchar* g,int nv,int nak,CXairoProxyServer* s):
		  m_name(n),m_gloss(g),m_nvals(nv),m_nak(nak),m_id(id),m_server(s) {}
		const xairo_uchar* getName() const {return m_name;}
		const xairo_uchar* getGloss() const {return m_gloss;}
		int getValueCount() const {return m_nvals;}
		const xairo_uchar* getValue(int) const;
		const Addkey* getAddkey() const;
	private:
		const xairo_uchar* m_id;
		const xairo_uchar* m_name;
		const xairo_uchar* m_gloss;
		int m_nvals;
		int m_nak;
		CXairoProxyServer* m_server;
	};

	const Region* getRegion(int) const;

	class Codebook : public CXairoIGrammar::Codebook
	{
	public:
		Codebook(const xairo_uchar* id,CXairoProxyServer* s):
		  m_id(id),m_server(s) {}
		const xairo_uchar* lookup(const xairo_uchar*) const;
		int getCount() const;
		void getPair(int,const xairo_uchar**,const xairo_uchar**) const;
	private:
		const xairo_uchar* m_id;
		CXairoProxyServer* m_server;
	};

	class Attribute : public CXairoIGrammar::Attribute
	{
	public:
		Attribute(const xairo_uchar* id,const xairo_uchar* n,const xairo_uchar* g,
			const xairo_uchar* url,const xairo_uchar* cb,const xairo_uchar* e,
			ip p, const xairo_uchar* tar, bool bB,
			CXairoProxyServer* s):
		  m_id(id),m_name(n),m_gloss(g),m_url(url),m_cbid(cb),m_elt(e),
			  m_ip(p),m_target(tar),m_bInBib(bB),m_server(s) {}
		const xairo_uchar* getName() const {return m_name;}
		const xairo_uchar* getGloss() const {return m_gloss;}
		const Namespace* getNamespace() const  {return new Namespace(m_url);}
		const Codebook* getCodebook() const {if (wcscmp(m_cbid,L("null"))==0) return NULL; else return new Codebook(m_cbid,m_server);}
		ip getIP() const {return m_ip;}
		const xairo_uchar* getTarget() const {return m_target;}
		bool inBib() const {return m_bInBib;}
		const xairo_uchar* getElt() const {return m_elt;}
		const xairo_uchar* getId() {return m_id;}
		const xairo_uchar* getCBId() {return m_cbid;}
	private:
		const xairo_uchar* m_id;
		const xairo_uchar* m_name;
		const xairo_uchar* m_gloss;
		const xairo_uchar* m_url;
		const xairo_uchar* m_cbid;
		const xairo_uchar* m_elt;
		CXairoProxyServer* m_server;
		ip m_ip;
		bool m_bInBib;
		const xairo_uchar* m_target;
	};
	int getElementCount() const;

	class Element : public CXairoIGrammar::Element
	{
	public:
		Element(const xairo_uchar* id,const xairo_uchar* n,const xairo_uchar* g,int ac,const xairo_uchar* url,bool bB,CXairoProxyServer* s):
		  m_id(id),m_name(n),m_gloss(g),m_url(url),m_natts(ac),m_bInBib(bB),m_server(s) {}
		const xairo_uchar* getName() const {return m_name;}
		const xairo_uchar* getGloss() const  {return m_gloss;}
		int getAttributeCount() const {return m_natts;}
		const Attribute* getAttribute(int) const;
		const Namespace* getNamespace() const {return new Namespace(m_url);}
		bool inBib() const {return m_bInBib;}
	private:
		const xairo_uchar* m_id;
		const xairo_uchar* m_name;
		const xairo_uchar* m_gloss;
		const xairo_uchar* m_url;
		int m_natts;
		CXairoProxyServer* m_server;
		bool m_bInBib;

	};
	const Element* getElement(int) const;

	int getGlobalAttributeCount() const;
	const Attribute* getGlobalAttribute(int) const;

	class Partition : public CXairoIGrammar::Partition
	{
	public:
		Partition(const xairo_uchar* id,const xairo_uchar* n,const xairo_uchar* d,
			int c,bool b,CXairoProxyServer* s):m_id(id),m_name(n),m_desc(d),m_count(c),
			m_bhasClasses(b),m_server(s) {}
		const xairo_uchar* getName() const {return m_name;}
		const xairo_uchar* getDescription() const  {return m_desc;}
		int getCount() const {return m_count;}
		const Partition* getDivision(int) const;
		bool hasClasses() const {return m_bhasClasses;}
		int getWordCount() const;
		int getIndex() const;
		const xairo_uchar* getId() {return m_id;}
	private:
		const xairo_uchar* m_id;
		const xairo_uchar* m_name;
		const xairo_uchar* m_desc;
		int m_count;
		bool m_bhasClasses;
		CXairoProxyServer* m_server;
	};
	const Partition* getRootPartition() const;

	class Label : public CXairoIGrammar::Label
	{
	public:
		Label(const xairo_uchar* e,const xairo_uchar* a):m_elt(e),m_att(a) {}
		const xairo_uchar* getElement() const {return m_elt;}
		const xairo_uchar* getAttribute() const {return m_att;}
	private:
		const xairo_uchar* m_elt;
		const xairo_uchar* m_att;
	};
	int getLabelCount() const ;
	const Label* getLabel(int) const;

	int getCodePageCount() const;
	const xairo_uchar* getCodePage(int) const;

	int getCharacterEntityCount() const;
	void getCharacterEntity(int,const xairo_uchar**,int*) const; 


	int getWTagCount() const ;
	const xairo_uchar* getWTag(int) const;

	const char* getDSC() const;

	class CXairoProxyServer* m_server;
};
