#include "stdafx.h"
#include "XairoProxyServer.h"
#include "XairoProxyWordList.h"
#include "XairoProxySolution.h"
#include "XairoProxySolveError.h"
#include "XairoProxyCorpusTextList.h"
#include "XairoProxyCorpusTextFilter.h"
#include "XairoProxyAttributeValueList.h"
#include "XairoProxyExtendedForm.h"
#include "XairoProxyAddkeyFilter.h"
#include "XairoProxyAddkeyValueList.h"
//#include "XairoStubServerI.h"

#ifdef _WIN32
#include <Ws2tcpip.h>
#include <Wspiapi.h>
#endif

const xairo_uchar* ErrString(int dw);

void CXairoProxyServer::Error(const xairo_uchar* s)
{
	throw new CXairoProxyError(CXairoIError::ERR_UNKNOWN,s);
}

CXairoProxyServer::CXairoProxyServer()
{
	m_server=XAIRO_NAMESPACE::CXairoImplementation::getStubServer();
	m_grammar.m_server=this;
	RPCSetErrorHandler(Error);
}

CXairoProxyServer::CXairoProxyServer(CXairoInterface* s)
{
	m_server=XAIRO_NAMESPACE::CXairoImplementation::getStubServer(s);
	m_grammar.m_server=this;
	m_transport=NULL;
}

CXairoProxyServer::~CXairoProxyServer(void)
{
	m_transport->End();
}

bool CXairoProxyServer::Configure(CXairoTransport* t,XAIRO_NAMESPACE::CXairoIBootMonitor* bm)
{
	XMLTransService::Codes res;
	m_utf8=XMLPlatformUtils::fgTransService->makeNewTranscoderFor("UTF-8",res,1024);
	m_transport=t;
	try {
		m_transport->Start();
	}
	catch (CXairoIError* e) {
		bm->Log(e->getMessage());
		return false;
	}
	return true;
}

bool CXairoProxyServer::Configure(bool bLocal,const char* address,unsigned port,XAIRO_NAMESPACE::CXairoIBootMonitor* bm)
{
	CXairoTransport* t;
	if (bLocal)
		t=new CXairoLocalDefault(m_server);
	else
		t=new CXairoSocketDefault(address,port);
	return Configure(t,bm);
}

class CXairoIWordList* CXairoProxyServer::getWordList(int np,const xairo_uchar** re,const xairo_uchar* l)
{
	const xairo_uchar* s;
	DOMElement* d;
	xmlrpc_arg x[3];
	x[0].t=XMLRPC_VECTORS; x[0].vecs=re; x[0].nvec=np;
	x[1].t=XMLRPC_STRING; x[1].s=l;
	s=RPCCall(L("getWordList"),2,x);
//	s.Format(L"<methodCall><methodName>getWordList</methodName><params><param><value><string>%s</string></value></param><param><value><string>%s</string></value></param></params></methodCall>",re,l);
	d=RPCRemote(s); delete [] s;
//	int k=RPCInt(d,L"status");
//	if (k) return NULL;
	const xairo_uchar* id=RPCStringParam(d,0);
	if (wcscmp(id,L("null"))==0) return NULL;
	return new CXairoProxyWordList(id,this);
}

class CXairoIWordList* CXairoProxyServer::getFrequencyList(int np,const xairo_uchar** re,int n,int m,int p,const xairo_uchar* l)
{
	const xairo_uchar* s;
	DOMElement* d;
	xmlrpc_arg x[5];
	x[0].t=XMLRPC_VECTORS; x[0].vecs=re; x[0].nvec=np;
	x[1].t=XMLRPC_I4; x[1].i=n;
	x[2].t=XMLRPC_I4; x[2].i=m;
	x[3].t=XMLRPC_I4; x[3].i=p;
	x[4].t=XMLRPC_STRING; x[4].s=l;
	s=RPCCall(L("getFrequencyList"),5,x);
//	s.Format(L"<methodCall><methodName>getFrequencyList</methodName><params><param><value><string>%s</string></value></param><param><value><i4>%d</i4></value></param><param><value><i4>%d</i4></value></param><param><value><i4>%d</i4></value></param><param><value><string>%s</string></value></param></params></methodCall>",);
	d=RPCRemote(s); delete [] s;
//	int k=RPCInt(d,L"status");
//	if (k) return NULL;
	const xairo_uchar* id=RPCStringParam(d,0);
	if (wcscmp(id,L("null"))==0) return NULL;
	return new CXairoProxyWordList(id,this);
}



class CXairoIWordList* CXairoProxyServer::getFilteredFrequencyList(int np,const xairo_uchar** re,int n,int m,int p,const xairo_uchar* l,const CXairoICorpusTextFilter* f,const CXairoIAddkeyFilter* ff)
{
	const xairo_uchar* s;
	DOMElement* d;
	xmlrpc_arg x[7];
	x[0].t=XMLRPC_VECTORS; x[0].vecs=re; x[0].nvec=np;
	x[1].t=XMLRPC_I4; x[1].i=n;
	x[2].t=XMLRPC_I4; x[2].i=m;
	x[3].t=XMLRPC_I4; x[3].i=p;
	x[4].t=XMLRPC_STRING; x[4].s=l;
	x[5].t=XMLRPC_STRING; x[5].s=(f==NULL)?L("null"):((CXairoProxyCorpusTextFilter*) f)->getId();
	x[6].t=XMLRPC_STRING; x[6].s=(ff==NULL)?L("null"):((CXairoProxyAddkeyFilter*) ff)->getId();
	s=RPCCall(L("getFilteredFrequencyList"),7,x);
	d=RPCRemote(s); delete [] s;
	const xairo_uchar* id=RPCStringParam(d,0);
	if (wcscmp(id,L("null"))==0) return NULL;
	return new CXairoProxyWordList(id,this);
}

CXairoIAttributeValueList* CXairoProxyServer::getAttributeValueList(const CXairoIGrammar::Attribute* att)
{
	const xairo_uchar* s;
	DOMElement* d;
	xmlrpc_arg x[2];
	x[0].t=XMLRPC_STRING; x[0].s=((CXairoProxyGrammar::Attribute*) att)->getId();
	x[1].t=XMLRPC_STRING; x[1].s=((CXairoProxyGrammar::Attribute*) att)->getElt();
	const xairo_uchar* cbid=((CXairoProxyGrammar::Attribute*) att)->getCBId();
	s=RPCCall(L("getAttributeValueList"),2,x);
	d=RPCRemote(s); delete [] s;
	const xairo_uchar* id=RPCStringParam(d,0);
	if (wcscmp(id,L("null"))==0) return NULL;
	else return new CXairoProxyAttributeValueList(id,cbid,this);
}

CXairoIAddkeyValueList* CXairoProxyServer::getAddkeyValueList(const CXairoIGrammar::Addkey* a)
{
	const xairo_uchar* s;
	DOMElement* d;
	xmlrpc_arg x[1];
	x[0].t=XMLRPC_STRING; x[0].s=((CXairoProxyGrammar::Addkey*) a)->getId();
	s=RPCCall(L("getAddkeyValueList"),1,x);
	d=RPCRemote(s); delete [] s;
	const xairo_uchar* id=RPCStringParam(d,0);
	if (wcscmp(id,L("null"))==0) return NULL;
	else return new CXairoProxyAddkeyValueList(id,this);
}


CXairoISolution* CXairoProxyServer::solve(const xairo_uchar* q)
{
	const xairo_uchar* s;
	DOMElement* d;
	xmlrpc_arg  x[1];
	x[0].t=XMLRPC_STRING; x[0].s=q;
	s=RPCCall(L("solve"),1,x);
//	s.Format(L"<methodCall><methodName>solve</methodName><params><param><value><string>%s</string></value></param></params></methodCall>",RPCEscape(q));
	d=RPCRemote(s); delete [] s;
//	int k=RPCInt(d,L"status");
//	if (k) return NULL;
	const xairo_uchar* id=RPCStringParam(d,0);
	if (wcscmp(id,L("null"))==0) return NULL;
	else return new CXairoProxySolution(id,this);
}
/*
CXairoISolution* CXairoProxyServer::solveInDivision(const xairo_uchar* q,CXairoIGrammar::Partition* p)
{
	const xairo_uchar* s;
	DOMElement* d;
	xmlrpc_arg  x[2];
	x[0].t=XMLRPC_STRING; x[0].s=q;
	x[1].t=XMLRPC_STRING; x[1].s=((CXairoProxyGrammar::Partition*) p)->getId();
	s=RPCCall(L("solveInDivision"),2,x);
//	s.Format(L"<methodCall><methodName>solve</methodName><params><param><value><string>%s</string></value></param></params></methodCall>",RPCEscape(q));
	d=RPCRemote(s); delete [] s;
//	int k=RPCInt(d,L"status");
//	if (k) return NULL;
	const xairo_uchar* id=RPCStringParam(d,0);
	if (wcscmp(id,L("null"))==0) return NULL;
	else return new CXairoProxySolution(id,this);
}

*/

int CXairoProxyServer::getWordCount()
{
	const xairo_uchar* s;
	DOMElement* d;
	s=RPCCall(L("getWordCount"),0,NULL);
	d=RPCRemote(s); delete [] s;
	return RPCIntParam(d,0);
}

const xairo_uchar* CXairoProxyServer::getCorpusName()
{
	const xairo_uchar* s;
	DOMElement* d;
	s=RPCCall(L("getCorpusName"),0,NULL);
	d=RPCRemote(s); delete [] s;
	return RPCStringParam(d,0);
}

const xairo_uchar* CXairoProxyServer::getDefaultLanguage()
{
	const xairo_uchar* s;
	DOMElement* d;
	s=RPCCall(L("getDefaultLanguage"),0,NULL);
	d=RPCRemote(s); delete [] s;
	return RPCStringParam(d,0);
}

CXairoICorpusTextList* CXairoProxyServer::getCorpusTextList()
{
	return new CXairoProxyCorpusTextList(this);
}

CXairoIExtendedForm* CXairoProxyServer::getUsage(unsigned int i)
{
	return new CXairoProxyExtendedForm(i,this);
}

CXairoIAddkeyFilter* CXairoProxyServer::getAddkeyFilter(const xairo_uchar* key,int nvals,const xairo_uchar** vals)
{
	const xairo_uchar* s;
	DOMElement* d;
	xmlrpc_arg x[2];
	x[0].t=XMLRPC_STRING; x[0].s=key;
	x[1].t=XMLRPC_VECTORS; x[1].nvec=nvals; x[1].vecs=vals;
 	s=RPCCall(L("getAddkeyFilter"),2,x);
	d=RPCRemote(s); delete [] s;
	return new CXairoProxyAddkeyFilter(RPCStringParam(d,0),this);
}


void* CXairoProxyServer::deserialize(const xairo_uchar* c)
{
	if (wcsncmp(c,L("sol"),3)==0)
		return new CXairoProxySolution(c,this);
	else if (wcsncmp(c,L("ctf"),3)==0)
		return new CXairoProxyCorpusTextFilter(c,this);
	else return NULL;
}
	
char* CXairoSocketDefault::DoTransport(const char* s)
{
	return Transport(s);
}

char* CXairoLocalDefault::DoTransport(const char* s)
{
	return m_server->Transport(s);
}

// This has been changed to use UTF-8 across the net

char* CXairoSocketDefault::Transport(const char* s)
{                   
	int k1,k2;
	k1=(int) strlen(s)+1;
	k2=send(m_socket,s,k1,0);
	assert(k2<0||k1==k2);
	int rn;
	char* t;
	if (k2<0) rn=k2;
	Receive(t);
	return t;
}

#define BSIZE 500000

char bu[BSIZE];

int CXairoSocketDefault::Receive(char*& w)
{            
	int n=0; char* x=NULL; bool bClosed=false;
	w=NULL;
	for (;;) {
     	int k = recv(m_socket, bu, BSIZE, 0); 
		if (k==0) {
			throw new CXairoProxyError(CXairoIError::ERR_UNKNOWN,L("Connection closed by server"));
		}
		if (k==SOCKET_ERROR) {
#ifndef _WIN32
			int e=errno;
			throw new CXairoProxyError(CXairoIError::ERR_UNKNOWN,strerror(e));
#else
			int e=WSAGetLastError();
			throw new CXairoProxyError(CXairoIError::ERR_UNKNOWN,ErrString(e));
#endif
		}
		char* xx=new char[n+k];
		if (x) memcpy(xx,x,n);
		memcpy(xx+n,bu,k);
		if (x) delete [] x;
		x=xx;
		n+=k;
		w=x;
		int nw=n;
		if (w[nw-1]==0) break;
	}
	if (bClosed) return -1;
	return (int) strlen(w);
}

#define IPV6_SUPPORT


CXairoSocketDefault::CXairoSocketDefault(const char* dns,unsigned port)
{
	m_port=port;
	m_address=dns;
}

CXairoLocalDefault::CXairoLocalDefault(CXairoIStubServer* s)
{
	m_server=s;
}

void CXairoLocalDefault::Start()
{
}

void CXairoSocketDefault::Start()
{     
#ifdef _WIN32
	WORD wVersionRequested;
	WSADATA wsaData;
	int err;
	wVersionRequested = MAKEWORD( 2, 2 );
	err = WSAStartup( wVersionRequested, &wsaData );
	if ( err != 0 ) {
		throw new CXairoProxyError(CXairoIError::ERR_UNKNOWN,L("Cannot initialise Winsock"));
	}
#endif
	m_socket=socket(AF_INET, SOCK_STREAM, 0);
	if (m_socket==INVALID_SOCKET) {
		throw new CXairoProxyError(CXairoIError::ERR_UNKNOWN,L("Error opening socket"));
    }
	int e;
#ifdef IPV6_SUPPORT // this is the approved IPv6 way of doing this
	struct addrinfo a; struct addrinfo* aa;
	char port[NI_MAXSERV];
	sprintf(port,"%d",m_port);
	memset(&a, 0, sizeof(a));
	a.ai_family = AF_INET;
	a.ai_socktype = SOCK_STREAM;
	a.ai_protocol = IPPROTO_TCP;
	e=getaddrinfo(m_address,port,&a,&aa);
	if (e==0) {
		while (aa) {
			int e=connect(m_socket,aa->ai_addr,(int) aa->ai_addrlen);
			if (e != SOCKET_ERROR) break;
			aa=aa->ai_next;
		}
	}
#else
	hostent* h=NULL;
	if (isalpha(m_address[0])){
		h=gethostbyname(m_address);
	} else {
		unsigned addr = inet_addr(m_address);
		h = gethostbyaddr((char *)&addr, 4, AF_INET);
	}
	if (h==NULL) {
		throw new CXairoProxyError(L("Cannot resolve hostname"));
	}
	struct sockaddr_in service;
    service.sin_family = AF_INET;
    memcpy(&service.sin_addr,h->h_addr_list[0],h->h_length);
    service.sin_port = htons(m_port);
//    if (bind(m_socket, (struct sockaddr*)&passive,sizeof(struct sockaddr))==SOCKET_ERROR) {
//		throw new CXairoProxyError(L("Error binding socket"));
  //  }
	int e=connect(m_socket,(struct sockaddr* )&service,sizeof(struct sockaddr));
#endif
	if (e==SOCKET_ERROR)
	{	
#ifndef _WIN32
		int e=errno;
		throw new CXairoProxyError(CXairoIError::ERR_UNKNOWN,strerror(e));
#else
		int e=WSAGetLastError();
		throw new CXairoProxyError(CXairoIError::ERR_UNKNOWN,ErrString(e));
#endif
	}
	const char* s=DoTransport("<?xaira-client?>");
	if (strncmp(s,"<?xaira-server",14) != 0)
		throw new CXairoProxyError(CXairoIError::ERR_UNKNOWN,"Server is not Xaira 1.0");
}

void CXairoLocalDefault::End()
{
}

void CXairoSocketDefault::End()
{
#ifndef _WIN32
	close(m_socket);
#else
	closesocket(m_socket);
#endif
}


class EH : public HandlerBase
{
	void error  (  const SAXParseException &    exc  ) {throw exc;}
	void fatalError  (  const SAXParseException &    exc  ) {throw exc;}
};


DOMElement* CXairoProxyServer::RPCRemote(const xairo_uchar* s)
{
	int bs=4*((int) wcslen(s)+1);
	char* x=new char[bs]; XMLSize_t  ce;
	XMLSize_t k=m_utf8->transcodeTo(s,(int) wcslen(s),(XMLByte*) x,bs,ce,XMLTranscoder::UnRep_Throw);
	x[k]=0;
	char* xx=m_transport->DoTransport(x);
	delete [] x;
	MemBufInputSource* is=new MemBufInputSource((XMLByte*) xx,(int) strlen(xx),L("xml-rpc")); 
	is->setEncoding(L("UTF-8"));
	EH eh;
	m_parser.setErrorHandler(&eh);
	try {
		m_parser.parse(*is);
		delete is;
		free(xx); // wasn't allocated by new
	}
	catch (SAXParseException& e) {
		throw new CXairoProxyError(CXairoIError::ERR_XMLRPC,e.getMessage());
	}
	DOMElement* d=m_parser.getDocument()->getDocumentElement();
	DOMNodeList* l=d->getElementsByTagName(L("fault"));
	if (l->getLength()) {
		DOMElement* f=(DOMElement*) l->item(0);
		DOMElement* s=RPCStruct(RPCValue(f));
		DOMNodeList* nl=s->getElementsByTagName(L("member"));
		const xairo_uchar* msg; int n=0;
		for (int i=0;i<(int) nl->getLength();i++) {
			DOMElement* m=(DOMElement*) nl->item(i);
			if (wcscmp(RPCName(m),L("faultString"))==0)
				msg=RPCString(RPCValue(m));
			if (wcscmp(RPCName(m),L("faultCode"))==0)
				n=RPCInt(RPCValue(m));
		}
		throw new CXairoProxyError((CXairoIError::code) n,msg); 
	}
	return d;
}

#if 0

DOMElement* CXairoProxyServer::RPCParam(DOMElement* e,int n)
{
	DOMNodeList* nl=e->getElementsByTagName(L"param");
	int k=nl->getLength();
	if (n >= k) return NULL;
	DOMNode* nd=nl->item(n);
	return (DOMElement*) nd;
}

DOMElement* CXairoProxyServer::RPCValue(DOMElement* e)
{
	DOMNodeList* nl=e->getElementsByTagName(L"value");
	int k=nl->getLength();
	if (k==0) return NULL;
	DOMNode* nd=nl->item(0);
	return (DOMElement*) nd;
}	

const xairo_uchar* CXairoProxyServer::RPCName(DOMElement* e)
{
	DOMNodeList* nl=e->getElementsByTagName(L"name");
	int k=nl->getLength();
	if (k==0) return NULL;
	DOMNode* nd=nl->item(0);
	return nd->getTextContent();
}	

DOMElement* CXairoProxyServer::RPCStruct(DOMElement* e)
{
	DOMNodeList* nl=e->getElementsByTagName(L"struct");
	int k=nl->getLength();
	if (k==0) return NULL;
	DOMNode* nd=nl->item(0);
	return (DOMElement*) nd;
}	

const xairo_uchar* CXairoProxyServer::RPCString(DOMElement* e)
{
	DOMNodeList* nl=e->getElementsByTagName(L"string");
	int k=nl->getLength();
	if (k==0) return L"";
	DOMNode* nd=nl->item(0);
	return nd->getTextContent();
}
	
int CXairoProxyServer::RPCInt(DOMElement* e)
{
	DOMNodeList* nl=e->getElementsByTagName(L"i4");
	int k=nl->getLength();
	if (k==0) return 0;
	DOMNode* nd=nl->item(0);
	const xairo_uchar* x=nd->getTextContent();
	return _wtoi(x);
}

double CXairoProxyServer::RPCDouble(DOMElement* e)
{
	DOMNodeList* nl=e->getElementsByTagName(L"double");
	int k=nl->getLength();
	if (k==0) return 0;
	DOMNode* nd=nl->item(0);
	const xairo_uchar* x=nd->getTextContent();
	return _wtof(x);
}

bool CXairoProxyServer::RPCBool(DOMElement* e)
{
	DOMNodeList* nl=e->getElementsByTagName(L"boolean");
	int k=nl->getLength();
	if (k==0) return 0;
	DOMNode* nd=nl->item(0);
	const xairo_uchar* x=nd->getTextContent();
	return _wtoi(x) != 0;
}

int CXairoProxyServer::RPCVectorParam(DOMElement* d,int i,int *& v)
{
	return RPCVector(RPCValue(RPCParam(d,i)),v);
}

int CXairoProxyServer::RPCVector(DOMElement* e,int*& v)
{
	DOMNodeList* nl=e->getElementsByTagName(L"data");
	if (nl->getLength()==0) return 0;
	DOMElement* ex=(DOMElement*) nl->item(0);
	DOMNodeList* nll=ex->getChildNodes();
	int k=0;
	for (int i=0;i<(int) nll->getLength();i++) {
		DOMNode* n=nll->item(i);
		if (n->getNodeType() != DOMNode::ELEMENT_NODE) continue;
		DOMElement* ne=(DOMElement*) n;
		if (wcscmp(ne->getTagName(),L"value") != 0) continue;
		k++;
	}
	v=new int [k];
	k=0;
	for (int i=0;i<(int) nll->getLength();i++) {
		DOMNode* n=nll->item(i);
		if (n->getNodeType() != DOMNode::ELEMENT_NODE) continue;
		DOMElement* ne=(DOMElement*) n;
		if (wcscmp(ne->getTagName(),L"value") != 0) continue;
		v[k++]=RPCInt(ne);
	}
	return k;
}

int CXairoProxyServer::RPCIntParam(DOMElement* d,int i)
{
	return RPCInt(RPCValue(RPCParam(d,i)));
}

const xairo_uchar* CXairoProxyServer::RPCStringParam(DOMElement* d,int i)
{
	return RPCString(RPCValue(RPCParam(d,i)));
}

bool CXairoProxyServer::RPCBoolParam(DOMElement* d,int i)
{
	return RPCBool(RPCValue(RPCParam(d,i)));
}

double CXairoProxyServer::RPCDoubleParam(DOMElement* d,int i)
{
	return RPCDouble(RPCValue(RPCParam(d,i)));
}

const xairo_uchar* CXairoProxyServer::RPCEscape(const xairo_uchar* s)
{
	int t=0;
	for (int i=0;i<(int) wcslen(s);i++) {
		if (s[i]=='<') t+=4;
		else if (s[i]=='&') t+=5;
		else t++;
	}
	xairo_uchar* w=new xairo_uchar [t+1]; w[0]=0;
	xairo_uchar* ww=w;
	for (int i=0;i<(int) wcslen(s);i++) {
		if (s[i]=='<') {
			wcscpy(ww,L"&lt;"); ww+=4;
		}
		else if (s[i]=='&') {
			wcscpy(ww,L"&amp;"); ww+=5;
		}
		else {
			*(ww++)=s[i]; 
		}
	}
	*ww=0;
	return w;
}

const xairo_uchar* CXairoProxyServer::RPCUnescape(const xairo_uchar* s)
{
	xairo_uchar* w=new xairo_uchar [wcslen(s)+1];
	xairo_uchar* ww=w;
	for (int i=0;i<(int) wcslen(s);i++) {
		if (s[i]=='&') {
			if (wcsncmp(s+i,L"&lt;",4)==0) {
				*ww++='<'; i+=3;
			} else if (wcsncmp(s+i,L"&amp;",5)==0) {
				*ww++='&'; i+=4;
			}
			else *ww++=s[i];
		}
	}
	*ww=0;
	return w;
}


#define P1 L"<params>"
#define P2 L"</params>"
#define A1 L"<param><value><array><data>"
#define A2 L"</data></array></value></param>"

const xairo_uchar* CXairoProxyServer::RPCParams(int n,const xmlrpc_arg* x)
{
	int tx=wcslen(P1)+wcslen(P2);
	for (int i=0;i<n;i++) {
		switch (x[i].t) {
			int t;
			case XMLRPC_STRING:
				t=_scwprintf(L"<param><value><string>%s</string></value></param>",RPCEscape(x[i].s));
				tx+=t;
				break;
			case XMLRPC_I4:
				t=_scwprintf(L"<param><value><i4>%d</i4></value></param>",x[i].i);
				tx+=t;
				break;
			case XMLRPC_BOOL:
				t=_scwprintf(L"<param><value><boolean>%d</i4></boolean></param>",x[i].i);
				tx+=t;
				break;
			case XMLRPC_DOUBLE:
				t=_scwprintf(L"<param><value><double>%f</double></boolean></param>",x[i].d);
				tx+=t;
				break;
			case XMLRPC_VECTOR:
				tx+=wcslen(A1)+wcslen(A2);
				for (int j=0;j<x[i].nvec;j++) {
					tx+=_scwprintf(L"<value><i4>%d</i4></value>",x[i].vec[j]);
				}
				break;
			case XMLRPC_VECTORS:
				tx+=wcslen(A1)+wcslen(A2);
				for (int j=0;j<x[i].nvec;j++) {
					tx+=_scwprintf(L"<value><string>%s</string></value>",x[i].vecs[j]);
				}
				break;
			default:
				ASSERT(FALSE);
		}
	}
	xairo_uchar* w=new xairo_uchar [tx+1];
	wcscpy(w,P1); xairo_uchar* ww=w+wcslen(P1);
	for (int i=0;i<n;i++) {
		switch (x[i].t) {
			case XMLRPC_STRING:
				ww+=swprintf(ww,L"<param><value><string>%s</string></value></param>",RPCEscape(x[i].s));
				break;
			case XMLRPC_I4:
				ww+=swprintf(ww,L"<param><value><i4>%d</i4></value></param>",x[i].i);
				break;
			case XMLRPC_BOOL:
				ww+=swprintf(ww,L"<param><value><boolean>%d</boolean></value></param>",x[i].i);
				break;
			case XMLRPC_DOUBLE:
				ww+=swprintf(ww,L"<param><value><double>%f</double></value></param>",x[i].d);
				break;
			case XMLRPC_VECTOR:
				wcscpy(ww,A1);
				ww+=wcslen(A1);
				for (int j=0;j<x[i].nvec;j++) {
					ww+=swprintf(ww,L"<value><i4>%d</i4></value>",x[i].vec[j]);
				}
				wcscpy(ww,A2);
				ww+=wcslen(A2);
				break;
			case XMLRPC_VECTORS:
				wcscpy(ww,A1);
				ww+=wcslen(A1);
				for (int j=0;j<x[i].nvec;j++) {
					ww+=swprintf(ww,L"<value><string>%s</string></value>",x[i].vecs[j]);
				}
				wcscpy(ww,A2);
				ww+=wcslen(A2);
				break;
		}
	}
	wcscpy(ww,P2);
	return w;
}

#define MR1 L"<methodResponse>"
#define MR2 L"</methodResponse>"
#define MN1 L"<methodName>"
#define MN2 L"</methodName>"
#define MC1 L"<methodCall>"
#define MC2 L"</methodCall>"

const xairo_uchar* CXairoProxyServer::RPCRet(int n,xmlrpc_arg * x)
{
	const xairo_uchar* p=RPCParams(n,x);
	int tx=wcslen(p)+wcslen(MR1)+wcslen(MR2);
	xairo_uchar* w=new xairo_uchar[tx+1];
	wcscpy(w,MR1);
	wcscat(w,p);
	wcscat(w,MR2);
	return w;
}

const xairo_uchar* CXairoProxyServer::RPCCall(const xairo_uchar* m,int n,const xmlrpc_arg* x)
{
	const xairo_uchar* p=RPCParams(n,x);
	int tx=wcslen(m)+wcslen(p)+wcslen(MC1)+wcslen(MC2)+wcslen(MN1)+wcslen(MN2);
	xairo_uchar* w=new xairo_uchar[tx+1];
	wcscpy(w,MC1);
	wcscat(w,MN1);
	wcscat(w,m);
	wcscat(w,MN2);
	wcscat(w,p);
	wcscat(w,MC2);
	return w;
}
#endif

#ifdef _WIN32
const xairo_uchar* ErrString(int dw)
{
    LPVOID lpMsgBuf;
    FormatMessage(
        FORMAT_MESSAGE_ALLOCATE_BUFFER | 
        FORMAT_MESSAGE_FROM_SYSTEM,
        NULL,
        dw,
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
        (LPTSTR) &lpMsgBuf,
        0, NULL );
	const xairo_uchar* m=_wcsdup((xairo_uchar*) lpMsgBuf);
    LocalFree(lpMsgBuf);
	return m;
}
#endif

