#include "stdafx.h"
#include "XairoProxySolution.h"
#include "XairoProxyHit.h"
#include "XairoProxyCollocationTable.h"
#include "XairoProxyCorpusTextFilter.h"
#include "XairoProxyAddkeyFilter.h"

CXairoProxySolution::CXairoProxySolution(const xairo_uchar* id,CXairoProxyServer* s):
	m_id(id),m_server(s)
{
}

CXairoProxySolution::~CXairoProxySolution(void)
{
	const xairo_uchar* s;
	xmlrpc_arg x[1];
	x[0].t=XMLRPC_STRING;
	x[0].s=m_id;
	s=RPCCall(L("Solution_release"),1,x);
	m_server->RPCRemote(s);
}

int CXairoProxySolution::getCount()
{
	const xairo_uchar* s;
	DOMElement* d;
	xmlrpc_arg x[1];
	x[0].t=XMLRPC_STRING;
	x[0].s=m_id;
	s=RPCCall(L("Solution_getCount"),1,x);
	d=m_server->RPCRemote(s); delete [] s;
	return RPCIntParam(d,0);
}

CXairoIHit* CXairoProxySolution::getHit(int i)
{
	const xairo_uchar* s;
	DOMElement* d;
	xmlrpc_arg x[2];
	x[0].t=XMLRPC_STRING; x[0].s=m_id;
	x[1].t=XMLRPC_I4; x[1].i=i;
	s=RPCCall(L("Solution_getHit"),2,x);
	d=m_server->RPCRemote(s); delete [] s;
	const xairo_uchar* u=RPCStringParam(d,0);
	int doc=RPCIntParam(d,1);
	int f=RPCIntParam(d,2);
	const xairo_uchar* url=RPCStringParam(d,3);
	unsigned int c1=RPCIntParam(d,4);
	unsigned int c2=RPCIntParam(d,5);
	return new CXairoProxyHit(m_id,i,u,doc,f,url,c1,c2,m_server);
}

int CXairoProxySolution::getCorpusTextCount()
{
	const xairo_uchar* s;
	DOMElement* d;
	xmlrpc_arg x[1];
	x[0].t=XMLRPC_STRING; x[0].s=m_id;
	s=RPCCall(L("Solution_getCorpusTextCount"),1,x);
	d=m_server->RPCRemote(s); delete [] s;
	return RPCIntParam(d,0);
}

CXairoISolution* CXairoProxySolution::thinInit(int n)
{
	const xairo_uchar* s;
	DOMElement* d;
	xmlrpc_arg x[2];
	x[0].t=XMLRPC_STRING; x[0].s=m_id;
	x[1].t=XMLRPC_I4; x[1].i=n;
	s=RPCCall(L("Solution_thinInit"),2,x);
	d=m_server->RPCRemote(s); delete [] s;
	const xairo_uchar* id=RPCStringParam(d,0);
	if (wcscmp(id,L("null"))==0) return NULL;
	else return new CXairoProxySolution(id,m_server);
}

CXairoISolution* CXairoProxySolution::thinRand(int n,int seed)
{
	const xairo_uchar* s;
	DOMElement* d;
	xmlrpc_arg x[3];
	x[0].t=XMLRPC_STRING; x[0].s=m_id;
	x[1].t=XMLRPC_I4; x[1].i=n;
	x[2].t=XMLRPC_I4; x[2].i=seed;
	s=RPCCall(L("Solution_thinRand"),3,x);
	d=m_server->RPCRemote(s); delete [] s;
	const xairo_uchar* id=RPCStringParam(d,0);
	if (wcscmp(id,L("null"))==0) return NULL;
	else return new CXairoProxySolution(id,m_server);
}

CXairoISolution* CXairoProxySolution::thinOnePerText()
{
	const xairo_uchar* s;
	DOMElement* d;
	xmlrpc_arg x[1];
	x[0].t=XMLRPC_STRING; x[0].s=m_id;
	s=RPCCall(L("Solution_thinOnePerText"),1,x);
	d=m_server->RPCRemote(s); delete [] s;
	const xairo_uchar* id=RPCStringParam(d,0);
	if (wcscmp(id,L("null"))==0) return NULL;
	else return new CXairoProxySolution(id,m_server);
}

CXairoISolution* CXairoProxySolution::thinSelection(int n,int* v)
{
	const xairo_uchar* s;
	DOMElement* d;
	xmlrpc_arg x[2];
	x[0].t=XMLRPC_STRING; x[0].s=m_id;
	x[1].t=XMLRPC_VECTOR; x[1].nvec=n; x[1].vec=v;
	s=RPCCall(L("Solution_thinSelection"),2,x);
	d=m_server->RPCRemote(s); delete [] s;
	const xairo_uchar* id=RPCStringParam(d,0);
	if (wcscmp(id,L("null"))==0) return NULL;
	else return new CXairoProxySolution(id,m_server);
}

CXairoISolution* CXairoProxySolution::copy()
{
	const xairo_uchar* s;
	DOMElement* d;
	xmlrpc_arg x[1];
	x[0].t=XMLRPC_STRING; x[0].s=m_id;
	s=RPCCall(L("Solution_copy"),1,x);
	d=m_server->RPCRemote(s); delete [] s;
	const xairo_uchar* id=RPCStringParam(d,0);
	if (wcscmp(id,L("null"))==0) return NULL;
	else return new CXairoProxySolution(id,m_server);
}

/*CXairoISolution* CXairoProxySolution::thinSubcorpus(int nt,int* v)
{
	const xairo_uchar* s;
	DOMElement* d;
	xmlrpc_arg x[2];
	x[0].t=XMLRPC_STRING; x[0].s=m_id;
	x[1].t=XMLRPC_VECTOR; x[1].nvec=nt; x[1].vec=v;
	s=RPCCall(L("Solution_thinSubcorpus"),2,x);
	d=m_server->RPCRemote(s); delete [] s;
	const xairo_uchar* id=RPCStringParam(d,0);
	if (wcscmp(id,L("null"))==0) return NULL;
	else return new CXairoProxySolution(id,m_server);
}
*/

CXairoISolution* CXairoProxySolution::thinCollocation(const xairo_uchar* key,CXairoICollocationTable::_cwin w,int n,const xairo_uchar** sel)
{
	const xairo_uchar* s;
	DOMElement* d;
	xmlrpc_arg x[8];
	x[0].t=XMLRPC_STRING; x[0].s=m_id;
	x[1].t=XMLRPC_VECTORS; x[1].nvec=n; x[1].vecs=sel;
	x[2].t=XMLRPC_STRING; x[2].s=key;
	x[3].t=XMLRPC_BOOL; x[3].i=w.bX?1:0;
	x[4].t=XMLRPC_I4; x[4].i=w.wl;
	x[5].t=XMLRPC_I4; x[5].i=w.wr;
	x[6].t=XMLRPC_STRING; x[6].s=w.bX?w.s:L("");
	x[7].t=XMLRPC_STRING; x[7].s=w.bX?w.ns:L("");
	s=RPCCall(L("Solution_thinCollocation"),8,x);
	d=m_server->RPCRemote(s); delete [] s;
	const xairo_uchar* id=RPCStringParam(d,0);
	if (wcscmp(id,L("null"))==0) return NULL;
	else return new CXairoProxySolution(id,m_server);
}

CXairoISolution* CXairoProxySolution::thinCorpusTextFilter(const CXairoICorpusTextFilter* f)
{
	const xairo_uchar* s;
	DOMElement* d;
	xmlrpc_arg x[2];
	x[0].t=XMLRPC_STRING; x[0].s=m_id;
	x[1].t=XMLRPC_STRING; x[1].s=((CXairoProxyCorpusTextFilter*) f)->getId();
	s=RPCCall(L("Solution_thinCorpusTextFilter"),2,x);
	d=m_server->RPCRemote(s); delete [] s;
	const xairo_uchar* id=RPCStringParam(d,0);
	if (wcscmp(id,L("null"))==0) return NULL;
	else return new CXairoProxySolution(id,m_server);
}

CXairoISolution* CXairoProxySolution::thinAddkeyFilter(const CXairoIAddkeyFilter* f)
{
	const xairo_uchar* s;
	DOMElement* d;
	xmlrpc_arg x[2];
	x[0].t=XMLRPC_STRING; x[0].s=m_id;
	x[1].t=XMLRPC_STRING; x[1].s=((CXairoProxyAddkeyFilter*) f)->getId();
	s=RPCCall(L("Solution_thinAddkeyFilter"),2,x);
	d=m_server->RPCRemote(s); delete [] s;
	const xairo_uchar* id=RPCStringParam(d,0);
	if (wcscmp(id,L("null"))==0) return NULL;
	else return new CXairoProxySolution(id,m_server);
}

CXairoICollocationTable* CXairoProxySolution::makeCollocationTable(const xairo_uchar* pattern,const xairo_uchar* key,CXairoICollocationTable::cmeth m,CXairoICollocationTable::_cwin w,int maxn,double mins,double maxs,int lowf)
{
	const xairo_uchar* s;
	DOMElement* d;
	xmlrpc_arg x[13];
	x[0].t=XMLRPC_STRING; x[0].s=m_id;
	x[1].t=XMLRPC_STRING; x[1].s=pattern;
	x[2].t=XMLRPC_STRING; x[2].s=key;
	x[3].t=XMLRPC_I4; x[3].i=m;
	x[4].t=XMLRPC_BOOL; x[4].i=w.bX?1:0;
	x[5].t=XMLRPC_I4; x[5].i=w.wl;
	x[6].t=XMLRPC_I4; x[6].i=w.wr;
	x[7].t=XMLRPC_STRING; x[7].s=w.bX?w.s:L("");
	x[8].t=XMLRPC_STRING; x[8].s=w.bX?w.ns:L("");
	x[9].t=XMLRPC_I4; x[9].i=maxn;
	x[10].t=XMLRPC_DOUBLE; x[10].d=mins;
	x[11].t=XMLRPC_DOUBLE; x[11].d=maxs;
	x[12].t=XMLRPC_I4; x[12].i=lowf;
	s=RPCCall(L("Solution_makeCollocationTable"),13,x);
	d=m_server->RPCRemote(s); delete [] s;
	const xairo_uchar* id=RPCStringParam(d,0);
	if (wcscmp(id,L("null"))==0) return NULL;
	else return new CXairoProxyCollocationTable(id,m_server);
}

void CXairoProxySolution::getAnalysis(int np,int* p,int n,int* a,int* b)
{
	const xairo_uchar* s;
	DOMElement* d;
	xmlrpc_arg x[3];
	x[0].t=XMLRPC_STRING; x[0].s=m_id;
	x[1].t=XMLRPC_VECTOR; x[1].nvec=np; x[1].vec=p;
    x[2].t=XMLRPC_I4; x[2].i=n;
	s=RPCCall(L("Solution_getAnalysis"),3,x);
	d=m_server->RPCRemote(s); delete [] s;
	int* v;
	int m=RPCVectorParam(d,0,v);
	for (int i=0;i<m;i++)
		a[i]=v[i];
	int* vt;
	int mt=RPCVectorParam(d,1,vt);
	for (int i=0;i<mt;i++)
		b[i]=vt[i];
}

void CXairoProxySolution::getAnalysisScope(int np,int* p,int n,int* a,int* b,const xairo_uchar* e)
{
	const xairo_uchar* s;
	DOMElement* d;
	xmlrpc_arg x[4];
	x[0].t=XMLRPC_STRING; x[0].s=m_id;
	x[1].t=XMLRPC_VECTOR; x[1].nvec=np; x[1].vec=p;
    x[2].t=XMLRPC_I4; x[2].i=n;
	x[3].t=XMLRPC_STRING; x[3].s=e;
	s=RPCCall(L("Solution_getAnalysisScope"),4,x);
	d=m_server->RPCRemote(s); delete [] s;
	int* v;
	int m=RPCVectorParam(d,0,v);
	for (int i=0;i<m;i++)
		a[i]=v[i];
	int* vt;
	int mt=RPCVectorParam(d,1,vt);
	for (int i=0;i<mt;i++)
		b[i]=vt[i];
}

const xairo_uchar* CXairoProxySolution::serialize()
{
	return m_id;
}



