#include "stdafx.h"
#include "XairoProxyWord.h"
#include "XairoProxyForm.h"
#include "XairoProxyCorpusTextFilter.h"
#include "XairoProxyAddkeyFilter.h"

CXairoProxyWord::CXairoProxyWord(const xairo_uchar* id,int f,int fc,const xairo_uchar* norm0,CXairoProxyServer *s):
	m_id(id),m_f(f),m_fc(fc),m_norm0(norm0),m_server(s)
{
	m_bCache=true;
}

CXairoProxyWord::CXairoProxyWord(const xairo_uchar* id,CXairoProxyServer *s):
	m_id(id),m_server(s)
{
	m_bCache=false;
}

CXairoProxyWord::~CXairoProxyWord(void)
{
	delete [] m_id;
}


const xairo_uchar* CXairoProxyWord::getNorm(int i) 
{
	if (i==0 && m_bCache) return m_norm0;
	const xairo_uchar* s;
	DOMElement* d;
	xmlrpc_arg x[2];
	x[0].t=XMLRPC_STRING; x[0].s=m_id;
	x[1].t=XMLRPC_I4; x[1].i=i;
	s=RPCCall(L("Word_getNorm"),2,x);
	d=m_server->RPCRemote(s); delete [] s;
	return RPCStringParam(d,0);
}

int CXairoProxyWord::getFrequency() 
{
	if (m_bCache) return m_f;
	const xairo_uchar* s;
	DOMElement* d;
	xmlrpc_arg x[1];
	x[0].t=XMLRPC_STRING; x[0].s=m_id;
	s=RPCCall(L("Word_getFrequency"),1,x);
	d=m_server->RPCRemote(s); delete [] s;
	return RPCIntParam(d,0);
}

int CXairoProxyWord::getFilteredFrequency(const CXairoICorpusTextFilter* f,const CXairoIAddkeyFilter* ff) 
{
	const xairo_uchar* s;
	DOMElement* d;
	xmlrpc_arg x[3];
	x[0].t=XMLRPC_STRING; x[0].s=m_id;
	x[1].t=XMLRPC_STRING; x[1].s=(f==NULL)?L("null"):((CXairoProxyCorpusTextFilter*) f)->getId();
	x[2].t=XMLRPC_STRING; x[2].s=(ff==NULL)?L("null"):((CXairoProxyAddkeyFilter*) ff)->getId();
	s=RPCCall(L("Word_getFilteredFrequency"),3,x);
	d=m_server->RPCRemote(s); delete [] s;
	return RPCIntParam(d,0);
}

int CXairoProxyWord::getFormCount()  
{
	if (m_bCache) return m_fc;
	const xairo_uchar* s;
	DOMElement* d;
	xmlrpc_arg x[1];
	x[0].t=XMLRPC_STRING; x[0].s=m_id;
	s=RPCCall(L("Word_getFormCount"),1,x);
	d=m_server->RPCRemote(s); delete [] s;
	return RPCIntParam(d,0);
}

CXairoIForm* CXairoProxyWord::getForm(int i)
{
	const xairo_uchar* s;
	DOMElement* d;
	xmlrpc_arg x[2];
	x[0].t=XMLRPC_STRING; x[0].s=m_id;
	x[1].t=XMLRPC_I4; x[1].i=i;
	s=RPCCall(L("Word_getForm"),2,x);
	d=m_server->RPCRemote(s); delete [] s;
	return new CXairoProxyForm(RPCStringParam(d,0),m_server);
}

